#include "./ConditionalVariable.h"
#if defined(__OS_X__)
#include <sys/time.h>
#endif

#if defined(__POSIX__)
#include <errno.h>
#endif

#include <stdio.h>

ConditionalVariable::ConditionalVariable(void)
{
#if defined(_MSC_VER) || defined(__BORLANDC__)
#if defined(WINAPI_FAMILY) && WINAPI_FAMILY==WINAPI_FAMILY_PHONE_APP
    m_Event = CreateEventEx(NULL, NULL, 0, 0);
#else
    m_Event = CreateEvent(NULL, FALSE, FALSE, NULL);
#endif
    ResetEvent(m_Event);
#endif

#if defined(_POSIX_THREADS)
	m_Signaled = false;
	pthread_mutex_init(&m_MutexId, NULL);
 	pthread_cond_init(&m_CondVar, NULL);
#endif

#if defined(__FREERTOS__)
	vSemaphoreCreateBinary(m_Semaphore);
	xSemaphoreTake(m_Semaphore, 0);	
#endif
}

//-------------------------------------------------------------------------------------------------

ConditionalVariable::~ConditionalVariable(void)
{
#if defined(_MSC_VER) || defined(__BORLANDC__)
	CloseHandle(m_Event);
#endif

#if defined(_POSIX_THREADS)
	pthread_mutex_destroy(&m_MutexId);
	pthread_cond_destroy(&m_CondVar);
#endif

#if defined(__FREERTOS__)
	vQueueDelete(m_Semaphore);
#endif
}

//-------------------------------------------------------------------------------------------------

void ConditionalVariable::Wait(void)
{
#if defined(_MSC_VER) || defined(__BORLANDC__)
#if WINAPI_FAMILY==WINAPI_FAMILY_PHONE_APP
    WaitForSingleObjectEx(m_Event, INFINITE, FALSE);
#else
    DWORD res = WaitForSingleObject(m_Event, INFINITE);
#endif
#endif

#if defined(_POSIX_THREADS)
	pthread_mutex_lock(&m_MutexId);
	if (!m_Signaled)
	{
		pthread_cond_wait(&m_CondVar, &m_MutexId);
	}
	m_Signaled = false;
	pthread_mutex_unlock(&m_MutexId);
#endif

#if defined(__FREERTOS__)	
	xSemaphoreTake(m_Semaphore, portMAX_DELAY);	
#endif
}

// Vraci true, pokud doslo k udalosti
bool ConditionalVariable::Wait(uint32b milisecsTimeout)
{
#if defined(_MSC_VER) || defined(__BORLANDC__)
#if WINAPI_FAMILY==WINAPI_FAMILY_PHONE_APP
    DWORD res = WaitForSingleObjectEx(m_Event, milisecsTimeout, FALSE);
#else
    DWORD res = WaitForSingleObject(m_Event, milisecsTimeout);
#endif
    return (res == WAIT_OBJECT_0);
#endif

#if defined(_POSIX_THREADS)
	bool isTimedOut = false;
	
	pthread_mutex_lock(&m_MutexId);

	struct timespec ts;

#if !defined(__OS_X__)
    clock_gettime(CLOCK_REALTIME, &ts);
#else
    struct timeval tv;
    gettimeofday(&tv, NULL);
    ts.tv_sec = tv.tv_sec + 0;
    ts.tv_nsec = 0;
#endif
    ts.tv_sec += milisecsTimeout / 1000;
	ts.tv_nsec += (milisecsTimeout % 1000) * 1000;
	
	if (!m_Signaled)
	{
        if (pthread_cond_timedwait(&m_CondVar, &m_MutexId, &ts) == ETIMEDOUT)
		{
            isTimedOut = true;
		}
	}
	
	m_Signaled = false;
	pthread_mutex_unlock(&m_MutexId);

	return (!isTimedOut);
#endif

#if defined(__FREERTOS__)	
	return (xSemaphoreTake(m_Semaphore, milisecsTimeout / portTICK_RATE_MS));
#endif
}

void ConditionalVariable::Signal(void)
{
#if defined(_MSC_VER) || defined(__BORLANDC__)
	SetEvent(m_Event);
#endif

#if defined(_POSIX_THREADS)
	pthread_mutex_lock(&m_MutexId);
	m_Signaled = true;
	pthread_cond_signal(&m_CondVar);
	pthread_mutex_unlock(&m_MutexId);
#endif

#if defined(__FREERTOS__)		
	xSemaphoreGive(m_Semaphore);
#endif
}

#if defined(CONDITIONALVARIABLE_SELFTEST)

static CriticalSection printfCS;

void ConditionalVariable::Test()
{
	class WaitThread : public Thread
	{
	protected:
		ConditionalVariable *m_CondVar;
		const char *m_Id;
	public:
		WaitThread(ConditionalVariable *condVar, const char *id)
		{
			m_CondVar = condVar;
			m_Id = id;
		}

		void Execute(void)
		{
			printfCS.Enter();
			SharkPrintf("Thread '%s' is going to wait for conditional variable\n", m_Id);
			printfCS.Leave();
			m_CondVar->Wait();
			printfCS.Enter();
			SharkPrintf("Thread '%s' was waked up on conditional variable\n", m_Id);
			printfCS.Leave();
		}
	};

	class SignalThread : public Thread
	{
	protected:
		ConditionalVariable *m_CondVar;
		const char *m_Id;
	public:
		SignalThread(ConditionalVariable *condVar, const char *id)
		{
			m_CondVar = condVar;
			m_Id = id;
		}

		void Execute(void)
		{
			printfCS.Enter();
			SharkPrintf("Thread '%s' is going to signal conditional variable\n", m_Id);
			printfCS.Leave();
			m_CondVar->Signal();			
		}
	};

	ConditionalVariable condVar;
	WaitThread waitThread1(&condVar, "Thread 1");
	WaitThread waitThread2(&condVar, "Thread 2");
	SignalThread signalThread(&condVar, "Signal Thread");

	SharkPrintf("ConditionalVariable::Test 1 - first wait then signal\n");
	waitThread1.Start();
	signalThread.Start();
	signalThread.Join();
	waitThread1.Join();

	SharkPrintf("ConditionalVariable::Test 2  - first signal then wait\n");
	signalThread.Start();
	waitThread1.Start();
	signalThread.Join();
	waitThread1.Join();

	SharkPrintf("ConditionalVariable::Test 3 - start 2 threads to wait then signal wait 3 sec and signal again\n");
	waitThread1.Start();
	waitThread2.Start();
	signalThread.Start();
	signalThread.Join();
	SystemCalls::Sleep(3000);
	signalThread.Start();
	waitThread1.Join();
	waitThread2.Join();
	signalThread.Join();
}

#endif
