#ifndef __ConditionalVariable_h
#define __ConditionalVariable_h

//#define CONDITIONALVARIABLE_SELFTEST

#include "./PlatformUtils.h"

#if defined(CONDITIONALVARIABLE_SELFTEST)
#include "./Thread.h"
#include "./SharkPrintf.h"
#include "./SystemCalls.h"
#include "./CriticalSection.h"
#endif

#if defined(_POSIX_THREADS)
#include <pthread.h>
#endif

#if defined(__FREERTOS__)
#include <FreeRTOS.h>
#include <semphr.h>
#endif

/*!
 * \brief Class emulating the event
 * \details Class represent simple event functionality.
 * \code{.cpp}
 *  class WaitThread : public Thread
 *  {
 *  protected:
 *  	ConditionalVariable *m_CondVar;
 *  public:
 *  	WaitThread(ConditionalVariable *condVar)
 *  	{
 *          m_CondVar = condVar;
 *  	}
 *
 *  	void Execute(void)
 *  	{
 *          // Some code
 *          // Wait until continue
 *          m_CondVar->Wait();
 *          // Some code
 *  	}
 *  };
 *
 *  class Test
 *  {
 *  protected:
 *  	ConditionalVariable m_CondVar;
 *  public:
 *  	void Execute(void)
 *  	{
 *          // Create and run wait thread
 *          WaitThread waitThread(&m_CondVar);
 *          waitThread.Start();
 *          // Some code
 *          // Signal
 *          m_CondVar.Signal();
 *          waitThread.Join();
 *  	}
 *  };
 * \endcode
 */
class ConditionalVariable
{
private:
#if defined(_MSC_VER) || defined(__BORLANDC__)
    /*!
     * \brief Event object
     * \details Object used for rise event. Used only with MS VS and borland C compiler.
     */
	HANDLE m_Event;
#endif

#if defined(_POSIX_THREADS)
    /*!
     * \brief Mutex ID
     * \details ID for created mutex. Used only with gcc compiler (pthread).
     */
    pthread_mutex_t m_MutexId;
    /*!
     * \brief Condition variable
     * \details Synchronization device that allows threads to suspend execution and relinquish the processors until some predicate on shared data is satisfied. Used only with gcc compiler (pthread).
     */
    pthread_cond_t m_CondVar;
    /*!
     * \brief Signalized flag
     * \details Flag for calling \see Signal function.
     */
    volatile bool m_Signaled;
#endif

#if defined(__FREERTOS__)
    /*!
     * \brief Semaphore
     * \details Semaphore for synchronize threads. Used only for FreeRTOS compiler.
     */
    xSemaphoreHandle m_Semaphore;
#endif

public:
    /*!
     * \brief Constructor
     * \details Create all instances and initilazations.
     */
    ConditionalVariable(void);
    /*!
     * \brief Destructor
     * \details Destroy all created istances and made cleenups.
     */
    ~ConditionalVariable(void);

    /*!
     * \brief Wait
     * \details Infinite wait for Signal function. This function resets signaled flag.
     * \see Signal
     */
    void Wait(void);

    /*!
     * \brief Configured wait
     * \details This function wait for signal until \c milisecsTimeout or signal reached. This function resets signaled flag.
     * \param[in] milisecsTimeout Timeout in ms.
     * \returns
     * \li \c true - Signal reached
     * \li \c false - Timeout reached
     * \see Signal
     */
    bool Wait(uint32b milisecsTimeout);
    /*!
     * \brief Signal
     * \details Sets signaled flag for interrupt Wait function.
     * \see Wait
     */
    void Signal(void);

#if defined(CONDITIONALVARIABLE_SELFTEST)
    /*!
     * \brief Test function
     * \details Function for testing class functionality. Used only with CONDITIONALVARIABLE_SELFTEST definition.
     */
    static void Test();
#endif
};
#endif
