#ifndef __CriticalSection_h
#define __CriticalSection_h

#include "./PlatformUtils.h"

#if defined(_POSIX_THREADS)
#include <pthread.h>
#endif

#if defined(__FREERTOS__)
extern "C" {
#include <FreeRTOS.h>
#include <semphr.h>
#include <task.h>
}
#endif

// fimxme: michale, proc je tady stdio?
// snazme se nezanaset .h spoustou #includu, jinak z toho vznikne propletenec, kde nebudeme vedet, odkud to ma vlastne .cpp naincludovano
// a vymazanim nejakeho nevinneho includu to ele rozbijeme (napriklad tenhle vidim jako zbytecny v .h)
#include <stdio.h>

/*!
 * \brief Class for manage critical sections
 * \details Class for thread synchronization. This class is independet on compiler.
 * \code{.cpp}
 *  class MainThread : public Thread
 *  {
 *  protected:
 *  	CriticalSection *m_CritSec;
 *  public:
 *  	WaitThread(CriticalSection *critSec)
 *  	{
 *          m_CritSec = critSec;
 *  	}
 *
 *  	void Execute(void)
 *  	{
 *          // Some code
 *          m_CritSec->Enter();
 *          // Some code
 *          m_CritSec->Leave();
 *  	}
 *  };
 *
 *  class Test
 *  {
 *  protected:
 *  	CriticalSection m_CritSec;
 *  public:
 *  	void Execute(void)
 *  	{
 *          // Create threads
 *          MainThread thread1(&m_CritSec);
 *          MainThread thread2(&m_CritSec);
 *          thread1.Start();
 *          thread2.Start();
 *          thread1.Join();
 *          thread2.Join();
 *  	}
 *  };
 * \todo SERVER!!!! - Odkomentovat
 * \endcode
 */
class CriticalSection
{
private:
#if defined(_MSC_VER) || defined(__BORLANDC__)
    /*!
     * \brief Critical Section
     * \details Object used for manage critical section. Used only with MS VS and borland C compiler.
     */
	CRITICAL_SECTION m_CriticalSection;
#endif

#if defined(_POSIX_THREADS)
    /*!
     * \brief m_MutexId
     * \details Object used for manage critical section. Used only with gcc compiler.
     */
	pthread_mutex_t m_MutexId;
#endif

#if defined(__FREERTOS__)
    /*!
     * \brief m_Semaphore
     * \details Semaphore for manage critical section. Used only for FreeRTOS compiler.
     */
	xSemaphoreHandle m_Semaphore;
#endif

public:
    /*!
     * \brief Constructor
     * \details Creates object instances for managing critical section.
     */
	CriticalSection(void);
    /*!
     * \brief Destructor
     * \details Release all created instancess and cler memory
     */
    ~CriticalSection(void);

    /*!
     * \brief Enter
     * \details Lock object for critical section. After call Leave function, next Enter call pass.
     * \see Leave
     */
	void Enter(void);
    /*!
     * \brief Leave
     * \details Release object for critical section.
     */
	void Leave(void);
};

#endif

