/*!
 * \file EngineDefinitions.h
 * File denines enums and struct for SSCP communication
 */

#ifndef __EngineDefinitions_h
#define __EngineDefinitions_h

#if defined(STANDALONE_SUPPORT)
#include "./OperationStatus.h"
#else
#include "./../../Shared/Utils/OperationStatus.h"
#endif

#define SEX_IMAGE_GUID_LEN 16

/*!
 * \brief System commands
 * \details Enums for usable commands for PLC
 * \todo SERVER!!!! - Odkomentovat
 */
enum SystemCommand
{
    /*!
     * \brief Warm restart
     * \details Restarts the device in a "warm mode" (permanent variables keep their values after the boot).
     *
     * Parameters: N/A
     */
    SystemCommand_WarmRestart = 0,
    /*!
     * \brief Cold restart
     * \details Restarts the device in a "cold mode" (permanent variables get initial values after the boot)
     *
     * Parameters: N/A
     */
    SystemCommand_ColdRestart = 1,
    /*!
     * \brief Accept new image
     * \details RT is reinitialized with actual image in the memory.
     *
     * Parameters:
     * - Flags (1B)
     *      - \c D0 - Try online changes
     *      - \c D1 - Clear retain variables
     */
    SystemCommand_AcceptNewImage = 2,
    /*!
     * \brief Create boot project
     * \details RT is reinitialized with actual image in the memory.
     *
     * Parameters:
     * - image area nr. (1 B)
     */
    SystemCommand_CreateBootProject = 3,
    /*!
     * \brief Set active partition
     * \details Sets the specified partition area as active.
     *
     * Parameters:
     * - image area nr. (1 B)
     */
    SystemCommand_SetActivePartition = 4,
    /*!
     * \brief Start PLC application
     * \details Starts the application if it is not running. Reasons of not running previous call to stop, excepion state etc.
     *
     * Parameters:
     * - Run mode (1 B)
     *      - \c Full
     *      - \c Comm
     *      - \c Eval
     *
     * - Flags (1 B)
     *      - \c DO - Clear variables saved to flash
     */
    SystemCommand_StartPlcApplication = 5,
    /*!
     * \brief Stop PLC application
     * \details Stops all tasks gracefully. Data are not available in this mode.
     *
     * Parameters: N/A
     */
    SystemCommand_StopPlcApplication = 6,
    /*!
     * \brief Set MAC address
     * \details Sets the MAC Addres of PLC. Only valid for uPLC.
     *
     * Parameters:
     * - MAC address (8 B)
     */
    SystemCommand_SetMacAddress = 7,
    /*!
     * \brief Clear WEB driectory
     * \details Clear Web directory.
     *
     * Parameters: N/A
     */
    SystemCommand_ClearWebDirectory = 8,
    /*!
     * \brief Format filesystem
     * \details Format selected filesystem area.
     *
     * Parameters:
     * - File system ID (1 B)
     *      - \c System
     *      - \c User
     */
    SystemCommand_FormatFilesystem = 9,
    /*!
     * \brief Reset to factory defaults
     * \details Set all properties to factory defaults
     *
     * Parameters: N/A
     */
    SystemCommand_ResetToFactoryDefaults = 10,
    /*!
     * \brief Save history
     * \details Save history
     *
     * Parameters: N/A
     * \todo Doplnit detailní popis
     */
    SystemCommand_SaveHistory = 11,
    /*!
     * \brief Port monitor command
     * \details Sends commands for port monitors defined canal. GetStatus function returns permissions PM (if enabled, return error \c SSCP \c Already \c running)
     *
     * Parameters:
     * - Channel ID (4 B)
     * - Command (1 B)
     *      - \c Enable
     *      - \c Disable
     *      - \c Clear
     *      - \c GetStatus
     */
    SystemCommand_PortMonitorCommand = 12,
    /*!
     * \brief Delete boot project
     * \details Erase area for specified boot image
     *
     * Parameters:
     * - image area nr. (1 B)
     */
    SystemCommand_DeleteBootProject = 13,
    /*!
     * \brief Dummy query
     * \details Used to query state of last operation
     *
     * Parameters: N/A
     */
    SystemCommand_DummyQuery = 0xff // command for querying if command was processed
};

/*!
 * \brief Run mode
 * \details Mode for running application in RT
 * \todo SERVER!!!! - Odkomentovat
 */
enum RunMode
{
    /*!
     * \brief Full run
     * \details RT runs in full mode (regulation is on)
     */
    RunMode_FullRun = 0,
    /*!
     * \brief Communication only
     * \details RT runs in communication mode (regulation is off)
     */
    RunMode_CommunicationOnly = 1,
    /*!
     * \brief Communications with transform
     * \details Like RunMode_CommunicationOnly, but property transformation is enabled
     */
    RunMode_CommunicationsWithTransform = 4,
    /*!
     * \brief Evaluation only
     * \details RT runs in evaluation mode
     */
    RunMode_EvaluationOnly = 2,
    /*!
     * \brief Commissioning
     * \details RT is running in commissioning mode
     */
    RunMode_Commissioning = 3
};

/*!
 * \brief System command params
 * \details Parametres for sytem commands
 * \sa SystemCommand
 * \todo SERVER!!!! - Odkomentovat
 */
union SystemCommandParams
{
    /*!
     * \brief sl
     * \todo Doplnit k čemu to je
     */
	struct
	{
		bool OnlineChanges;
		bool ForceClearRetain;
		bool AutomaticStart; // pouzito z MyIO request klienta, jinde pokud mozno nepouzivat !
	} s1;
    /*!
     * \brief Boot image area number
     * \todo Doplnit k čemu to je
     */
    uint8b BootImageAreaNumber;
    /*!
     * \brief Parameter long
     * \todo Doplnit k čemu to je
     */
    uint64b ParameterLong;
    /*!
     * \brief Start plc run mode
     * \details Run mode for command SystemCommand_StartPlcApplication
     * \sa RunMode, SystemCommand_StartPlcApplication
     */
    RunMode StartPlcRunMode;
    /*!
     * \brief File system id
     * \details File system ID for command SystemCommand_FormatFilesystem
     *
     * - \c System = 0
     * - \c User = 1
     *
     * \sa SystemCommand, SystemCommand_FormatFilesystem
     */
    uint8b FileSystemId;
};

/*!
 * \brief Command request
 * \details Struct for queue commad request. It holds information about command (type, params, return code)
 * \todo SERVER!!!! - Odkomentovat
 */
struct CommandRequest
{
    /*!
     * \brief Request active
     * \details Active flag. If is it true, command is executed otherwise is skipped
     */
	bool RequestActive;

    /*!
     * \brief Command
     * \details Executing command
     * \sa SystemCommand
     */
	SystemCommand Command;
    /*!
     * \brief Command params
     * \details Parameters for executed commad
     * \sa SystemCommandParams
     */
    SystemCommandParams CommandParams;
    /*!
     * \brief Result
     * \details Returned value for called command
     * \todo Error codes???
     */
	uint32b Result;
};

/*!
 * \brief Evaluator state
 * \details Evaluator state for get PLC Statisics function positive respond
 * \todo SERVER!!!! - Odkomentovat
 */
enum EvaluatorState
{
    /*!
     * \brief Stopped
     * \todo Detaily??
     */
    EvaluatorState_Stopped = 0,
    /*!
     * \brief Running normal tasks
     * \todo Detaily??
     */
    EvaluatorState_RunningNormalTasks = 1,
    /*!
     * \brief Stopping execution
     * \todo Detaily??
     */
    EvaluatorState_StoppingExecution = 2,
    /*!
     * \brief Running exception state task
     * \todo Detaily??
     */
    EvaluatorState_RunningExceptionStateTask = 3,
    /*!
     * \brief Exception state task failed
     * \todo Detaily??
     */
    EvaluatorState_ExceptionStateTaskFailed = 4,
    /*!
     * \brief No exception state task defined
     * \todo Detaily??
     */
    EvaluatorState_NoExceptionStateTaskDefined = 5,
    /*!
     * \brief Commissioning
     * \todo Detaily??
     */
    EvaluatorState_Commissioning = 6,
    /*!
     * \brief Invalid image
     * \todo Detaily??
     */
    EvaluatorState_InvalidImage = 7,
    /*!
     * \brief No image
     * \todo Detaily??
     */
    EvaluatorState_NoImage = 8
};

//set read only variables - error kody a error chovani
//s cistou hlavou zase projet stavy pri soucasnem vypinani RT a padu do exception apod.
//signalizace neuspechu pri komunikaci pres SSCP

/*!
 * \brief Evaluator state mask
 * \details Masks for Evaluator state
 * \todo SERVER!!!! - Odkomentovat
 */
enum EvaluatorStateMask
{
	MaskStopped = (uint32b)1 << (uint32b)EvaluatorState_Stopped,
	MaskRunningNormalTasks = (uint32b)1 << (uint32b)EvaluatorState_RunningNormalTasks,
	MaskStoppingExecution = (uint32b)1 << (uint32b)EvaluatorState_StoppingExecution,
	MaskRunningExceptionStateTask = (uint32b)1 << (uint32b)EvaluatorState_RunningExceptionStateTask,
	MaskExceptionStateTaskFailed = (uint32b)1 << (uint32b)EvaluatorState_ExceptionStateTaskFailed,
	MaskNoExceptionStateTaskDefined = (uint32b)1 << (uint32b)EvaluatorState_NoExceptionStateTaskDefined,
	MaskCommissioning = (uint32b)1 << (uint32b)EvaluatorState_Commissioning,
	MaskInvalidImage = (uint32b)1 << (uint32b)EvaluatorState_InvalidImage,
	MaskNoImage = (uint32b)1 << (uint32b)EvaluatorState_NoImage,

	AllStates = MaskStopped | MaskRunningNormalTasks | MaskStoppingExecution | MaskRunningExceptionStateTask | 
		MaskExceptionStateTaskFailed | MaskNoExceptionStateTaskDefined | MaskCommissioning | MaskInvalidImage | MaskNoImage,

	StatesForStartPlcApplication = MaskStopped | MaskCommissioning,
	StatesForStopPlcApplication = MaskRunningNormalTasks | MaskRunningExceptionStateTask | MaskExceptionStateTaskFailed | MaskNoExceptionStateTaskDefined,
	StatesForStartCommissioning = StatesForStartPlcApplication | StatesForStopPlcApplication,
	StatesForAcceptNewImage = MaskStopped | MaskRunningNormalTasks | MaskStoppingExecution | MaskRunningExceptionStateTask | 
		MaskExceptionStateTaskFailed | MaskNoExceptionStateTaskDefined | MaskCommissioning | MaskInvalidImage | MaskNoImage,
	StatesForCreateBootProject = MaskStopped | MaskRunningNormalTasks | MaskStoppingExecution | MaskRunningExceptionStateTask | 
		MaskExceptionStateTaskFailed | MaskNoExceptionStateTaskDefined | MaskCommissioning,
	StatesForGetTaskStatistics = MaskRunningNormalTasks | MaskRunningExceptionStateTask | MaskExceptionStateTaskFailed | MaskNoExceptionStateTaskDefined,
	StatesForGetStacks = MaskRunningNormalTasks | MaskRunningExceptionStateTask | MaskExceptionStateTaskFailed | MaskNoExceptionStateTaskDefined,
	StatesForGetImageInfo = MaskRunningNormalTasks | MaskRunningExceptionStateTask | MaskExceptionStateTaskFailed | MaskNoExceptionStateTaskDefined |
		MaskCommissioning,
	StatesForWorkWithValues = MaskRunningNormalTasks | MaskRunningExceptionStateTask | MaskExceptionStateTaskFailed | MaskNoExceptionStateTaskDefined,
};
/*!
 * \brief Stack type
 * \details Debug info call stack type
 * \todo SERVER!!!! - Odkomentovat
 */
enum StackType
{
    /*!
     * \brief 0x80
     * \details Exception call stack
     */
	ExceptionCallStack = 0x80,
    /*!
     * \brief 0x81
     * \details Exception computational stack
     */
    ExceptionCompStack = 0x81
};
/*!
 * \brief The Get stacks error
 * \details Error codes
 * \todo SERVER!!!! - Odkomentovat
 */
enum GetStacksError
{
	NoError = 0,
	InvalidTask = 1,
	InvalidStackType = 2,
	OffsetOutOfRange = 3
};

// fixme: vypinani vsech apod.
/*!
 * \brief Set read only mode params
 * \details
 * \todo SERVER!!!! - Odkomentovat
 */
enum SetReadOnlyModeParams
{
	ForceReadOnly,
	ForceWriteable,
	SetWritableOnlyIfNotReadOnlyByDefault,
	SetToDefault,	
};

namespace StatisticsParams
{
    enum StatisticsParams
	{
        /*!
         * \brief Total heap
         * \details Total heap memory available for runtime
         */
        TotalHeap = 0,
        /*!
         * \brief Free heap
         * \details Free heap after image load and process
         */
        FreeHeap,
        /*!
         * \brief Total space for program
         * \details Total space available for image
         */
        TotalSpaceForProgram,
        /*!
         * \brief Free space
         * \details Free space after image save
         */
        FreeSpace,
        /*!
         * \brief Section VMEX used
         * \details Heap memory consumed by VMEX section (VM image)
         */
        SectionVmexUsed,
        /*!
         * \brief Section CIDM used
         * \details
         * \todo doplnit detailní popis
         */
        SectionCidmUsed,
        /*!
         * \brief Section RTCM used
         * \details Heap memory consumed by communication section
         */
        SectionRtcmUsed,
	};
}

#endif

