/*!
 * \file NetworkingHelpers.h
 * Network communications definitions and classes
 */

#ifndef __NetworkingHelpers_h
#define __NetworkingHelpers_h

#include "./datatypes.h"
#include "PlatformUtils.h"
#if !defined(STANDALONE_SUPPORT)
#include "../Utils/TimeTransformations.h"
#endif
#include "./OperationStatus.h"
#if defined(__FREERTOS__)
#include "WatchdogManager.h"
#endif

#if defined(_MSC_VER) || defined(__BORLANDC__)

#include <winsock2.h>

typedef SOCKET SharkSocket;

#define FD_SET_TYPE	struct fd_set
#else

#if defined (__PIKEOS__) || defined (__FREERTOS__)

#include <lwip/inet.h>
#include <lwip/sockets.h>

// Definice funkci z lwIP stacku
#define closesocket	lwip_close
#define ioctlsocket	lwip_ioctl

#else
#if defined(__POSIX__)
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>

#define closesocket	close
#define ioctlsocket	ioctl

#endif
#endif

typedef sint32b SharkSocket;

#define INVALID_SOCKET	-1
#define SOCKET_ERROR	-1

#define FD_SET_TYPE	fd_set
#define SOCKADDR	struct sockaddr

#endif

// fixme: pro jine platformy definovat INVALID_SOCKET jako hodnotu SharkSocket

#if defined(__FREERTOS__)
/*!
 * \brief Networking manager
 * \details Used only with __FREERTOS__ definition
 * \todo Doplnit!
 */
class NetworkingManager : public Thread
{
protected:
	unsigned long m_IpAddr;
	unsigned long m_Netmask;
	unsigned long m_Gateway;
	unsigned long m_PrimaryDns;
	bool m_UseDhcp;
	bool m_Initialized;

public:
	NetworkingManager(unsigned long ip_addr, unsigned long netmask, unsigned long gateway, unsigned long primaryDns, bool useDhcp);
	~NetworkingManager();

	void Execute(void);
};
#endif

#if !defined(SSCP_STANDALONE_CLIENT)
/*!
 * \brief Networking helpers
 * \details Used only without SSCP_STANDALONE_CLIENT definition
 * \todo SERVER!!!! - Odkomentovat
 */
class NetworkingHelpers
{
public:
	// fixme: vylepsit error handling u vsech fci
	static bool InitializeNetworking(unsigned long ip_addr, unsigned long netmask, unsigned long gateway, unsigned long primaryDns, bool useDhcp);
	static bool DeinitializeNetworking(void);
	static OperationStatus SimpleConnect(uint32b host, uint32b port, bool tcp, SharkSocket *s);
	static OperationStatus StartListening(uint32b port, uint8b queueLength, bool tcp, SharkSocket *s);
	static SharkSocket AcceptConnection(SharkSocket listenSocket); // warning - this can block
	static SharkSocket AcceptConnection(SharkSocket listenSocket, const FD_SET_TYPE * readerSet);
	static uint8b WaitForSocketEvent(SharkSocket s, bool read, bool write, bool exception, TimeSpan_t timeout);
	static void CloseSocket(SharkSocket *s, bool tcp);

#if defined(__FREERTOS__)
	static void SetMac(uint64b mac);
	static void PhyAliveCheck(WatchdogManager *wdtManager, uint16b wdtClientId, uint32b wdtPeriod);
#endif

	static void GetCurrentConfiguration(uint32b *ip, uint32b *gw, uint32b *dns, uint32b *mask);
};
#endif
/*!
 * \brief Communication buffer
 * \details Class for working about buffer.
 */
struct CommBuffer
{
    /*!
     * \brief Buffer
     * \details Buffer data array
     */
	uint8b * Buffer;
    /*!
     * \brief Buffer size
     * \details Maximum size of buffered data
     */
	uint32b BufferSize;
    /*!
     * \brief Offset
     * \details Current data offset for pointer
     */
	sint32b Offset;
    /*!
     * \brief Data length
     * \details Current data length
     */
	uint32b DataLen;

    /*!
     * \brief Data offset
     * \details Data offset for data writing
     */
    uint32b DataOffset;

    /*!
     * \brief Constructor
     * \details Initialize all properties.
     */
	CommBuffer()
	{
		Buffer = NULL;
		BufferSize = 0;
		Offset = 0;
		DataLen = 0;
		DataOffset = 0;
	}

    /*!
     * \brief Destructor
     * \details Delete buffer and free memmory.
     */
    ~CommBuffer()
	{
		if (Buffer != NULL)
		{
			delete [] Buffer;
			Buffer = NULL;
		}
	}

    /*!
     * \brief Allocate buffer
     * \param[in] bufferSize New buffer size
     * \details Create new buffer memory and allocate size for them
     */
	void AllocateBuffer(uint32b bufferSize)
	{
		Buffer = new uint8b[BufferSize = bufferSize];
	}

    /*!
    * \brief Clear
    * \details Clears buffer from data. This function doesn't release memmory. only resets all pointers
    */
	void Clear()
	{
		Offset = 0;
		DataLen = 0;
		DataOffset = 0;
	}

    /*!
     * \brief GetRemain
     * \return Remaining buffer capacity
     * \details Get remaining buffer capacity
     */
	uint32b GetRemain(void) { return (DataLen - DataOffset); }
     /*!
     * \brief GetFree
     * \return Free buffer size
     * \details Get free buffer size
     */
	uint32b GetFree(void) { return (BufferSize - DataOffset); }
    /*!
     * \brief operator []
     * \param[in] index Data index
     * \return Value on selected index
     * \details Get data from buffer on specific index
     */
	uint8b& operator[](int index)
	{
		return(Buffer[index]);
	}

	/*const uint8b& operator[](int index) const
	{
		return(Buffer[index]);
	}*/

    /*!
     * \brief operator +=
     * \param[in] right Value
     * \return Whole buffer
     * \details This operation adds specific data at the end of buffer
     */
	CommBuffer& operator+=(const uint8b& right)
	{
		Buffer[DataOffset++] = right;
		return (*this);
	}

    /*!
     * \brief operator <<=
     * \param[in] right Value
     * \return Whole buffer
     * \details This operation adds specific data at the end of buffer
     */
	CommBuffer& operator<<=(const uint8b& right)
	{
		Buffer[DataOffset++] = right;
		return (*this);
	}

    /*!
     * \brief operator <<=
     * \param[in] right Value
     * \return Whole buffer
     * \details This operation adds specific data at the end of buffer (uint16b)
     */
    CommBuffer& operator<<=(const uint16b& right)
	{
		DataOffset = SetUint16ToBuffer(Buffer, DataOffset++, right);
		return (*this);
	}

    /*!
     * \brief operator <<=
     * \param[in] right Value
     * \return Whole buffer
     * \details This operation adds specific data at the end of buffer (uint32b)
     */
    CommBuffer& operator<<=(const uint32b& right)
	{
		DataOffset = SetUint32ToBuffer(Buffer, DataOffset++, right);
		return (*this);
	}

    /*!
     * \brief operator <<=
     * \param[in] right Value
     * \return Whole buffer
     * \details This operation adds specific data at the end of buffer (uint64b)
     */
    CommBuffer& operator<<=(const uint64b& right)
	{
		DataOffset = SetUint64ToBuffer(Buffer, DataOffset++, right);
		return (*this);
	}

    /*!
     * \brief operator >>
     * \param[out] right
     * \return Whole buffer
     * \details Read data from current position
     */
	CommBuffer& operator>>(uint8b& right)
	{
		right = Buffer[DataOffset++];
		return (*this);
	}

    /*!
     * \brief operator >>
     * \param[out] right
     * \return Whole buffer
     * \details Read data from current position (uint16b)
     */
    CommBuffer& operator>>(uint16b& right)
	{
		right = GetUint16FromBuffer(Buffer, DataOffset);
		DataOffset += 2;
		return (*this);
	}

    /*!
     * \brief operator >>
     * \param[out] right
     * \return Whole buffer
     * \details Read data from current position (uint32b)
     */
    CommBuffer& operator>>(uint32b& right)
	{
		right = GetUint32FromBuffer(Buffer, DataOffset);
		DataOffset += 4;
		return (*this);
	}

    /*!
     * \brief operator >>
     * \param[out] right
     * \return Whole buffer
     * \details Read data from current position (uint64b)
     */
    CommBuffer& operator>>(uint64b& right)
	{
		right = GetUint64FromBuffer(Buffer, DataOffset);
		DataOffset += 8;
		return (*this);
	}

	#pragma region Funkce pro vytahovani dat z bufferu
    /*!
     * \brief Get uint16 value from buffer
     * \param[in] buffer Data buffer
     * \param[in] offset Data position
     * \return Value
     * \details Reads value from \c buffer on from position \c offset and return readed value
     */
	static uint16b GetUint16FromBuffer(uint8b * buffer, sint32b offset)
	{
		return (buffer[offset] << 8) + buffer[offset + 1];
	}

    /*!
     * \brief Get uint32 value from buffer
     * \param[in] buffer Data buffer
     * \param[in] offset Data position
     * \return Value
     * \details Reads value from \c buffer on from position \c offset and return readed value
     */
    static uint32b GetUint32FromBuffer(uint8b * buffer, sint32b offset)
	{
		return (buffer[offset] << 24) + (buffer[offset + 1] << 16) + (buffer[offset + 2] << 8) + buffer[offset + 3];
	}

    /*!
     * \brief Get uint64 value from buffer
     * \param[in] buffer Data buffer
     * \param[in] offset Data position
     * \return Value
     * \details Reads value from \c buffer on from position \c offset and return readed value
     */
    static uint64b GetUint64FromBuffer(uint8b * buffer, sint32b offset)
	{
		uint32b upper = GetUint32FromBuffer(buffer, offset);
		uint32b lower = GetUint32FromBuffer(buffer + 4, offset);

		return ((uint64b)upper << 32) + lower;
	}

    /*!
     * \brief Get int64 value from buffer
     * \param[in] buffer Data buffer
     * \param[in] offset Data position
     * \return Value
     * \details Reads value from \c buffer on from position \c offset and return readed value
     */
    static sint64b GetInt64FromBuffer(uint8b * buffer, sint32b offset)
	{
		uint32b upper = GetUint32FromBuffer(buffer, offset);
		uint32b lower = GetUint32FromBuffer(buffer + 4, offset);

		return ((sint64b)upper << (sint64b)32) + lower;
	}

    /*!
     * \brief Set uint32 value to buffer
     * \param buffer Data buffer
     * \param[in] offset Data position
     * \param[in] value Value
     * \return Writen data size
     * \details Write \c value to \c buffer from position \c offset and return writen data size
     */
    static sint32b SetUint16ToBuffer(uint8b * buffer, sint32b offset, uint16b value)
	{
		buffer[offset] = value >> 8;
		buffer[offset + 1] = value & 0xFF;
		return(offset + 2);
	}

    /*!
     * \brief Set uint32 value to buffer
     * \param buffer Data buffer
     * \param[in] offset Data position
     * \param[in] value Value
     * \return Writen data size
     * \details Write \c value to \c buffer from position \c offset and return writen data size
     */
    static sint32b SetUint32ToBuffer(uint8b * buffer, sint32b offset, uint32b value)
	{
		buffer[offset] = value >> 24;
		buffer[offset + 1] = (value >> 16) & 0xFF;
		buffer[offset + 2] = (value >> 8) & 0xFF;
		buffer[offset + 3] = value & 0xFF;
		return(offset + 4);
	}

    /*!
     * \brief Set uint32 value to buffer
     * \param buffer Data buffer
     * \param[in] offset Data position
     * \param[in] value Value
     * \return Writen data size
     * \details Write \c value to \c buffer from position \c offset and return writen data size
     */
    static sint32b SetUint64ToBuffer(uint8b * buffer, sint32b offset, uint64b value)
	{
		offset = SetUint32ToBuffer(buffer, offset, (uint32b)(value >> 32));
		offset = SetUint32ToBuffer(buffer, offset, (uint32b)(value & 0xFFFFFFFF));
		return(offset);
	}
	#pragma endregion
};


#endif

