#include "./SscpClient.h"
#include "./ModbusCrc.h"
#include "./SystemLogger.h"

#if defined(__BORLANDC__) || defined(__POSIX__)
#include <stdio.h>
#endif

#if (defined(__POSIX__) && !defined(__PIKEOS__)) || (defined(WINAPI_FAMILY) && WINAPI_FAMILY==WINAPI_FAMILY_PHONE_APP)
#include <openssl/md5.h>
#endif

SscpDevice::SscpDevice(uint32b ipv4, uint16b port, uint8b slaveAddress, char *username, char *password,
	char *proxyId, bool md5PassFormat)
{
	Ipv4 = ipv4;
	Port = port;
	SlaveAddress = slaveAddress;
	Username = username;

	if (!md5PassFormat)
	{
		CalculateMD5(Password, (uint8b*)password);
	}
	else
	{
		memcpy(Password, password, SharkMin(16, strlen(password)));
	}

	ProxyId = proxyId;

	MaxDataSize = 0;
}

SscpDevice::SscpDevice(char *ipv4, uint16b port, uint8b slaveAddress, char *username, char *password,
	char *proxyId, bool md5PassFormat)
{
	Ipv4 = inet_addr(ipv4);
	Port = port;
	SlaveAddress = slaveAddress;
	Username = username;

	if (!md5PassFormat)
	{
		CalculateMD5(Password, (uint8b*)password);
	}
	else
	{
		memcpy(Password, password, SharkMin(16, strlen(password)));
	}

	ProxyId = proxyId;

	MaxDataSize = 0;
}

void SscpDevice::CalculateMD5(unsigned char *destination, unsigned char *password)
{
#if (defined(_MSC_VER) || defined(__BORLANDC__)) && (!defined(WINAPI_FAMILY) || WINAPI_FAMILY!=WINAPI_FAMILY_PHONE_APP)
	HCRYPTPROV hCryptProv;
	HCRYPTHASH hHash;

	BYTE bHash[0x7f];
	DWORD dwHashLen = 16; // The MD5 algorithm always returns 16 bytes.

	if (CryptAcquireContext(&hCryptProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_MACHINE_KEYSET))
	{
		if (CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hHash))
		{
			if (CryptHashData(hHash, password, strlen((char *)password), 0))
			{
				if (CryptGetHashParam(hHash, HP_HASHVAL, bHash, &dwHashLen, 0))
					memcpy(destination, bHash, 16);
			}
		}
	}
#endif
#if (defined(__POSIX__) && !defined(__PIKEOS__)) || (defined(WINAPI_FAMILY) && WINAPI_FAMILY==WINAPI_FAMILY_PHONE_APP)
	MD5(password, strlen((const char*)password), destination);
#endif
}

bool SscpClient::CommonCheckAndPart(SscpDevice *device)
{
	if ((!m_Parser) || (!device) || (device->EndPoint < 0))
		return (false);

	m_EPContext[device->EndPoint]->Buffer.DataOffset = 1;
	m_EPContext[device->EndPoint]->Buffer[0] = device->SlaveAddress;
	return (true);
}

bool SscpClient::SendAndReceivedResponse(SscpDevice *device, bool response)
{
	if (m_EPContext[device->EndPoint]->State == DriverState_Idle)
	{
		m_EPContext[device->EndPoint]->CompleteEvent.Wait(1);
		m_EPContext[device->EndPoint]->State = DriverState_SendRequest;
		if (response)
		{
			bool eventSignaled = m_EPContext[device->EndPoint]->CompleteEvent.Wait(m_Timeout);
			bool dataReceived = m_EPContext[device->EndPoint]->Buffer.DataLen > 0;
			bool correctState = m_EPContext[device->EndPoint]->State != DriverState_Reset;
			if (!correctState)
				m_EPContext[device->EndPoint]->State = DriverState_Idle;
			return (eventSignaled && dataReceived && correctState);
		}
		return (true);
	}
	else
	{
		if (m_EPContext[device->EndPoint]->State == DriverState_Reset)
			m_EPContext[device->EndPoint]->State = DriverState_Idle;
	}
	return (false);
}

bool SscpClient::AssignEndPoint(uint8b epIndex, uint32b ipv4, uint16b port, uint32b timeout)
{
	if (m_TcpClient)
	{
		TcpClientEndpoint *ep = new TcpClientEndpoint(ipv4, timeout);
		if (ep)
		{
			ep->SetParameters(port, this, epIndex);
			m_TcpClient->SetEndpoint(epIndex, ep);
			return (true);
		}
	}
	return (false);
}

SscpClient::SscpClient(uint8b maxEndpoints, uint32b timeout, uint32b maxBufferSize)
{
	m_Timeout = SharkMax(timeout, 1000);
	m_MaxEndpoints = maxEndpoints;
	m_AssignedEpCount = 0;

	m_TcpClient = new TcpUdpClient(m_MaxEndpoints);
	m_Parser = new SscpProtocolParser();
	m_EPContext = new SscpEndpointContext*[m_MaxEndpoints];
	for (uint32b i = 0; i < m_MaxEndpoints; i++)
	{
		m_EPContext[i] = new SscpEndpointContext(maxBufferSize + 40);
		m_EPContext[i]->State = DriverState_Idle;
		m_EPContext[i]->ErrorFlags = 0;
	}
}

SscpClient::~SscpClient()
{
	if (m_EPContext)
	{
		for (uint32b i = 0; i < m_MaxEndpoints; i++)
		{
			if (m_EPContext[i])
			{
				delete (m_EPContext[i]);
				m_EPContext[i] = NULL;
			}
		}
		delete[] m_EPContext;
		m_EPContext = NULL;
	}
	if (m_TcpClient)
	{
		delete (m_TcpClient);
		m_TcpClient = NULL;
	}
	if (m_Parser)
	{
		delete (m_Parser);
		m_Parser = NULL;
	}
}

bool SscpClient::AssignDevice(SscpDevice *device, bool forceCreateNewEndpoint /*= false (from declaration)*/)
{
	if (m_TcpClient)
	{
		if (forceCreateNewEndpoint)
		{
			sint32b epIndex = m_AssignedEpCount;
			device->EndPoint = AssignEndPoint(m_AssignedEpCount++, device->Ipv4, device->Port, m_Timeout - 100) ? epIndex : -1;

			return(device->EndPoint != -1);
		}
		else
		{
			sint32b epIndex = m_TcpClient->FindEndpoint(device->Ipv4, device->Port);
			if (epIndex == -1)
			{
				// pridame pouze pokud EP neexistuje
				epIndex = m_AssignedEpCount;
				device->EndPoint = AssignEndPoint(m_AssignedEpCount++, device->Ipv4, device->Port, m_Timeout - 100) ? epIndex : -1;

				return(device->EndPoint != -1);
			}
			else
			{
				device->EndPoint = epIndex;
				return (true);
			}
		}
	}
	return (false);
}

bool SscpClient::Initialize(void)
{
	if (m_TcpClient)
	{
		m_TcpClient->Start(1);
		return (true);
	}
	return (false);
}

bool SscpClient::DeInitialize(void)
{
	if (m_TcpClient)
	{
		m_TcpClient->SetStopSignal();
		m_TcpClient->Join(0xff, 0, 0);
		m_TcpClient->ClearAndFreeEndpoint();
		m_AssignedEpCount = 0;
		return (true);
	}
	return (false);
}

void SscpClient::Reset(uint32b tag)
{
	if (m_EPContext[tag]->State != DriverState_Idle)
	{
		m_EPContext[tag]->State = DriverState_Reset;
		m_EPContext[tag]->CompleteEvent.Signal();
	}
}

ICallbackCommunicationDriver::CallbackAction SscpClient::GetNextAction(uint32b tag)
{
	ICallbackCommunicationDriver::CallbackAction action = ICallbackCommunicationDriver::CallbackAction_None;

	DriverState state = m_EPContext[tag]->State;
	switch (state)
	{
		case DriverState_SendRequest:
			{
				action = ICallbackCommunicationDriver::CallbackAction_WriteData;
				m_EPContext[tag]->State = DriverState_ReadConfirmation;
			} break;
		case DriverState_ReadConfirmation:
			{
				const uint16b headerLen = 5;
				const uint16b lengthOffset = 3;

				if (m_EPContext[tag]->Buffer.DataLen >= headerLen)
				{
					uint16b expectedLength = m_EPContext[tag]->Buffer[lengthOffset] << 8 | m_EPContext[tag]->Buffer[lengthOffset + 1] + headerLen;

					if (expectedLength == m_EPContext[tag]->Buffer.DataLen)
					{
						m_EPContext[tag]->State = DriverState_CheckConfirmation;
					}
					else
					{
						action = ICallbackCommunicationDriver::CallbackAction_ReadData;
					}
				}
				else
					action = ICallbackCommunicationDriver::CallbackAction_ReadData;
			} break;
		case DriverState_CheckConfirmation:
			{
				if (m_EPContext[tag]->Buffer.DataLen > 0)
				{
					m_EPContext[tag]->Buffer.DataOffset = 1;
					m_EPContext[tag]->State = DriverState_Idle;
					m_EPContext[tag]->CompleteEvent.Signal();
				}
			} break;
		case DriverState_Idle:
			{
				// bud pozadavek pripojeni nebo prijata asynchronne data
			} break;
		case DriverState_Reset:
			{
				// stav pokud je nejaka chyba pri spojeni - zmena z tohoto stavu je v sendandreceive
			} break;
	}
	return(action);
}

CommBuffer* SscpClient::GetCommunicationBuffer(uint32b tag)
{
	return (&(m_EPContext[tag]->Buffer));
}

SscpError::SscpError SscpClient::Login(SscpDevice *device)
{
	if (CommonCheckAndPart(device))
	{
		m_Parser->Login_Request(m_EPContext[device->EndPoint]->Buffer, device->Username,
			(const char*)device->Password, device->ProxyId);

		if (SendAndReceivedResponse(device, true))
		{
			SscpError::SscpError res = m_Parser->Login_Confirmation(m_EPContext[device->EndPoint]->Buffer,
				&(device->ProtocolVersion),
				&(device->ServerDataSize),
				&(device->Rights),
				device->ImageGuid);

			device->MaxDataSize = SharkMin(device->ServerDataSize, m_EPContext[device->EndPoint]->Buffer.BufferSize - 40);

			return res;
		}

		return (SscpError::NoResponse);
	}
	return (SscpError::Unknown);
}

SscpError::SscpError SscpClient::Logout(SscpDevice *device)
{
	return (SscpError::NoError);
}

SscpError::SscpError SscpClient::ReadSingleVariable(SscpDevice *device, uint32b commUid, uint8b *buffer, uint32b offset, uint32b length)
{
	SscpError::SscpError result = SscpError::Unknown;
	if (CommonCheckAndPart(device))
	{
		m_Parser->ReadVariablesDirectlyHeader_Request(m_EPContext[device->EndPoint]->Buffer, length == 0, false, 1);
		if (length == 0)
			m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid);
		else
			m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid, offset, length);

		if (SendAndReceivedResponse(device, true))
		{
			uint8b *value;
			uint32b valueLength = 0;

			result = m_Parser->ReadVariablesDirectlyVars_Confirmation(m_EPContext[device->EndPoint]->Buffer, &value, &valueLength,
				&(m_EPContext[device->EndPoint]->ErrorFlags));
			if (result == SscpError::NoError)
				memcpy(buffer, value, valueLength);
		}
		else
			result = SscpError::NoResponse;
	}
	if (result == SscpError::TooLongUseFileTransfer)
	{
		char tmpPath[32];
		uint32b valuesLength = 0;
		strcpy(tmpPath, "/var/direct");
		result = ReadFile(device, tmpPath, buffer, &valuesLength);

		return (result);
	}
	return (result);
}

SscpError::SscpError SscpClient::ReadMultipleVariables(SscpDevice *device, uint32b *commUid, uint8b *buffer, uint32b *offset, uint32b *length, uint32b count)
{
	SscpError::SscpError result = SscpError::Unknown;
	if ((CommonCheckAndPart(device)) && (device->MaxDataSize > 0))
	{
		m_Parser->ReadVariablesDirectlyHeader_Request(m_EPContext[device->EndPoint]->Buffer, length == NULL, false, count);
		uint32b i;
		for (i = 0; (i < count) && (m_EPContext[device->EndPoint]->Buffer.DataOffset < device->MaxDataSize); i++)
		{
			if (length == NULL)
				m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid[i]);
			else
				m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid[i], offset[i], length[i]);
		}
		if (i < count)
			return (SscpError::DataTooLong);

		if (SendAndReceivedResponse(device, true))
		{
			uint8b *value;
			uint32b valueLength = 0;

			result = m_Parser->ReadVariablesDirectlyVars_Confirmation(m_EPContext[device->EndPoint]->Buffer, &value, &valueLength,
				&(m_EPContext[device->EndPoint]->ErrorFlags));
			if (result == SscpError::NoError)
				memcpy(buffer, value, valueLength);
		}
		else
			result = SscpError::NoResponse;
	}
	else
		result = SscpError::DeviceNotInitialized;

	if (result == SscpError::TooLongUseFileTransfer)
	{
		char tmpPath[32];
		uint32b valuesLength = 0;
		strcpy(tmpPath, "/var/direct");
		result = ReadFile(device, tmpPath, buffer, &valuesLength);

		return (result);
	}
	return (result);
}

SscpError::SscpError SscpClient::RegisterVariableGroup(SscpDevice *device, uint8b groupId, VariableDefinition *variables, uint32b count)
{
	if (CommonCheckAndPart(device))
	{
		uint32b len = 0;
		for (uint32b i = 0; i < count; i++)
			len += variables[i].Length;

		m_Parser->RegisterGroupHeader_Request(m_EPContext[device->EndPoint]->Buffer, groupId, len == 0, false, count);
		uint32b i;
		for (i = 0; (i < count) && (m_EPContext[device->EndPoint]->Buffer.DataOffset < device->MaxDataSize); i++)
		{
			if (len == 0)
				m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, variables[i].Uid);
			else
				m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, variables[i].Uid, variables[i].Offset, variables[i].Length);
		}
		if (i < count)
			return (SscpError::DataTooLong);

		if (SendAndReceivedResponse(device, true))
			return (SscpError::NoError);
		return (SscpError::NoResponse);
	}
	return (SscpError::Unknown);
}

SscpError::SscpError SscpClient::ReadVariableGroup(SscpDevice *device, uint8b groupId, VarAccessFlags::VarAccessFlags flags, uint8b *buffer,
	bool *forceFileTransfer, uint32b *length)
{
	if (length)
		*length = 0;

	SscpError::SscpError result = SscpError::Unknown;
	if (CommonCheckAndPart(device) && ((!forceFileTransfer) || (*forceFileTransfer == false)))
	{
		m_Parser->ReadVariableGroupDirectly_Request(m_EPContext[device->EndPoint]->Buffer, groupId, flags);

		if (SendAndReceivedResponse(device, true))
		{
			uint8b *value;
			uint32b valueLength = 0;

			result = m_Parser->ReadVariableGroupDirectly_Confirmation(m_EPContext[device->EndPoint]->Buffer, &value, &valueLength,
				&(m_EPContext[device->EndPoint]->ErrorFlags));
			if (result == SscpError::NoError)
			{
				memcpy(buffer, value, valueLength);
				if (length)
					*length = valueLength;
			}
		}
		else
			return (SscpError::NoResponse);
	}
	if ((result == SscpError::DataTooLong) || ((forceFileTransfer) && (*forceFileTransfer)))
	{
		if (forceFileTransfer)
			*forceFileTransfer = true;

		char tmpPath[32];
		uint32b valuesLength = 0;
		sprintf(tmpPath, "/var/group/%d/%d", groupId, flags & VarAccessFlags::GetFormatMask);
		result = ReadFile(device, tmpPath, buffer, &valuesLength);
		if (length)
			*length = valuesLength;
	}
	return (result);
}

SscpError::SscpError SscpClient::WriteSingleVariable(SscpDevice *device, uint32b commUid, uint32b offset, uint32b length, bool commUidOnly, uint8b *buffer)
{
	SscpError::SscpError result = SscpError::Unknown;
	if (CommonCheckAndPart(device))
	{
		bool bFileMode = length > device->MaxDataSize;

		if (bFileMode)
		{
			result = WriteFile(device, "/var/direct", buffer, length, 0);
			if (result != SscpError::NoError)
				return (result);
		}

		m_Parser->WriteVariableDirectly_Request(m_EPContext[device->EndPoint]->Buffer, commUidOnly, false, (VarAccessFlags::VarAccessFlags)0,
			1, bFileMode ? 0 : length);

		if (commUidOnly)
			m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid);
		else
			m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid, offset, length);

		if (!bFileMode)
			memcpy(m_EPContext[device->EndPoint]->Buffer.Buffer + m_EPContext[device->EndPoint]->Buffer.DataOffset, buffer, length);

		if (SendAndReceivedResponse(device, true))
			result = m_Parser->WriteVariableDirectly_Confirmation(m_EPContext[device->EndPoint]->Buffer,
			&(m_EPContext[device->EndPoint]->ErrorFlags));
		else
			result = SscpError::NoResponse;
	}
	return (result);
}
// JK: Funkce opravena, aby se odesílala nejdříve velká data a poté požadavek na zapsání
SscpError::SscpError SscpClient::WriteMultipleVariables(SscpDevice *device, uint32b *commUid, uint32b *offset, uint32b *length, uint32b count,
	bool commUidOnly, uint8b *bufferWithAllValues)
{
	SscpError::SscpError result = SscpError::Unknown;
	if (CommonCheckAndPart(device) && commUid && length)
	{
		uint32b totalValueLength = 0, i = 0, totalRequestLength = 0;
		for (i = 0; i < count; i++)
			totalValueLength += length[i];

		m_Parser->WriteVariableDirectly_Request(m_EPContext[device->EndPoint]->Buffer, commUidOnly, false, (VarAccessFlags::VarAccessFlags)0,
			count, totalValueLength, &totalRequestLength);

		bool bFileMode = totalRequestLength > device->MaxDataSize;
		if (bFileMode)  // JK:Zapsání dat do souboru
		{
#if !defined(SSCP_STANDALONE_CLIENT)
			Logger.LogDebug(NULL, "File transfer");
#endif
			result = WriteFile(device, "/var/direct", bufferWithAllValues, totalValueLength, 0);
			if (result != SscpError::NoError)
				return (result);
			m_Parser->WriteVariableDirectly_Request(m_EPContext[device->EndPoint]->Buffer, commUidOnly, false, (VarAccessFlags::VarAccessFlags)0,
				count, 0, &totalRequestLength);
		}

		// JK: Doplněno ((bFileMode) ? 0 : totalValueLength)), protože velikost není ovlivněna délkou dat, pokud se zapisuje přes soubor
		for (i = 0; (i < count) && ((m_EPContext[device->EndPoint]->Buffer.DataOffset + ((bFileMode) ? 0 : totalValueLength)) < device->MaxDataSize); i++)
		{
			if (commUidOnly)
				m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid[i]);
			else
				m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid[i], offset[i], length[i]);
		}
		if (i < count)
			return (SscpError::DataTooLong);

		if (!bFileMode)
			memcpy(m_EPContext[device->EndPoint]->Buffer.Buffer + m_EPContext[device->EndPoint]->Buffer.DataOffset, bufferWithAllValues, totalValueLength);

		if (SendAndReceivedResponse(device, true))
			result = m_Parser->WriteVariableDirectly_Confirmation(m_EPContext[device->EndPoint]->Buffer,
			&(m_EPContext[device->EndPoint]->ErrorFlags));
		else
			result = SscpError::NoResponse;
	}
	return (result);
}

SscpError::SscpError SscpClient::WriteVariableGroup(SscpDevice *device, uint8b groupId, VarAccessFlags::VarAccessFlags flags, uint8b *indexes, uint8b idxCount,
	uint8b *buffer, uint32b length)
{
	SscpError::SscpError result = SscpError::Unknown;
	if (CommonCheckAndPart(device))
	{
		uint8b writeMode = flags & VarAccessFlags::SetFormatMask;

		bool bFileMode = length > device->MaxDataSize;

		if (bFileMode)
		{
			result = WriteFile(device, "/var/direct", buffer, length, 0);
			if (result != SscpError::NoError)
				return (result);
		}

		m_Parser->WriteVariableGroupDirectly_Request(m_EPContext[device->EndPoint]->Buffer, groupId, flags, idxCount, bFileMode ? 0 : length);
		if (writeMode == VarAccessFlags::SetSelectedValues)
		{
			memcpy(m_EPContext[device->EndPoint]->Buffer.Buffer + m_EPContext[device->EndPoint]->Buffer.DataOffset, indexes, idxCount);
			m_EPContext[device->EndPoint]->Buffer.DataOffset += idxCount;
		}

		if (!bFileMode)
			memcpy(m_EPContext[device->EndPoint]->Buffer.Buffer + m_EPContext[device->EndPoint]->Buffer.DataOffset, buffer, length);

		if (SendAndReceivedResponse(device, true))
			result = m_Parser->WriteVariableGroupDirectly_Confirmation(m_EPContext[device->EndPoint]->Buffer,
			&(m_EPContext[device->EndPoint]->ErrorFlags));
		else
			result = SscpError::NoResponse;
	}
	return (result);
}

SscpError::SscpError SscpClient::WriteVariableGroup(SscpDevice *device, uint8b groupId, VarAccessFlags::VarAccessFlags flags, uint8b *buffer, uint32b length)
{
	return(WriteVariableGroup(device, groupId, flags, NULL, 0, buffer, length));
}

SscpError::SscpError SscpClient::ReadVariableStructure(SscpDevice *device, uint32b commUid, uint8b *buffer, uint32b *length)
{
	SscpError::SscpError result = SscpError::Unknown;
	if (CommonCheckAndPart(device))
	{
		m_Parser->ReadVariableStructure_Request(m_EPContext[device->EndPoint]->Buffer, 0, 1);
		m_Parser->AppendVariable_Request(m_EPContext[device->EndPoint]->Buffer, commUid);

		if (SendAndReceivedResponse(device, true))
		{
			uint8b *structure;
			uint32b structLength = 0;

			result = m_Parser->ReadVariableStructure_Confirmation(m_EPContext[device->EndPoint]->Buffer, &structure, &structLength);
			if (result == SscpError::NoError)
			{
				memcpy(buffer, structure, structLength);
				if (length)
					*length = structLength;
			}
		}
	}
	return (result);
}

SscpError::SscpError SscpClient::ReadFile(SscpDevice *device, const char *filename, uint8b* buffer, uint32b *fileSize)
{
	if (CommonCheckAndPart(device))
	{
		m_Parser->InitiateReceiveData_Request(m_EPContext[device->EndPoint]->Buffer, filename);
		if (SendAndReceivedResponse(device, true))
		{
			DataTransferProperties dataProperties;

			SscpError::SscpError result = m_Parser->InitiateReceiveData_Confirmation(m_EPContext[device->EndPoint]->Buffer, &dataProperties);
			if (result == SscpError::NoError)
			{
				if (fileSize)
					*fileSize = dataProperties.Size;

				uint32b dataOffset = 0;
				bool transmissionCompleted = false;

				result = SscpError::NoError;
				if (buffer)
				{
					result = SscpError::Unknown;
					do
					{
						CommonCheckAndPart(device);
						m_Parser->ReceiveDataChunk_Request(m_EPContext[device->EndPoint]->Buffer, dataOffset);

						if (SendAndReceivedResponse(device, true))
						{
							result = m_Parser->ReceiveDataChunk_Confirmation(m_EPContext[device->EndPoint]->Buffer, &dataOffset, buffer,
								&transmissionCompleted, &dataProperties);
							if (result != SscpError::NoError)
							{
								transmissionCompleted = true;
								break;
							}
						}
						else
							break;

					}
					while (!transmissionCompleted);

#if 0
					if (dataProperties.CalculatedCrc != dataProperties.Crc)
						return (SscpError::InvalidCrc);
#endif
				}
			}
			return (result);
		}
	}
	return (SscpError::Unknown);
}

SscpError::SscpError SscpClient::WriteFile(SscpDevice *device, const char *filename, uint8b* buffer, uint32b dataSize, uint64b timeStamp)
{
	if (CommonCheckAndPart(device))
	{
		uint16b crc = ModbusCrc::CRC16(buffer, dataSize);

		DataTransferProperties dataProperties;
		dataProperties.Fill(dataSize, timeStamp, crc);

		m_Parser->InitiateSendData_Request(m_EPContext[device->EndPoint]->Buffer, filename, &dataProperties);
		if (SendAndReceivedResponse(device, true))
		{
			if (m_Parser->InitiateSendData_Confirmation(m_EPContext[device->EndPoint]->Buffer) == SscpError::NoError)
			{
				uint32b dataOffset = 0;
				bool transmissionCompleted = false;

				SscpError::SscpError result = SscpError::Unknown;
				do
				{
					CommonCheckAndPart(device);
					m_Parser->SendDataChunk_Request(m_EPContext[device->EndPoint]->Buffer, device->MaxDataSize, buffer, dataSize, &dataOffset, &transmissionCompleted);

					if (SendAndReceivedResponse(device, true))
					{
						uint32b confirmedDataOffset = 0;
						result = m_Parser->SendDataChunk_Confirmation(m_EPContext[device->EndPoint]->Buffer, &confirmedDataOffset);
						if (result != SscpError::NoError)
						{
							transmissionCompleted = true;
							break;
						}
					}
					else
						break;

				}
				while (!transmissionCompleted);

				CommonCheckAndPart(device);
				m_Parser->FinishDataSend_Request(m_EPContext[device->EndPoint]->Buffer, crc);
				if (SendAndReceivedResponse(device, true))
				{
					return (m_Parser->FinishDataSend_Confirmation(m_EPContext[device->EndPoint]->Buffer));
				}
			}
		}
	}
	return (SscpError::NoError);
}

SscpError::SscpError SscpClient::TimeSync(SscpDevice *device, uint64b time, uint64b *returnedTime, TimeCommand::TimeCommand *cmd)
{
	SscpError::SscpError result = SscpError::Unknown;
	if (CommonCheckAndPart(device))
	{
		if (device->ProtocolVersion < 7)
		{
			if (cmd != NULL) *cmd = (time > 0) ? TimeCommand::SetUTC : TimeCommand::GetUTC;
			m_Parser->TimeSetup_Request(m_EPContext[device->EndPoint]->Buffer, time);
		}
		else
		{
			TimeCommand::TimeCommand c = (cmd == NULL) ? TimeCommand::GetUTC : *cmd;
			m_Parser->TimeSetupEx_Request(m_EPContext[device->EndPoint]->Buffer, static_cast<uint8b>(c), 0, time);
		}
		if (SendAndReceivedResponse(device, true))
		{
			result = (device->ProtocolVersion < 7) ?
				m_Parser->TimeSetup_Confirmation(m_EPContext[device->EndPoint]->Buffer, returnedTime) :
				m_Parser->TimeSetupEx_Confirmation(m_EPContext[device->EndPoint]->Buffer, returnedTime);
		}
	}
	return (result);
}

uint64b SscpClient::GetErrorFlagsForDestination(SscpDevice *device)
{
	if ((device) && (device->EndPoint < m_MaxEndpoints))
		return (m_EPContext[device->EndPoint]->ErrorFlags);
	return (0);
}

#if defined(SSCP_CLIENT_EXTENDED_CONVERSION)

template <class T> bool ConvertSscpValue(void *ptr, bool boolValue, StTypes::StType returnType, void *destination)
{
	T value = *(T*)ptr;
	if (boolValue)
		value = value != 0;

	switch (returnType)
	{
		case StTypes::Bool:		*(bool*)destination = (value != 0) ? 1 : 0; break;
		case StTypes::Byte:
		case StTypes::Usint:	*(uint8b*)destination = (uint8b)value; break;
		case StTypes::Sint:		*(sint8b*)destination = (sint8b)value; break;
		case StTypes::Word:
		case StTypes::Uint:		*(uint16b*)destination = (uint16b)value; break;
		case StTypes::Int:		*(sint16b*)destination = (sint16b)value; break;
		case StTypes::Dword:
		case StTypes::Udint:	*(uint32b*)destination = (uint32b)value; break;
		case StTypes::Dint:		*(sint32b*)destination = (sint32b)value; break;
		case StTypes::Ulint:	*(uint64b*)destination = (uint64b)value; break;
		case StTypes::Date:
		case StTypes::Time:
		case StTypes::TOD:
		case StTypes::DT:
		case StTypes::Lword:
		case StTypes::Lint:		*(sint64b*)destination = (sint64b)value; break;
		case StTypes::Real:		*(float*)destination = (float)value; break;
		case StTypes::Lreal:	*(double*)destination = (double)value; break;
		default:
			return (false);
	}
	return (true);
}

bool SscpClient::ConvertVariableFromPlc(uint8b *buffer, StTypes::StType type, bool bigEndian, void *destination,
	StTypes::StType returnType)
{
	uint8b tmp[8];
	uint8b *ptr = buffer;
	if (!bigEndian)
	{
		switch (type)
		{
			case StTypes::Bool:
			case StTypes::Usint:
			case StTypes::Sint:
			case StTypes::Byte:
				{
					tmp[0] = *ptr;
				} break;
			case StTypes::Int:
			case StTypes::Uint:
			case StTypes::Word:
				{
					tmp[0] = ptr[1];
					tmp[1] = ptr[0];
				} break;
			case StTypes::Dint:
			case StTypes::Udint:
			case StTypes::Real:
			case StTypes::Dword:
				{
					tmp[0] = ptr[3];
					tmp[1] = ptr[2];
					tmp[2] = ptr[1];
					tmp[3] = ptr[0];
				} break;
			case StTypes::Lint:
			case StTypes::Ulint:
			case StTypes::Lreal:
			case StTypes::Time:
			case StTypes::DT:
			case StTypes::Date:
			case StTypes::Lword:
				{
					tmp[0] = ptr[7];
					tmp[1] = ptr[6];
					tmp[2] = ptr[5];
					tmp[3] = ptr[4];
					tmp[4] = ptr[3];
					tmp[5] = ptr[2];
					tmp[6] = ptr[1];
					tmp[7] = ptr[0];
				} break;
		}
		ptr = tmp;
	}

	switch (type)
	{
		case StTypes::Sint:		return ConvertSscpValue<sint8b>(ptr, false, returnType, destination); break;
		case StTypes::Int:		return ConvertSscpValue<sint16b>(ptr, false, returnType, destination); break;
		case StTypes::Dint:		return ConvertSscpValue<sint32b>(ptr, false, returnType, destination); break;
		case StTypes::Time:
		case StTypes::Date:
		case StTypes::TOD:
		case StTypes::DT:
		case StTypes::Lint:		return ConvertSscpValue<sint64b>(ptr, false, returnType, destination); break;
		case StTypes::Bool:		return ConvertSscpValue<bool>(ptr, true, returnType, destination); break;
		case StTypes::Byte:
		case StTypes::Usint:	return ConvertSscpValue<uint8b>(ptr, false, returnType, destination); break;
		case StTypes::Word:
		case StTypes::Uint:		return ConvertSscpValue<uint16b>(ptr, false, returnType, destination); break;
		case StTypes::Dword:
		case StTypes::Udint:	return ConvertSscpValue<uint32b>(ptr, false, returnType, destination); break;
		case StTypes::Lword:
		case StTypes::Ulint:	return ConvertSscpValue<uint64b>(ptr, false, returnType, destination); break;
		case StTypes::Real:  	return ConvertSscpValue<float>(ptr, false, returnType, destination); break;
		case StTypes::Lreal:	return ConvertSscpValue<double>(ptr, false, returnType, destination); break;
	}
	return (true);
}

bool SscpClient::ConvertVariableToPlc(uint8b *buffer, StTypes::StType type, bool bigEndian, void *source,
	StTypes::StType sourceType)
{
	uint8b tmp[8];
	uint8b *ptr = bigEndian ? buffer : tmp;

	switch (sourceType)
	{
		case StTypes::Sint:		ConvertSscpValue<sint8b>(source, false, type, ptr); break;
		case StTypes::Int:		ConvertSscpValue<sint16b>(source, false, type, ptr); break;
		case StTypes::Dint:		ConvertSscpValue<sint32b>(source, false, type, ptr); break;
		case StTypes::Time:
		case StTypes::Date:
		case StTypes::TOD:
		case StTypes::DT:
		case StTypes::Lint:		ConvertSscpValue<sint64b>(source, false, type, ptr); break;
		case StTypes::Bool:		ConvertSscpValue<bool>(source, true, type, ptr); break;
		case StTypes::Byte:
		case StTypes::Usint:	ConvertSscpValue<uint8b>(source, false, type, ptr); break;
		case StTypes::Word:
		case StTypes::Uint:		ConvertSscpValue<uint16b>(source, false, type, ptr); break;
		case StTypes::Dword:
		case StTypes::Udint:	ConvertSscpValue<uint32b>(source, false, type, ptr); break;
		case StTypes::Lword:
		case StTypes::Ulint:	ConvertSscpValue<uint64b>(source, false, type, ptr); break;
		case StTypes::Real:  	ConvertSscpValue<float>(source, false, type, ptr); break;
		case StTypes::Lreal:	ConvertSscpValue<double>(source, false, type, ptr); break;
	}

	if (!bigEndian)
	{
		switch (type)
		{
			case StTypes::Bool:
			case StTypes::Usint:
			case StTypes::Sint:
			case StTypes::Byte:
				{
					buffer[0] = *ptr;
				} break;
			case StTypes::Int:
			case StTypes::Uint:
			case StTypes::Word:
				{
					buffer[0] = ptr[1];
					buffer[1] = ptr[0];
				} break;
			case StTypes::Dint:
			case StTypes::Udint:
			case StTypes::Real:
			case StTypes::Dword:
				{
					buffer[0] = ptr[3];
					buffer[1] = ptr[2];
					buffer[2] = ptr[1];
					buffer[3] = ptr[0];
				} break;
			case StTypes::Lint:
			case StTypes::Ulint:
			case StTypes::Lreal:
			case StTypes::Time:
			case StTypes::DT:
			case StTypes::Date:
			case StTypes::Lword:
				{
					buffer[0] = ptr[7];
					buffer[1] = ptr[6];
					buffer[2] = ptr[5];
					buffer[3] = ptr[4];
					buffer[4] = ptr[3];
					buffer[5] = ptr[2];
					buffer[6] = ptr[1];
					buffer[7] = ptr[0];
				} break;
		}
		ptr = tmp;
	}

	return (true);
}
#endif
