/*!
 * \mainpage Shark Slave Communication Protocol (SSCP) Client
 *
 * SDK used for communication with PLC via SSCP protocol.
 *
 * \section Prerequirements
 * - OpenSLL for selected OS:
 *      - \__ANDROID__
 *      - \__OS_X__
 *      - \Windows phone OS (see below)
 *      - \__POSIX__
 *
 * \section Definitions
 * \todo dopsat co která definice znamená + doplnit chybějící negenerické přepínače!!!!
 *
 * \li \c GLOBAL_WDTMANAGER
 * \li \c _SSCP_DEBUG
 * \li \c STANDALONE_SUPPORT
 * \li \c SSCP_STANDALONE_CLIENT
 * \li \c SSCP_DISABLE_DELAYED_CONNECTION
 * \li \c \__POSIX__
 * \li \c \__FREERTOS__ - Builder for FREERTOS OS
 * \li \c _POSIX_THREADS
 * \li \c \__THREAD_ID__
 * \li \c \__ANDROID__ - Build for Android OS
 * \li \c \__OS_X__ - Build for iOS/OS X
 * \li \c \WINAPI_FAMILY=WINAPI_FAMILY_PHONE_APP - Build for Windows phone OS
 *
 * \section Usage
 * \code{.cpp}
 * #include <WinSock2.h>
 * #include <windows.h>
 * #include "SscpClient.h"
 * #include <conio.h>
 *
 * void printarray(uint8b *buff, uint32b size, uint8b div)
 * {
 *     for (uint32b i = 0; i < size; i++)
 *     {
 *         printf("%02x ", buff[i]);
 *         if ((div > 4) && ((i + 1) % 4 == 0))
 *             printf ("| ");
 *         if ((i + 1) % div == 0)
 *             printf("\n");
 *     }
 *     printf("\n");
 * }
 *
 * int main(int argc, char* argv[])
 * {
 * #if defined(_MSC_VER)
 *     WSADATA wsaData;
 *
 *     if ((WSAStartup(MAKEWORD(2,2), &wsaData)) != 0)
 *     {
 *         printf("ws32 init failed\n");
 *         exit(-1);
 *     }
 * #endif
 *
 *     SscpClient *cl = new SscpClient(1, 20000);
 *     if (cl)
 *     {
 *         SscpDevice *dev = new SscpDevice("192.168.1.121", 12346, 1, "admin", "rw", NULL);
 *         cl->AssignDevice(dev);
 *         cl->Initialize();
 *         Sleep(1000);
 *         uint8b value[4096];
 *         uint32b length = 0;
 *         if (cl->Login(dev) == SscpError::NoError)
 *         {
 *             printf("Connected\n");
 *             SscpError::SscpError res = SscpError::Unknown;
 *
 *             printf("** ReadSingleVariable\n");
 *             for (int i = 0; i < RETRIES; i++)
 *             {
 *                 SscpError::SscpError res = cl->ReadSingleVariable(dev, 5201, value);
 *                 if (res == SscpError::NoError)
 *                 {
 *                     printf("Data:\n");
 *                     printarray(value, 128, 16);
 *                 }
 *                 else
 *                 {
 *                     printf("Error: %s (%x)\n", GetSscpErrorText(res), cl->GetErrorFlagsForDestination(dev));
 *                     if (res == SscpError::NotLogged)
 *                     {
 *                         res = cl->Login(dev);
 *                     }
 *                 }
 *                 Sleep(10);
 *             }
 *
 *             printf("** ReadMultipleVariable\n");
 *             uint32b tmpComms[] = { 6067 };
 *
 *             for (int i = 0; i < RETRIES; i++)
 *             {
 *                 SscpError::SscpError res = cl->ReadMultipleVariables(dev, tmpComms, value, NULL, NULL, 1);
 *                 if (res == SscpError::NoError)
 *                 {
 *                     printf("Data:\n");
 *                     printarray(value, 128, 16);
 *                 }
 *                 else
 *                 {
 *                     printf("Error: %s (%x)\n", GetSscpErrorText(res), cl->GetErrorFlagsForDestination(dev));
 *                     if (res == SscpError::NotLogged)
 *                     {
 *                         res = cl->Login(dev);
 *                     }
 *                 }
 *                 Sleep(10);
 *             }
 *             printf("** WriteSingleVariable int16\n");
 *             uint16b i = 2100;
 *             uint8b *tmp = (uint8b*)&i;
 *             value[0] = tmp[1];
 *             value[1] = tmp[0];
 *             res = cl->WriteSingleVariable(dev, 17039, 0, 2, true, value);
 *             printf("Result: %x\n", res);
 *
 *             printf("** WriteSingleVariable\n");
 *             for (uint32b i = 0; i < 101; i++)
 *             {
 *                 value[4 * i + 0] = 0;
 *                 value[4 * i + 1] = 0;
 *                 value[4 * i + 2] = 0;
 *                 value[4 * i + 3] = i * 2;
 *             }
 *             res = cl->WriteSingleVariable(dev, 14958, 0, 0x38, true, value);
 *             printf("Result: %x\n", res);
 *             cl->Logout(dev);
 *         }
 *
 *         cl->DeInitialize();
 *
 *         delete (cl);
 *         cl = NULL;
 *
 *         printf("Finished\n");
 *         getch();
 *     }
 *
 *     return 0;
 * }
 * \endcode
 */
#ifndef __SscpClientH
#define __SscpClientH

#if defined(SSCPCLIENTLIB_LIBRARY) && defined(EXPORT_LIBRARY)
#  define SSCPCLIENTLIBSHARED_EXPORT __declspec(dllexport)
#elif defined(EXPORT_LIBRARY)
#  define SSCPCLIENTLIBSHARED_EXPORT __declspec(dllimport)
#else
#  define SSCPCLIENTLIBSHARED_EXPORT
#endif


#include "./datatypes.h"
#include "./SscpProtocolParser.h"
#include "./TcpClient.h"
#include "./ConditionalVariable.h"
#include "./NetworkingHelpers.h"

#define SSCP_CLIENT_EXTENDED_CONVERSION 1
/*!
 * \brief Driver state
 * \details Describe drivers current state
 */
enum DriverState
{
	/*!
	 * \brief Request is sended
	 */
	DriverState_SendRequest = 0,
	/*!
	 * \brief Driver read confirmation
	 */
	 DriverState_ReadConfirmation,
	 /*!
	 * \brief Driver is checking confirmation
	 */
	 DriverState_CheckConfirmation,
	 /*!
	 * \brief Driver is in iddle state
	 */
	 DriverState_Idle,
	 /*!
	* \brief Driver is reseting itself
	*/
	DriverState_Reset
};

/*!
 * \brief Property types enum
 * \details Used protocol types definition
 */
#ifndef SSCPTYPES
#define SSCPTYPES
/*!
 * \brief SSCP basic types
 */
namespace StTypes
{
	typedef enum
	{
		Bool = 0,

		Sint,
		Int,
		Dint,
		Lint,

		Usint,
		Uint,
		Udint,
		Ulint,

		Byte,
		Word,
		Dword,
		Lword,

		Real,
		Lreal,

		Time,
		Date,
		TOD,
		DT
	} StType;
}
#endif // SSCPTYPES

#ifndef TIME_COMMAND
#define TIME_COMMAND
/*!
 * \brief Timefunction commands
 */
namespace TimeCommand
{
	typedef enum
	{
		/*!
		 * \brief Get RTC in UTC format
		 */
		GetUTC = 0x01,
		/*!
		 * \brief Get RTC in local format
		 */
		 GetLocal = 0x02,
		 /*!
		  * \brief Set RTC from UTC format (timestamp required)
		  */
		  SetUTC = 0x10,
		  /*!
		  * \brief Set RTC from local format (timestamp required)
		  */
		  SetLocal = 0x11,
		  /*!
		 * \brief Get timezone offset (returned as timestamp)
		 */
		 GetTimeZone = 0x20,
		 /*!
		  * \brief Get daylight saving offset (returned as timestamp)
		  */
		  GetDaylight = 0x21
	} TimeCommand;
}
#endif // TIME_COMMAND
/*!
 * \brief SSCP endpoint context
 * \details Communication driver for sending an receiving SSCP messages. This object is common communication object.
 */
class SSCPCLIENTLIBSHARED_EXPORT SscpEndpointContext
{
	public:
	/*!
	 * \brief Buffer
	 * \details Communication buffer
	 */
	CommBuffer Buffer;
	/*!
	 * \brief Complete event
	 * \details Object for event responds to sended messages
	 */
	ConditionalVariable CompleteEvent;
	/*!
	 * \brief State
	 * \details Current driver state
	 */
	DriverState State;
	/*!
	 * \brief Error flags
	 * \details Drivers error state
	 */
	uint64b ErrorFlags;

	/*!
	 * \brief Constructor
	 * \param[in] bufferSize Buffer size_t
	 * \details Initialize communication buffer
	 */
	SscpEndpointContext(uint32b bufferSize)
	{
		Buffer.AllocateBuffer(bufferSize);
		State = DriverState_Idle;
	}
};

/*!
 * \brief SSCP Device definition
 * \details Object represented SSCP device. This device represent current PLC for communication
 *
 * \ref Usage
 */
class SSCPCLIENTLIBSHARED_EXPORT SscpDevice
{
	public:
	/*!
	 * \brief IP address
	 * \details Current devices IP address
	 */
	uint32b Ipv4;
	/*!
	 * \brief IP port
	 * \details Current devices IP port
	 */
	uint16b Port;

	/*!
	 * \brief Endpoint id
	 * \details End point identifier
	 */
	sint8b EndPoint;

	/*!
	 * \brief Slave address
	 * \details Modbas slave address definition
	 */
	uint8b SlaveAddress;
	/*!
	 * \brief Username
	 * \details User name for connect to client
	 */
	char *Username;
	/*!
	 * \brief Password
	 * \details Hashed password for setted Username (MD5)
	 * \sa Username
	 */
	uint8b Password[16]; // MD5
	/*!
	 * \brief Proxy ID
	 * \details
	 * \todo Jak funguje tato proxy?? Co tam má být??
	 */
	char *ProxyId;

	/* setup via login function */
	/*!
	 * \brief Protocol version
	 * \details Protocol version from PLC. This version is running in PLC and is getting from Login function
	 * \sa SscpClient::Login
	 */
	uint8b ProtocolVersion; // TODO: ReadOnly???
	/*!
	 * \brief Server data size
	 * \details Server data size, buffer length limit on server side. This value is getting from Login function
	 * \sa SscpClient::Login
	 */
	uint16b ServerDataSize;    // TODO: ReadOnly???
	/*!
	* \brief Max data size
	* \details Calculated as minimum of server (ServerDataSize) and client (EndpointContext) buffer length after login.
	*/
	uint16b MaxDataSize;    // TODO: ReadOnly???
	/*!
	 * \brief Rights
	 * \details Users rights. This value is getting from Login function
	 * \sa SscpClient::Login
	 */
	ConnectionRights Rights;// TODO: ReadOnly???
	/*!
	 * \brief Image GUID
	 * \details GUID actually running application. This value is getting from Login function
	 * \sa SscpClient::Login
	 */
	uint8b ImageGuid[16];   // TODO: ReadOnly???

	/*!
	 * \brief Constructor
	 * \param[in] ipv4 IP address (uint32b format)
	 * \param[in] port IP Port
	 * \param[in] slaveAddress ModBus slave address (0 for current PLC)
	 * \param[in] username User name
	 * \param[in] password Password
	 * \param[in] proxyId Proxy ID
	 * \param[in] md5PassFormat Password is hashed (true)
	 * \details Standard constructor. Set current SSCP device class.
	 */
	SscpDevice(uint32b ipv4, uint16b port, uint8b slaveAddress, char *username, char *password, char *proxyId, bool md5PassFormat = false);
	/*!
	 * \brief Constructor
	 * \param[in] ipv4 IP address (string)
	 * \param[in] port IP Port
	 * \param[in] slaveAddress ModBus slave address (0 for current PLC)
	 * \param[in] username User name
	 * \param[in] password Password
	 * \param[in] proxyId Proxy ID
	 * \param[in] md5PassFormat Password is hashed (true)
	 * \details Standard constructor. Set current SSCP device class.
	 */
	SscpDevice(char *ipv4, uint16b port, uint8b slaveAddress, char *username, char *password, char *proxyId, bool md5PassFormat = false);

	/*!
	 * \brief Calculate MD5
	 * \param[out] destination Hash
	 * \param[in] password Password
	 * \details Function for calculate MD5 hash. Hash is calculated from \c password and result is written to \c destination. Destination must have at least 16bytes
	 */
	void CalculateMD5(unsigned char *destination, unsigned char *password);
};

/*!
 * \brief SSCP Client definition
 * \details Object represented SSCP Client. Client is only one and represents communication layer. One client can have more SSCP devices
 *
 * \ref Usage
 *
 * \sa SscpDevice
 */
class SSCPCLIENTLIBSHARED_EXPORT SscpClient : ICallbackCommunicationDriver
{
	private:
	/*!
	 * \brief Tcp Client
	 * \details Communication layer object
	 */
	TcpUdpClient *m_TcpClient;
	/*!
	 * \brief Parser
	 * \details SSCP data parser
	 */
	SscpProtocolParser *m_Parser;

	/*!
	 * \brief Max endpoints
	 * \details Maximum of connected devices
	 */
	uint8b m_MaxEndpoints;
	/*!
	 * \brief Assigned endpoints count
	 * \details Currently assigned devices endpoints. This value cannot be bigger than m_MaxEndpoints
	 * \sa m_MaxEndpoints, AssignEndPoint
	 */
	uint8b m_AssignedEpCount;
	/*!
	 * \brief EP Context
	 * \details Array of created SSCP device context. Every connected device has one object in this array. Array length coresponds with m_AssignedEpCount
	 * \sa m_AssignedEpCount, AssignEndPoint
	 */
	SscpEndpointContext **m_EPContext;

	/*!
	 * \brief Timeout
	 * \details Communication timeout in ms
	 */
	uint32b m_Timeout;

	/*!
	 * \brief Assign new endPoint
	 * \param[in] epIndex Endpoint identifier
	 * \param[in] ipv4 IP address
	 * \param[in] port IP port
	 * \param[in] timeout Communication timeout in ms
	 * \return
	 * \li \c true - Object created and assigned
	 * \li \c false - Error while creating or asignig the device
	 *
	 * \details Function create and assign new device endpoint to the client.
	 */
	bool AssignEndPoint(uint8b epIndex, uint32b ipv4, uint16b port, uint32b timeout = 4000);

	/*!
	 * \brief Common check and part
	 * \param[in] device Selectged device
	 * \return
	 * \li \c true - Checking success
	 * \li \c false - Checking failed (device not specified, device has no endpointID, missing SSCP data parser)
	 *
	 * \details Check function. This function check device properties and initialize communication with this device.
	 */
	bool CommonCheckAndPart(SscpDevice *device);
	/*!
	 * \brief Send and received response
	 * \param[in] device Selected device
	 * \param[in] response Flag for reading response
	 * \li \c true - Wait for response
	 * \li \c false - Don't wait to response
	 *
	 * \return
	 * \li \c true - Sending success
	 * \li \c false - Sending failed
	 *
	 * \details Send data to \c device and read responds. Function is blocked.
	 */
	bool SendAndReceivedResponse(SscpDevice *device, bool response);

	public:
	/*!
	 * \brief Constructor
	 * \param[in] maxEndpoints Presumed endpoints count
	 * \param[in] timeout Communication timeout
	 * \param[in] maxBufferSize Maximum buffer size for client
	 * \details Initialize instance of SscpClient object
	 */
	SscpClient(uint8b maxEndpoints, uint32b timeout = 4000, uint32b maxBufferSize = SSCP_MAX_DATA_SIZE);
	/*!
	 * \brief Destructor
	 * \details Deinitialize client and free allocated memmory.
	 */
	~SscpClient();

	/*!
	 * \brief Assign new device
	 * \param[in] device Device
	 * \param[in] forceCreateNewEndpoint Add device without existency check (true)
	 * \return
	 * \li \c true - Creation success
	 * \li \c false - Creation falied
	 * \details Assign new device to current client
	 * \sa AssignEndPoint
	 */
	bool AssignDevice(SscpDevice *device, bool forceCreateNewEndpoint = false);

	/*!
	 * \brief Initialize
	 * \return
	 * \li \c true - initialization success
	 * \li \c false - initialization failed
	 * \details Initialize current client. This function must be called first
	 *
	 * \ref Usage
	 */
	bool Initialize(void);
	/*!
	 * \brief DeInitialize
	 * \return
	 * \li \c true - initialization success
	 * \li \c false - initialization failed
	 * \details Deinitialize current client. After calling this function the client is not able to communicate wit defined devices.
	 */
	bool DeInitialize(void);

	/*!
	 * \overload ICallbackCommunicationDriver::Reset
	 */
	virtual void Reset(uint32b tag);

	/*!
	 * \overload ICallbackCommunicationDriver::GetNextAction
	 */
	virtual CallbackAction GetNextAction(uint32b tag);
	/*!
	 * \overload ICallbackCommunicationDriver::GetCommunicationBuffer
	 */
	virtual CommBuffer* GetCommunicationBuffer(uint32b tag);

	/*!
	 * \brief Login
	 * \param[in] device Selected device
	 * \return Communication error
	 * \details Call login function via SSCP protocol (__0x01 0x00__) on selected device with setted user name and password. This function must be called befor rest communication with device.
	 *
	 * There are 3 login level. Engineering , Full control and Readonly. The PLC will deny (insufficient rights) execution of functions that are not available at the level of current ogin.
	 *
	 * _Note_: Login is expected as the first command after connection establishment. Any other function will close the connection immediately .
	 *
	 * \sa SscpDevice, SscpFunction, SscpError::SscpError
	 */
	SscpError::SscpError Login(SscpDevice *device);
	/*!
	 * \brief Logout
	 * \param[in] device Selected device
	 * \return Communication error
	 * \details Call logout function via SSCP protocol (__0x01 0x01__) on selected device.
	 *
	 * This function doesn't break communication, only logout user and next communication falls. For next communication user must be logged in.
	 *
	 * There is no response to this message. The connection is closed by the PLC immediately.
	 *
	 * \sa SscpDevice, Login, SscpFunction, SscpError::SscpError
	 */
	SscpError::SscpError Logout(SscpDevice *device);

	/*!
	 * \brief Read single variable from device
	 * \param[in] device Selected device
	 * \param[in] commUid Variable ID
	 * \param[out] buffer Output buffer with readed data
	 * \param[in] offset Variable offset
	 * \param[in] length Variable length
	 * \return Communication error
	 * \details Call ReadVariablesDirectly function via SSCP protocol (__0x05 0x00__) on selected device.
	 *
	 * Every variable is assigned a ID number (4 bytes unsigned int) that does not change with program changes. Once assigned, the ID can only change when a specific action is taken (manual reordering or data type changes).
	 *
	 * _Note_: It is the responsibility of the development environment to keep the numbers correct.
	 *
	 * Clients must be aware of data types and sizes of all variables and layouts of complex variables (strings, structs, FBs, arrays). All information needed can be retrieved from the device as a file “/sys/vars”. The format is based on the VM’s image definition.
	 *
	 * If readed data was to large, this function would use method for geting larg data (/var/direct file transfer) automaticaly.
	 *
	 * \sa SscpFunction, SscpError::SscpError
	 */
	SscpError::SscpError ReadSingleVariable(SscpDevice *device, uint32b commUid, uint8b *buffer, uint32b offset = 0, uint32b length = 0);
	/*!
	 * \brief Read multiple variables from device
	 * \param[in] device Selected device
	 * \param[in] commUid Array of Variables ID
	 * \param[out] buffer Array of output data for selected device
	 * \param[in] offset Array of Variables offset
	 * \param[in] length Array of Variables loength
	 * \param[in] count Variables length
	 * \return Communication error
	 * \details This function is similar to ReadSingleVariable but for more than one variable.
	 *
	 * Arrays must be synchronous. It means, that offset[i], length[i] correspods to commUid[i].
	 *
	 * \sa SscpFunction, SscpError::SscpError, ReadSingleVariable
	 */
	SscpError::SscpError ReadMultipleVariables(SscpDevice *device, uint32b *commUid, uint8b *buffer, uint32b *offset, uint32b *length, uint32b count);
	/*!
	 * \brief Register variable group
	 * \param[in] device Selected device
	 * \param[in] groupId Group ID
	 * \param[in] variables Array of Variable definition
	 * \param[in] count Variables count
	 * \return Communication error
	 * \details Call RegisterVariableGroup function via SSCP protocol (__0x05 0x05__) on selected device.
	 *
	 * A client must create a group first and then read variables using group’s identification.
	 *
	 * \li Group registration
	 *
	 * Clients register a group of variables under a group ID. The maximal number of groups is device dependent. The maximal number of all variables in all groups is also device dependent. Variables can be arbitrarily distributed among groups and number of variables in a group is limited only by the total limit and number of variables in other groups.
	 *
	 * All clients have their own independent groups.
	 *
	 * Redefining a group is done by registering new variables. Registering zero variables means that the group has no variables registered.
	 *
	 * When a group is registered, clients can read the values directly or as binary transfer of files with values (see further for details). It depends on the data size.
	 *
	 * Registered groups can serve for group writes, too.
	 *
	 * \todo Co se stane, pokud je stejné GroupID??? Původní grupa se přepíše? Toto je nutné doplnit
	 *
	 * \sa SscpFunction, SscpError::SscpError, ReadVariableGroup, WriteVariableGroup
	 */
	SscpError::SscpError RegisterVariableGroup(SscpDevice *device, uint8b groupId, VariableDefinition *variables, uint32b count);
	/*!
	 * \brief Read variable group from device
	 * \param[in] device Selected device
	 * \param[in] groupId Group ID
	 * \param[in] flags Variable read flag
	 * \param[out] buffer Output buffer
	 * \param[in] forceFileTransfer Flag for force using large data block communication (/var/direct file transfer). This funcionality is used automaticaly
	 * \param[out] length Output data length
	 * \return Communication error
	 * \details Call ReadVariableGroupDirectly function via SSCP protocol (__0x05 0x06__) on selected device.
	 *
	 * This function is similar to ReadSingleVariable but for previously registred group.
	 *
	 * \sa SscpFunction, SscpError::SscpError, VarAccessFlags::VarAccessFlags, RegisterVariableGroup
	 */
	SscpError::SscpError ReadVariableGroup(SscpDevice *device, uint8b groupId, VarAccessFlags::VarAccessFlags flags, uint8b *buffer,
		bool *forceFileTransfer, uint32b *length);

	/*!
	 * \brief Write single variable data to device
	 * \param[in] device Selected device
	 * \param[in] commUid Variable ID
	 * \param[in] offset Variable offset
	 * \param[in] length Variable length
	 * \param[in] commUidOnly Flag for using commUid only for Variable detection (true -> offset = length = 0)
	 * \param[in] buffer Data to be written
	 * \return Communication error
	 * \details Call WriteVariableDirectly function via SSCP protocol (__0x05 0x10__) on selected device.
	 *
	 * Direct writes sets the variables directly without necessity of creation of groups. All writes that fit into maximal packet size can be set directly without accessing the file.
	 *
	 * \sa SscpFunction, SscpError::SscpError
	 */
	SscpError::SscpError WriteSingleVariable(SscpDevice *device, uint32b commUid, uint32b offset, uint32b length, bool commUidOnly, uint8b *buffer);
	/*!
	 * \brief Write multiple variables data to device
	 * \param[in] device Selected device
	 * \param[in] commUid Array of Variables ID
	 * \param[in] offset Array of Variables offset
	 * \param[in] length Array of Variables length
	 * \param[in] count Variables count
	 * \param[in] commUidOnly Flag for using commUid only for Variable detection (true -> offset = length = 0)
	 * \param[in] bufferWithAllValues Data to be written
	 * \return Communication error
	 * \details This function is similar to ReadSingleVariable but for more than one variable.
	 *
	 * Arrays must be synchronous. It means, that offset[i], length[i] correspods to commUid[i].
	 *
	 * \sa SscpFunction, SscpError::SscpError, WriteSingleVariable
	 */
	SscpError::SscpError WriteMultipleVariables(SscpDevice *device, uint32b *commUid, uint32b *offset, uint32b *length, uint32b count,
		bool commUidOnly, uint8b *bufferWithAllValues);
	/*!
	 * \brief Write variable data to group
	 * \param[in] device Selected device
	 * \param[in] groupId Group ID
	 * \param[in] flags Variable write flag
	 * \param[in] buffer Data to be written
	 * \param[in] length Data length
	 * \return Communication error
	 * \details Call WriteVariableGroupDirectly function via SSCP protocol (__0x05 0x11__) on selected device.
	 *
	 * This function is similar to WriteSingleVariable but for previously registred group.
	 *
	 * \sa SscpFunction, SscpError::SscpError, VarAccessFlags::VarAccessFlags, RegisterVariableGroup
	 */
	SscpError::SscpError WriteVariableGroup(SscpDevice *device, uint8b groupId, VarAccessFlags::VarAccessFlags flags, uint8b *buffer, uint32b length);
	/*!
	 * \brief Write variable data to group
	 * \param[in] device Selected device
	 * \param[in] groupId Group ID
	 * \param[in] flags Variable write flag
	 * \param[in] indexes
	 * \param[in] idxCount
	 * \param[in] buffer Data to be written
	 * \param[in] length Data length
	 * \return Communication error
	 * \details This function is similar to WriteVariableGroup but this function can writes some variables, not all.
	 *
	 * The request is a sequence of pairs {variable index; variable value}. Variable index is the index of the variable (and offset and length) during the registration.
	 *
	 * \sa SscpFunction, SscpError::SscpError, VarAccessFlags::VarAccessFlags, RegisterVariableGroup, WriteVariableGroup
	 */
	SscpError::SscpError WriteVariableGroup(SscpDevice *device, uint8b groupId, VarAccessFlags::VarAccessFlags flags, uint8b *indexes, uint8b idxCount,
		uint8b *buffer, uint32b length);

	/*!
	 * \brief Read variable structure from device
	 * \param[in] device Selected device
	 * \param[in] commUid Variable ID
	 * \param[out] buffer Output structure
	 * \param[out] length Output data length
	 * \return Communication error
	 * \details Call ReadVariableStructure function via SSCP protocol (__0x05 0x15__) on selected device.
	 *
	 * It is possible to read variable memory layout.
	 *
	 * Variable structure is constructed by string of tags. Tag is one octet and has following structure
	 *
	 * bit 7 | bit 6 | bit 5 | bit 4 | bit 3 | bit 2 | bit 1 | bit 0
	 * ----- | ----- | ----- | ----- | ----- | ----- | ----- | -----
	 * Tag#  | Tag#  | Tag#  | Tag#  | Ap/Cx | Len   | Len   | Len
	 *
	 * Legenda:
	 * __Tag#3-0__: 4 bit tag number
	 * __Ap/Cx__: Tag type - application(=0) or context (=0). Application tags are well known and contexts are extensions and specific data
	 * __Len2-0__: Data length
	 *
	 * Tag number range is from 0to 14, if it is required to have tag number greater than 14 then tag extending is used. In this case, the tag number is set to 15 and correct tag number is added as next octet, eg. tag number 120 is encoded as two octets 0xF? 0x78.
	 * Data length encoding is also specific. Lengths from zero to four is encoded directly in tag. Length greater than 4 and lower than 254 is encoded as two bytes, length in tag is set to five and next octet is set to correct length (if tag number has been extended then length follow after extended tag number, this is true in all situations). Length greater then 253 and lower than 65535 is encoded as four octets, length in tag is set to five again, next octet is set to 254 and next two octets contains correct length. Lastly, length from 65536 to 2 32 1 are encoded as six octets, length in tag is set to five again, next octet is set to 255 and next four octets contains correct length.
	 *
	 * _Next table show tag number for application tags used in actual version of Virtual machine_
	 *
	 * Data type | Tag number | Ap/Cx | Length
	 * --------- | ---------: | ----: | ------:
	 *  NULL     | 0          | 0     | 0
	 *  uint8    | 2          | 0     | 1
	 *  sint8    | 3          | 0     | 1
	 *  uint16   | 2          | 0     | 2
	 *  sint16   | 3          | 0     | 2
	 *  uint32   | 2          | 0     | 4
	 *  sint32   | 3          | 0     | 4
	 *  uint64   | 2          | 0     | 8
	 *  sint64   | 3          | 0     | 8
	 *  float32  | 4          | 0     | 4
	 *  float64  | 5          | 0     | 8
	 *  enum     | 9          | 0     | ?
	 *  addr     | 12         | 0     | 8
	 *
	 * _Next table show tag number for context tags used in actual version of Virtual machine_
	 *
	 * Data type    | Tag number | Ap/Cx | Length
	 * ------------ | ---------: | ----: | -------:
	 *  array       | 1          | 1     | array length
	 *  opening tag | 5          | 1     | 6
	 *  closing tag | 5          | 1     | 7
	 *
	 * Variable memory layout is always constructed from these tags, complex types (structures and function blocks) is always enclosed by opening and closing tag. Arrays is constructed as array tag followed by item structure enclosed in opening and closing tag.
	 *
	 * __Encoding examples__
	 *
	 * _Note_: All numbers are in hex
	 *
	 * _Returned structure_: 5E 32 94 31 1D 10 5E 5E 32 32 32 5F 5F 5F
	 *
	 * _Hierarchy_:
	 * 5E
	 *     32
	 *     94
	 *     31
	 *     1D 10 5E
	 *           5E
	 *               32
	 *               32
	 *               32
	 *           5F
	 *     5F
	 * 5F
	 *
	 * _This will be interpreted as_: structure with sint16, four bytes enumerated, sint8 and sixteen item array consisting of structure with three sint16.
	 *
	 * \sa SscpFunction, SscpError::SscpError
	 */
	SscpError::SscpError ReadVariableStructure(SscpDevice *device, uint32b commUid, uint8b *buffer, uint32b *length);

	/*!
	 * \brief Read file from device
	 * \param[in] device Selected device
	 * \param[in] filename File name with path
	 * \param[out] buffer File data
	 * \param[out] fileSize File size
	 * \return Communication error
	 * \details Call InitiateDataReceive function via SSCP protocol (__0x02 0x10__) on selected device.
	 *
	 * Read common file from device.
	 *
	 * Path                           | Description                                        | Comment
	 * ------------------------------ | -------------------------------------------------- | --------------------
	 *  /sys/sexm                     | In memory RT’s image.                              | According to capabilities (usually 2MB)
	 *  /sys/rt                       | Area for upgrade of RT                             | -
	 *  /sys/sex1                     | Area for the first shark image.                    | According to capabilities (usually 2MB)
	 *  /sys/sex2                     | Area for the second shark image.                   | According to capabilities (usually 2MB)
	 *  /sys/caps                     | PLC capabilites.                                   | See appendix for file format. This file describes the PLC as much as possible. Usually, it will be used by IDE.
	 *  /var/direct                   | Variable value/descriptions direct query           | -
	 *  /var/group/[groupId]/[flags]  | Variable value/description for registered group    | -
	 *  /sys/log                      | PLC log                                            | PLC Log.
	 *  /drv/result                   | Response for driver command longer than one packet | -
	 *  /d/...                        | Mountpoint for general filesystem                  | Web,...
	 *
	 * _Access privileges for file access_
	 *
	 * File name          | Minimal required privilege for read | Minimal required privilege for write
	 * ------------------ | ----------------------------------- | -------------------------------------
	 *  /sys/sex1         | Engineering                         | Engineering
	 *  /sys/sex2         | Engineering                         | Engineering
	 *  /sys/sexm         | Engineering                         | Engineering
	 *  /sys/rt           | Engineering                         | Engineering
	 *  /sys/config       | not to be readeable directly        | not to be written directly
	 *  /sys/cfgpass      | not to be readeable directly        | not to be written directly
	 *  /sys/caps         | Read                                | only not writable
	 *  /var/direct       | Read                                | only Full control
	 *  /var/group/\*     | Read                                | only Full control
	 *  /vars/flash       | not to be readeable directly        | not to be written directly
	 *  /log Engineering  | not writable                        | -
	 *  /sys/certstore    | Engineering                         | Engineering
	 *  /pm/\*            | Engineering                         | not writable
	 *
	 * \sa SscpFunction, SscpError::SscpError
	 */
	SscpError::SscpError ReadFile(SscpDevice *device, const char *filename, uint8b* buffer, uint32b *fileSize);
	/*!
	 * \brief Write file to device
	 * \param[in] device Selected device
	 * \param[in] filename File name with path
	 * \param[in] buffer File data
	 * \param[in] dataSize File size
	 * \param[in] timeStamp File creation datetime (DATE_TIME format)
	 * \return Communication error
	 * \details Call InitiateDataSend function via SSCP protocol (__0x02 0x00__) on selected device.
	 *
	 * Try to write file with \c filename to device. Directory access rights are described by function \c ReadFile.
	 *
	 * \sa SscpFunction, SscpError::SscpError, ReadFile
	 */
	SscpError::SscpError WriteFile(SscpDevice *device, const char *filename, uint8b* buffer, uint32b dataSize, uint64b timeStamp);

	/*!
	 * \brief Time sync
	 * \param[in] device Selected device
	 * \param[in] time Optional 64bit C# timestamp in UTC format for time preset. If time = 0, function reads actual time
	 * \param[out] returnedTime Setted time (64bit C# timestamp in UTC format)
	 * \param[in/out] cmd (Available from protocol version 7) Time function type. Set the requested command (as in) and get the returned value cmd type. Get RTC in UTC format is a default value
	 * \return Communication error
	 * \details Call TimeSetup function via SSCP protocol (__0x06 0x02__) on selected device.
	 *
	 * Sets/gets the datetime on device manualy
	 *
	 * \sa SscpFunction, SscpError::SscpError, TimeCommand::TimeCommand
	 */
	SscpError::SscpError TimeSync(SscpDevice *device, uint64b time, uint64b *returnedTime, TimeCommand::TimeCommand *cmd = NULL);

	/*!
	 * \brief Get last communication error flags for destination
	 * \param[in] device Selected device
	 * \return Device error state
	 * \details Get last communication error. Mask for readed properties. If bit is set, property is in error state
	 */
	uint64b GetErrorFlagsForDestination(SscpDevice *device);

#if defined(SSCP_CLIENT_EXTENDED_CONVERSION)
	/*!
	 * \brief Convert variable from PLC variable
	 * \param[in] buffer Readed value from PLC
	 * \param[in] type PLC variable type
	 * \param[in] bigEndian BigEndian flag
	 * \param[out] destination Pointer to local variable
	 * \param[in] returnType Local variable type
	 * \return
	 * \li \c true - conversion success
	 * \li \c false - cannot convert value
	 *
	 * \details Static function for conversion varible from PLC to local variable
	 */
	static bool ConvertVariableFromPlc(uint8b *buffer, StTypes::StType type, bool bigEndian, void *destination, StTypes::StType returnType);
	/*!
	 * \brief Convert variable to PLC variable
	 * \param[out] buffer PLC value
	 * \param[in] type PLC variable type
	 * \param[in] bigEndian BigEndian flag
	 * \param[in] source Pointer to local variable
	 * \param[in] sourceType Local variable type
	 * \return
	 * \li \c true - conversion success
	 * \li \c false - cannot convert value
	 *
	 * \details Static function for conversion varible to PLC from local variable
	 */
	static bool ConvertVariableToPlc(uint8b *buffer, StTypes::StType type, bool bigEndian, void *source, StTypes::StType sourceType);
#endif
};

#endif
