/*!
 * \file SscpProtocolCommon.h
 *
 * \todo Revidovat kvůli nepotřebným objektům a strukturám!!!!
 */
#ifndef __SSCPProtocolCommon_h
#define __SSCPProtocolCommon_h

#include <stdlib.h>
#include "./SSCPProtocolErrors.h"

#if !defined(SSCP_STANDALONE_CLIENT)
#define SSCP_ENABLE_SERVER	1
#define SSCP_ENABLE_CLIENT	1
#else
#undef SSCP_ENABLE_SERVER
#define SSCP_ENABLE_CLIENT	1
#endif

#if !defined(SSCP_STANDALONE_CLIENT)
#include "./../Engine/EngineDefinitions.h"
#else
#include "./EngineDefinitions.h"
#endif

/*!
 * \brief Namespace for SscpFunction enumeration
 */
namespace SscpFunction
{
    /*!
     * \brief SSCP function codes
     * \enum SscpFunction
     * \details Define SSCP functions.
     *
     * _Privileges for commands_
     *
     * Command                     | Minimal required privilege
     * --------------------------- | ----------------------------
     *  GetBasicInfo               | Engineering, if RequestedSize is equal to zero no valid login is required
     *  SetBasicInfo               | Engineering
     *  Login                      | not applicable
     *  Logout                     | not applicable
     *  InitiateDataSend           | For file transfer see \ref SscpClient::ReadFile
     *  SendDataChunk              | For file transfer see \ref SscpClient::ReadFile
     *  FinishDataSend             | For file transfer see \ref SscpClient::ReadFile
     *  InitiateDataReceive        | For file transfer see \ref SscpClient::ReadFile
     *  ReceiveDataChunk           | For file transfer see \ref SscpClient::ReadFile
     *  GetPlcStatistics           | Read only
     *  GetTaskStatistics          | Read only
     *  GetChannelStatistics       | Read only
     *  ReadVariablesDirectly      | Read only
     *  RegisterVariableGroup      | Read only
     *  ReadVariableGroupDirectly  | Read only
     *  WriteVariableDirectly      | Full control
     *  WriteVariableGroupDirectly | Full control
     *  ReadVariableStructure      | Read only
     *  SetResetReadOnlyMode       | Engineering / Full control
     *  ExecuteSystemCommand       | Engineering
     *  GetStacks                  | Read only
     *  TimeSetup                  | Read only or Full control for set
     */
	typedef enum
	{
        /*!
         * \brief 0x0000
         * \details Retrieves the basic information from a device.
         */
		GetBasicInfo = 0x0000,
        /*!
         * \brief 0x0001
         * \details Sets the basic information from a device
         */
        SetBasicInfo = 0x0001,

        /*!
         * \brief 0x0100
         * \details Login user to device and read setting from device. Also get users privileges.
         */
        Login = 0x0100,
        /*!
         * \brief 0x0101
         * \details Logout user from device. The connection is closed by the PLC immediately
         */
        Logout = 0x0101,

        /*!
         * \brief 0x0200
         * \details Trnasmit data from PC to device. Initislization function
         */
        InitiateDataSend = 0x0200,
        /*!
         * \brief 0x0201
         * \details Trnasmit data from PC to device. Data parts
         */
        SendDataChunk = 0x0201,
        /*!
         * \brief 0x0202
         * \details Trnasmit data from PC to device. End transmittion
         */
        FinishDataSend = 0x0202,

        /*!
         * \brief  0x0210
         * \details Trnasmit data from device to PC. Initislization function
         */
        InitiateDataReceive = 0x0210,
        /*!
         * \brief 0x0211
         * \details Trnasmit data from device to PC. Data parts
         */
        ReceiveDataChunk = 0x0211,

        /*!
         * \brief 0x0300
         * \details Gets devices state and other states
         */
        GetPlcStatistics = 0x0300,
        /*!
         * \brief 0x0301
         * \details Gets RT applications statistics
         */
        GetTaskStatistics = 0x0301,

        /*!
         * \brief 0x0500
         * \details Reading value from PLC directly
         */
        ReadVariablesDirectly = 0x0500,

        /*!
         * \brief 0x0505
         * \details Register variables group for faster reading/writing values to this attributes
         */
        RegisterVariableGroup = 0x0505,
        /*!
         * \brief 0x0506
         * \details Read data from variables in group
         */
        ReadVariableGroupDirectly = 0x0506,

        /*!
         * \brief 0x0510
         * \details Write data to variable directly
         */
        WriteVariableDirectly = 0x0510,
        /*!
         * \brief 0x0511
         * \details Write data from variables in group
         */
        WriteVariableGroupDirectly = 0x0511,

        /*!
         * \brief 0x0515
         * \details It is possible to read variable memory layout
         * \todo Zkontrolovat
         */
        ReadVariableStructure = 0x0515,

        /*!
         * \brief 0x0520
         * \details This function is used for manual control of variable value.
         */
        SetResetReadOnlyMode = 0x0520,

        /*!
         * \brief 0x0601
         * \details PLC control commands have direct influence on PLC’s behavior.
         */
        ExecuteSystemCommand = 0x0601,
        /*!
         * \brief 0x0602
         * \details Function for setting devices time
         */
        TimeSetup = 0x0602,
        /*!
         * \brief 0x0603
         * \details This command is used to send messages to communication channel.
         */
        DriverCommand = 0x0603,
        /*!
         * \brief 0x0602
         * \details Function for setting devices time
         */
        TimeSetupEx = 0x0604,

        /*!
         * \brief 0x0610
         * \details This function is used for retrieving call or computation stack content
         */
        GetStacks = 0x0610,

	} SscpFunction;
}

#if defined(_MSC_VER)
	#if defined(_DEBUG)
	#define SSCP_MAX_DATA_SIZE 228
	#else
	#define SSCP_MAX_DATA_SIZE 2048
	#endif
#elif defined(__POSIX__)
#define SSCP_MAX_DATA_SIZE 2048
#else
#define SSCP_MAX_DATA_SIZE 228
#endif

#define SSCP_PROTOCOL_VERSION											0x2

#define READ_ONLY_MODE_REQ_STATE_MASK							0x80
#define READ_ONLY_MODE_RESET_ALL_TO_DEFAULT_MASK	0x40

#if defined(_MSC_VER)
#include <stdio.h>
#endif

#define SSCP_READ_VARIABLES_FLAGS_RESPONSE_FORMAT_MASK 0x3F
#define SSCP_READ_VARIABLES_FLAGS_MODE_MASK 0x40
#define SSCP_READ_VARIABLES_FLAGS_USE_LENGTH_MASK 0x80

#define SSCP_UDP_LAST_SEGMENT		0x80
/*!
 * \brief Struct for variable definition
 * \details Struct for variable definition. This struct define one variable
 */
struct VariableDefinition
{
	public:
        /*!
         * \brief Variable ID
         * \details ID number (4 bytes unsigned int) that does not change with program changes
         */
		uint32b Uid;
        /*!
         * \brief Variable offset
         * \details Variable offset for complex data types
         */
		uint32b Offset;
        /*!
         * \brief Variable length
         * \details Variable length. Almost it coresponds with variable type
         */
		uint32b Length;

        /*!
         * \brief Changed
         * \details
         * \todo Co to je, k čemu je to dobrý?
         */
		bool Changed; // fixme: nejak vylepsit, tady to neni nejlepsi misto a je to zdvojeny s userdaty u CVariable

        /*!
         * \brief Constructor
         * \details Initialize current object
         */
		VariableDefinition()
		{
			Uid = Offset = Length = 0;
		}

        /*!
         * \brief Fill
         * \param[in] uid Variable ID
         * \param[in] offset Variable offset
         * \param[in] length Variable length
         * \details Set the object with selected values
         */
		void Fill(uint32b uid, uint32b offset, uint32b length)
		{
			Uid = uid;
			Offset = offset;
			Length = length;
		}
};
/*!
 * \brief Struct for definition variables group
 * \todo SERVER!!!! - Odkomentovat
 */
struct VariableGroup
{
	public:
        /*!
         * \brief Count
         * \details Variables count
         */
		uint32b Count;
        /*!
         * \brief Items
         * \details Array of variables
         */
		VariableDefinition *Items;
        /*!
         * \brief Max items
         * \details Maxumum items connected to this object
         * \todo Je to pravda?
         */
		uint8b MaxItems;
        /*!
         * \brief Registered
         * \details
         * \todo Doplnit, k čemu to tu je
         */
		bool Registered;
        /*!
         * \brief UseVmUid
         * \details
         * \todo Doplnit, k čemu to tu je
         */
        bool UseVmUid;

        /*!
         * \brief Constructor
         * \param[in] maxItems
         * \details Initialize object.
         */
		VariableGroup(uint8b maxItems)
		{
			Count = 0;
			Registered = false;
			UseVmUid = false;

			MaxItems = maxItems;
			Items = new VariableDefinition[maxItems];
		}	
        /*!
         * \brief Destructor
         * \details Clear all items and free assigned memmory
         */
        ~VariableGroup()
		{
			if (Items)
			{
				delete [] Items;
				Items = NULL;
			}
		}
};
/*!
 * \brief Data file information structure
 * \details Data file information structure
 */
struct DataTransferProperties
{
	public:
        /*!
         * \brief Size
         * \details File size
         */
		uint32b Size;
        /*!
         * \brief Timestamp
         * \details File timespan
         */
        uint64b Timestamp;
        /*!
         * \brief Crc
         * \details File CRC
         */
        uint16b Crc;
        /*!
         * \brief CalculatedCrc
         * \details Partial calculated data CRC
         */
        uint16b CalculatedCrc;
        /*!
         * \brief LastChunkSize
         * \details Last data received length
         */
        uint16b LastChunkSize;
        /*!
         * \brief FileOffset
         * \details Current file OffsetOutOfRange
         * \todo Co s tím?
         */
        uint32b FileOffset;

        /*!
         * \brief Fill
         * \param[in] size File size
         * \param[in] timestamp File timespan
         * \param[in] crc File CRC
         * \details Set defined values in object.
         */
		void Fill(uint32b size, uint64b timestamp, uint16b crc)
		{
			Size = size;
			Timestamp = timestamp;
			Crc = crc;
			CalculatedCrc = 0xffff;
			LastChunkSize = 0;
		}
};

//Udrzovat to tak aby mensi prava meli mensi ciselnou hodnotu
/*!
 * \brief Users Connection rights
 * \details Logged user rights
 */
enum ConnectionRights
{
    /*!
     * \brief 0x00
     * \details Unspecified rights
     */
	Nothing = 0x00, // None nemuze byt kvuli kolizi jineho enumu
    /*!
     * \brief 0x10
     * \details User has read only rights
     */
    ReadOnly = 0x10,
    /*!
     * \brief 0x80
     * \details User has full access
     */
    FullControl = 0x80,
    /*!
     * \brief 0xFF
     * \details User is detected as engeneer
     */
    Engineering = 0xFF
};

/*!
 * \brief Namespace for VarAccessFlags enumeration
 */
namespace VarAccessFlags
{
    /*!
     * \brief Communications flags
     * \details Communications flags
     */
	enum VarAccessFlags
	{
        /*!
         * \brief 0x80
         * \details Use Offset & Length
         */
        UseOffsetLength = 0x80,
        /*!
         * \brief 0x40
         * \details Use VM UID
         */
        UseVmUid = 0x40,
        /*!
         * \brief 0x20
         * \details File mode - read variable values from /var/direct file
         */
        WriteLargeData = 0x20,
        /*!
         * \brief 0x10
         * \details
         * \todo K čemu to je??
         */
        TaskIdPresent = 0x10,

        /*!
         * \brief 0
         * \details Get all values
         */
		GetAllValues = 0,
        /*!
         * \brief 1
         * \details Get all values with metadata - __NOT IMPLEMENTED__
         */
        GetAllValuesWithMeta = 1,	// not implemented
        /*!
         * \brief 2
         * \details Get changed values
         */
        GetChangedValues = 2,
        /*!
         * \brief 3
         * \details Get changed values with metadata - __NOT IMPLEMENTED__
         */
        GetChangedValuesWithMeta = 3, // not implemented
        /*!
         * \brief 4
         * \details Get compact values - __NOT IMPLEMENTED__
         */
        GetCompactValues = 4, // not implemented
        /*!
         * \brief Mask
         * \details Mask for all values for reading flags
         */
        GetFormatMask = 0x7,

        /*!
         * \brief 0
         * \details Set all values
         */
        SetAllValues = 0,
        /*!
         * \brief 1
         * \details Set selected values
         */
        SetSelectedValues = 1,
        /*!
         * \brief 2
         * \details Set compact values
         */
        SetCompactValues = 2,
        /*!
         * \brief Mask
         * \details Mask for all values for writing flags
         */
        SetFormatMask = 3,

	};
}

#if defined( __FREERTOS__ )
#pragma pack(1)
#else
#pragma pack(push, 1)
#endif
/*!
 * \brief SSCP protocol function header
 * \details Header for all functions sended by SSCP protocol
 */
struct SSCPProtocolHeader
{
    /*!
     * \brief FunctionID
     * \details Specified function code
     * \sa SscpFunction
     */
	uint16b m_FunctionId;
    /*!
     * \brief Data length
     * \details Function command length
     */
    uint16b m_DataLen;
};

//server responses
/*!
 * \brief Error respond
 * \details Base error respond message informations
 */
struct SSCPProtocolErrorResponse
{
    /*!
     * \brief Header
     * \details
     */
	SSCPProtocolHeader m_Header;

    /*!
     * \brief Error code
     * \details returned error code
     */
	uint32b m_ErrorCode;
};

/*!
 * \brief No such variable error response
 * \details No such variable error response
 * \todo Není to strohé?
 */
struct SSCPProtocolNoSuchVariableErrorResponse
{
    /*!
     * \brief Header
     * \details Error header
     */
	SSCPProtocolErrorResponse m_CommonErrorResponse;
    /*!
     * \brief UID
     * \details Varialble UID
     */
	uint32b m_Uid;
};

/*!
 * \brief System command error response
 * \details System command error response
 * \todo Není to strohé?
 */
struct SSCPProtocolSysCmdErrorResponse
{
    /*!
     * \brief Header
     * \details Error header
     */
    SSCPProtocolErrorResponse CommonErrorResponse;
    /*!
     * \brief Command
     * \details Command ID
     */
	SystemCommand Command;
};

/*!
 * \brief Login response message
 * \details Positive response to Login message
 * \todo Není to strohé?
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolLoginResponse
{
    /*!
     * \brief Header
     * \details Message header
     */
	SSCPProtocolHeader m_Header;

    /*!
     * \brief Protocol version
     * \details Protocol version
     */
	uint8b m_ProtocolVersion;
    /*!
     * \brief Maximum data size
     * \details Server indicates the maximal buffer size to the client. Client cannot create frames that are bigger than this.
     */
	uint16b m_MaxDataSize;
    /*!
     * \brief Users Rights
     * \details Get logged users rights
     * \sa ConnectionRights
     */
	uint8b m_Rights;
    /*!
     * \brief Image GUID
     * \details Image identification
     */
	uint8b m_ImageGuid[SEX_IMAGE_GUID_LEN];
};

/*!
 * \brief
 * \details
 * \todo Coto je?
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolSAckResponse
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Offset
     * \details Offset
     */
	uint32b m_Offset;
};

/*!
 * \brief Initiate data receive response
 * \details Initiate data receive response
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolInitiateDataReceiveResponse
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Total size
     * \details Size of the data.
     */
	uint32b	m_TotalSize;
    /*!
     * \brief Timestamp
     * \details DATE_TIME format
     */
	uint64b m_Timestamp;
    /*!
     * \brief CRC
     * \details CRC of the data
     */
	uint16b	m_Crc;
};

/*!
 * \brief Receive data chunk response
 * \todo SERVER!!!! - Odkomentovat
 * \details Receive data chunk response
 */
struct SSCPProtocolReceiveDataChunkResponse
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Offset
     * \details Offset to retrieve
     */
	uint32b m_Offset;
};

/*!
 * \brief Get PLC Statistics response
 * \todo SERVER!!!! - Odkomentovat
 * \details Get PLC Statistics response
 */
struct SSCPProtocolGetPlcStatisticsResponse
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Version
     * \details Following description describes the version 1
     */
	uint8b m_Version;
    /*!
     * \brief Normal task count
     * \details Corresponds with the max number of threads (i.e. tasks that run concurrently in normal operation).
     */
	uint8b m_NormalTaskCount;
    /*!
     * \brief Maximum task ID
     * \details Maximum task ID
     */
	uint8b m_MaxTaskId;
    /*!
     * \brief Evaluator state
     * \details Evaluator state
     * \sa EvaluatorState
     */
	uint8b m_EvalState;
    /*!
     * \brief Up time
     * \details Application running time (TIME)
     * \todo je to pravda, nebo je to time RT?
     */
	uint64b m_UpTime;
    /*!
     * \brief Running tasks mask
     * \details Bit mask of running task (bit position with 1 coresponds to task’s ID that runs)
     * \todo Naní to navázáno na nějaký enum?
     */
	uint64b m_RunningTasksMask;
    /*!
     * \brief Failed tasks mask
     * \details Bit mask of tasks that encountered an error.
     */
	uint64b m_FailedTasksMask;

	/* memory usage (in kB) */
    /*!
     * \brief Total heap
     * \details Total heap memory available for runtime. Memory usage (in kB)
     */
    uint16b TotalHeap;
    /*!
     * \brief Free heap before load
     * \details Free heap after runtime startup (memory used for runtime itself). Memory usage (in kB)
     */
    uint16b FreeHeapBeforeLoad;
    /*!
     * \brief Free heap
     * \details Free heap after image load and process. Memory usage (in kB)
     */
    uint16b FreeHeap;
    /*!
     * \brief Total code space
     * \details Total space available for image. Memory usage (in kB)
     */
    uint16b TotalCodeSpace;
    /*!
     * \brief Free code space
     * \details Free space after image save. Memory usage (in kB)
     */
    uint16b FreeCodeSpace;
    /*!
     * \brief Retain size
     * \details Retain area size. Memory usage (in kB)
     */
    uint16b RetainSize;

    /*!
     * \brief Section VMEX total
     * \details Heap memory consumed by VMEX section (VM image). Memory usage (in kB)
     */
    uint16b SectionVmexTotal;
    /*!
     * \brief Section RTCM total
     * \details Heap memory consumed by communication section. Memory usage (in kB)
     */
    uint16b SectionRtcmTotal;
    /*!
     * \brief Section other total
     * \details Heap memory consumed by other sections. Memory usage (in kB)
     */
    uint16b SectionOtherTotal;
};

/*!
 * \brief Time setup response
 * \details Time setup response
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolTimeSetupResponse
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader Header;

    /*!
     * \brief Time stamp
     * \details Actual PLC time in UTC format
     */
	uint64b TimeStamp;
};

// nepouzijeme primo TCallStack kvuli nedefinovanemu layoutu
/*!
 * \brief SSCP call stack
 * \details SSCP call stack data
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPCallStack
{
    /*!
     * \brief UID
     * \details Call stack UID
     */
	uint32b m_UID;
    /*!
     * \brief Offset
     * \details Call stack offset
     */
	uint32b m_Ofs;
};

// nepouzijeme primo TCallStack kvuli nedefinovanemu layoutu
// fixme: nijak se to zatim nepouziva
// chybi implementace prevraceni na big endian
/*!
 * \brief Stack element
 * \details SSCP call stack data
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPStackElem
{
    /*!
     * \brief Value
     * \details Value
     */
    union
	{
        /*!
         * \brief
         * \details
         */
        uint32b U32;
        /*!
         * \brief
         * \details
         */
        sint32b S32;
        /*!
         * \brief
         * \details
         */
        uint64b U64;
        /*!
         * \brief
         * \details
         */
        sint64b S64;
        /*!
         * \brief
         * \details
         */
        float32b F32;
        /*!
         * \brief
         * \details
         */
        float64b F64;

        /*!
         * \brief
         * \details
         */
        struct
		{
            /*!
             * \brief
             * \details
             */
            uint32b          aUID;
            /*!
             * \brief
             * \details
             */
            sint32b          aOfs;
		} Addr;
	} m_Data;

    /*!
     * \brief
     * \details
     */
    uint32b m_StType;
};

/*!
 * \brief Get stacks response
 * \details Debug info response
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolGetStacksResponse
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Version
     * \details Version
     */
	uint8b m_Version;
    /*!
     * \brief Exception error code
     * \details Valid only for exception snapshot.
     */
	uint32b m_ExceptionErrorCode; // valid only wen asking for exception stack snapshot
    /*!
     * \brief Stack type
     * \details Call stack type.
     * \sa StackType
     */
	uint8b m_StackType;
    /*!
     * \brief Task ID
     * \details Task number.
     */
	uint8b m_TaskId;
    /*!
     * \brief Offset
     * \details Offset from request).
     */
	uint16b m_Offset;
    /*!
     * \brief Count
     * \details Number of items.
     */
	uint8b m_Count;
    /*!
     * \brief Next offset
     * \details Next stack offset
     * \li \c 0xFFFF - No more data
     * \li \c 0x0000 - \c 0xFFFE - Next offset for another request
     */
	uint16b m_NextOffset;

	union
	{
        /*!
         * \brief Call stack
         */
		SSCPCallStack m_CallStack[1];
        /*!
         * \brief Comp stack
         */
		SSCPStackElem m_CompStack[1];
	};
};

/*!
 * \brief Get task statistics response
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolGetTaskStatisticsResponse
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Version
     * \details Following description describes the version 1
     */
	uint8b	m_Version;
    /*!
     * \brief Cycle count
     * \details Cyrcle count
     */
	uint64b m_CycleCount;
    /*!
     * \brief Last cycle duration
     * \details TIME
     */
	uint64b m_LastCycleDuration;
    /*!
     * \brief Average cycle duration
     * \details TIME
     */
    uint64b m_AverageCycleDuration;
    /*!
     * \brief Minimal cycle duration
     * \details TIME
     */
    uint64b m_MinimalCycleDuration;
    /*!
     * \brief Maximal cycle duration
     * \details TIME
     */
    uint64b m_MaximalCycleDuration;
};


//incomming requsts
/*!
 * \brief Login request message
 * \details Login request message
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolLoginRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader Header;

    /*!
     * \brief Requested version
     * \details Protocol version requested by client (Currently 2)
     */
	uint8b RequestedVersion;
    /*!
     * \brief Maximum data size
     * \details Client indicates the maximal buffer size to the server. Server cannot create frames that are bigger than this.
     */
	uint16b MaxDataSize;
    /*!
     * \brief User name length
     * \details Encoded user name is max 255 chars.
     */
	uint8b UserNameLen;
    /*!
     * \brief Password Length
     * \details Password is sended as MD5 hash - this value has to be 0x10 (16)
     */
	uint8b PassLen;				// !!! odsud uz neodpovida rozmisteni v telegramu
    /*!
     * \brief Proxy ID length
     * \details ProxyID Length from version 2 of the protocol. Max 255 chars.
     */
    uint8b ProxyIdLen;
    /*!
     * \brief User name
     * \details Short string UTF8 encoded.
     */
	sint8b *UserName;
    /*!
     * \brief Password
     * \details MD5 hash of password
     */
	sint8b *Pass;
    /*!
     * \brief Proxy ID
     * \details ProxyID String UTF8 Encoded from version 2 of the protocol.
     */
	sint8b *ProxyId;
};

/*!
 * \brief Initiate data send request
 * \details Initiate data send request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolInitiateDataSendRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Data ID length
     * \details File name length (Data identification == file name).
     */
	uint8b m_DataIdLen;	
    /*!
     * \brief File size
     * \details Size of the file data
     */
	uint32b m_DataSize;
    /*!
     * \brief Timestamp
     * \details Creation datetime (DATE_TIME format).
     */
    uint64b m_Timestamp;
    /*!
     * \brief Data ID
     * \details File name with path
     */
	uint8b *m_DataId;
};

/*!
 * \brief Send data chunk request
 * \todo SERVER!!!! - Odkomentovat
 * \details Send data chunk request
 */
struct SSCPProtocolSendDataChunkRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Data offset
     * \details The data that are being sent are from this offset
     */
	uint32b m_DataOffset;
};

/*!
 * \brief Finish send data request
 * \todo SERVER!!!! - Odkomentovat
 * \details Finish send data request
 */
struct SSCPProtocolFinishSendDataRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief CRC
     * \details CRC of the data
     */
	uint16b m_Crc;
};

/*!
 * \brief Initiate data receive request
 * \details Initiate data receive request
 * \todo incomplete - _Data identification - Encoded user name is max 128 chars_ missing.
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolInitiateDataReceiveRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Data ID length
     * \details File name length (Data identification == file name).
     */
    uint8b m_DataIdLen;
};

/*!
 * \brief The Receive data chunk request
 * \details The Receive data chunk request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolReceiveDataChunkRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Data offset
     * \details Offset to retrieve
     */
	uint32b m_DataOffset;
};

/*!
 * \brief Read variables directly request
 * \details Read variables directly request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolReadVariablesDirectlyRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Flags
     * \details One of VarAccessFlags::GetFormatMask flags. “Get Compact Values With Metadata” is not specified because the overhead of meta data is so big that it makes no sense to save individual bits.
     * \sa VarAccessFlags
     */
	uint8b m_Flags;
};

/*!
 * \brief Variable definition
 * \details Variable definition
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolVariableDefinition
{
    /*!
     * \brief UID
     * \details Variable UID
     */
	uint32b m_Uid;
    /*!
     * \brief Offset
     * \details Variable offset
     */
    uint32b m_Offset;
    /*!
     * \brief Length
     * \details Variable length
     */
    uint32b m_Length;
};

/*!
 * \brief Register variable group request
 * \details Register variable group request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolRegisterVariableGroupRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Group ID
     * \details Identification of a group.
     */
	uint8b m_GroupId;
    /*!
     * \brief Flags
     * \details Variabls flags VarAccessFlags::UseOffsetLength or VarAccessFlags::UseVmUid
     * \sa VarAccessFlags::UseOffsetLength, VarAccessFlags::UseVmUid
     */
	uint8b m_Flags;
};

/*!
 * \brief Read variable group directly request
 * \details Read variable group directly request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolReadVariableGroupDirectlyRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Group ID
     * \details Identification of a group.
     */
	uint8b m_GroupId;
    /*!
     * \brief Flags
     * \details One of VarAccessFlags::GetFormatMask flags.
     * \sa VarAccessFlags
     */
	uint8b m_Flags;
};

/*!
 * \brief Write variables directly request
 * \details Write variables directly request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolWriteVariablesDirectlyRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Flags
     * \details Use VarAccessFlags::UseOffsetLength, VarAccessFlags::UseVmUid or VarAccessFlags::WriteLargeData
     * \sa VarAccessFlags
     */
	uint8b m_Flags;
};

/*!
 * \brief Write variable group directly request
 * \details Write variable group directly request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolWriteVariableGroupDirectlyRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Group ID
     * \details Identification of a group.
     */
	uint8b m_GroupId;
    /*!
     * \brief Flags
     * \details One of VarAccessFlags::SetFormatMask
     */
	uint8b m_Flags;
};

/*!
 * \brief Read variable structure request
 * \details Read variable structure request
 * \obsolete
 * \todo incomplete - Not use
 */
struct SSCPProtocolReadVariableStructureRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Flags
     * \details Enable/Disable VarAccessFlags::UseVmUid flag
     * \sa VarAccessFlags::UseVmUid
     */
	uint8b m_Flags;
};

/*!
 * \brief Get plc statistics request
 * \details Get plc statistics request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolGetPlcStatisticsRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;
};

/*!
 * \brief Get task statistics request
 * \details Get task statistics request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolGetTaskStatisticsRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Task ID
     * \details Task ID
     */
	uint8b	m_TaskNumber;
};

/*!
 * \brief Accept new image system command parameters
 * \details Parameter for Accept new image system command
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolAcceptNewImageSystemCommandParameters
{
	//bool m_TryOnlineChanges;	
	//bool m_ForceClearRetain;
    /*!
     * \brief Flag
     * \details Reinitialize with actual image in the memory.
     * \li \c D0 - Online changes
     * \li \c D1 - Retain Clear
     */
	uint8b m_Flags_OC_CR; // Online changes, RetainClear
};

/*!
 * \brief Create boot project system command parameters
 * \details Parameter for Create boot project system command
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolCreateBootProjectSystemCommandParameters
{
    /*!
     * \brief Image area number
     * \details Image area number
     */
	uint8b m_ImageAreaNr;
};

/*!
 * \brief Set active partition system command parameters
 * \details Parameter for Set active partition system command
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolSetActivePartitionSystemCommandParameters
{
    /*!
     * \brief Image area number
     * \details Image area number
     */
    uint8b  m_ImageAreaNr;
};

/*!
 * \brief Set MAC address system command parameters
 * \details Parameter for Set MAC address system command
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolSetMacAddressSystemCommandParameters
{
    /*!
     * \brief MAC address
     * \details New MAC address
     */
	uint64b  m_MacAddress;
};

/*!
 * \brief Start PLC system command parameters
 * \details Parameter for Start PLC system system command
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolStartPlcSystemCommandParameters
{
    /*!
     * \brief Start PLC run mode
     * \details One of RunMode
     * \sa RunMode
     */
	uint8b StartPlcRunMode;
};

/*!
 * \brief Format filesystem command parameters
 * \details Parameter for Format filesystem system command
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolFormatFsSysCmdParameters
{
    /*!
     * \brief File system id
     * \details File system ID
     *
     * - \c System = 0
     * - \c User = 1
     */
    uint8b FileSystemId;
};

/*!
 * \brief System command request
 * \details System command request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolSystemCommandRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Command
     * \details Command ID
     * \sa SystemCommand
     */
    uint8b	m_Command;

    /*!
     * \brief Function parameter
     * \details Parameters for selected commad
     */
	union
	{
		SSCPProtocolAcceptNewImageSystemCommandParameters m_AcceptNewImage;
		SSCPProtocolCreateBootProjectSystemCommandParameters m_CreateBootProject;
		SSCPProtocolSetActivePartitionSystemCommandParameters m_SetActivePartition;
		SSCPProtocolSetMacAddressSystemCommandParameters m_SetMacAddress;
		SSCPProtocolStartPlcSystemCommandParameters m_StartPlc;
		SSCPProtocolFormatFsSysCmdParameters m_FormatFs;
	} m_Parameters;
};

/*!
 * \brief Time setup request
 * \details Time setup request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolTimeSetupRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader Header;
    /*!
     * \brief Time stamp
     * \details Optional 64bit C# timestamp in UTC format for time preset
     */
	uint64b TimeStamp;
};

#define SSCP_DRIVERCOMMAND_MAXLENGTH	128
/*!
 * \brief Driver command request
 * \details Driver command request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolDriverCommandRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader Header;
    /*!
     * \brief Channel ID
     * \details Channel ID constructed as 32 bit FNV1 hash from channel name
     */
	uint32b ChannelId;
    /*!
     * \brief Data
     * \details Sending data
     */
	uint8b Data[SSCP_DRIVERCOMMAND_MAXLENGTH];
    /*!
     * \brief Length
     * \details Data length
     */
	uint8b Length;
};

/*!
 * \brief Get stacks request
 * \details Get stacks request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolGetStacksRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Task number
     * \details Task number
     * \li \c 0xFF - exception snapshot
     * \li \c 0x00 - \c 0x20 - task thread
     */
	uint8b m_TaskId;
    /*!
     * \brief Stack type
     * \details Stack type
     * \li \c 0x00 call stack
     * \li \c 0x01 computational stack
     */
	uint8b m_StackType;
    /*!
     * \brief Offset
     * \details Offset of the top.
     */
	uint16b m_Offset;
};

/*!
 * \brief Set reset read only mode request
 * \todo SERVER!!!! - Odkomentovat
 */
struct SSCPProtocolSetResetReadOnlyModeRequest
{
    /*!
    * \brief Header
    * \details Message header
    */
    SSCPProtocolHeader m_Header;

    /*!
     * \brief Params
     * \details
     * \li \c B7 \c (0x80) 0 - reset, 1 set to read only
     * \li \c B6 \c (0x40) 1 - reset all to default (B7 is ignored)
     */
	uint8b m_Params;
    /*!
     * \brief Comm UID
     * \details UID repeats “var count” times
     */
	uint32b m_CommUid[1];
};

#if defined( __FREERTOS__ )
// fixme: Nastavi to pragma pack jako na zacatku kompilace, pokud by nekdo nastavoval pack drive tak to tento radek zrusi.
// fixme: Nemelo by se to ale stat, kazdy by mel po vlastnim nastaveni pragma pack vse zase vratit na puvodni hodnotu
#pragma pack()
#else
#pragma pack(pop)
#endif

//#define SSCP_SUCCESS_MASK												0x8000
//#define SSCP_ERROR_MASK													0xC000

#define SSCP_STATUS_SUCCESS	0x8000
#define SSCP_STATUS_ERROR	0xC000
#define SSCP_STATUS_MASK	0xC000

#endif
