#include "./ModbusCrc.h"
#if !defined(SSCP_STANDALONE_CLIENT)
#include "./../../Shared/Utils/SystemLogger.h"
#endif

// SharkEvaluator.h je tu kvuli StacksSnapshot
// v zadnem pripade ho nepouzivat primo, ale pouze pres IEngineAccess
#if defined(SSCP_ENABLE_SERVER)
#include "./../Engine/SharkEvaluator.h"
#endif

#if defined(_MSC_VER) || defined(__BORLANDC__)
#include <string.h>
#endif

#if defined(__POSIX__)
#if defined(__PIKEOS__)
extern "C" 
{
#include <vm.h>
}
#else
#include <unistd.h>
#if !defined(__OS_X__)
#include <sys/reboot.h>
#endif
#endif
#endif

#include "./SscpProtocolParser.h"


#if defined(SSCP_ENABLE_SERVER)
SscpProtocolParser::ContextSpecificData::ContextSpecificData(uint8b groupsCount, uint8b groupVariableCount)
{
	LoggedIn = false;
	ClientMaxDataSize = 0;
	Rights = ReadOnly;

	linkHeaderLength = 1;
	enableCrcCheck = false;
	allowOnlyBasicFunctions = false;

	// fixme: tady pak bude orisek s ostatnimi platformami - hlavne asi ARMem
	// dovedu si predstavit, ze ARM dovoli pouze jeden prenos (nebo nejak omezeny)
	// pak bude implementace binarniho prenosu slozitejsi a i treba nasdilene buffery apod. budemem muset resit
	// zatim vsak neresme to, co nas nepali...
	m_LargeDataTransfer = new LargeDataTransfer(NULL); // 0.1 MB

	VariableGroupCount = groupsCount;
	VariableGroups = new VariableGroup*[VariableGroupCount];
	for (uint8b i = 0; i < VariableGroupCount; i++)
		VariableGroups[i] = new VariableGroup(groupVariableCount);
	
	m_DataSendAllDataReceived = false;
}

//-------------------------------------------------------------------------------------------------

SscpProtocolParser::ContextSpecificData::~ContextSpecificData(void)
{
	if (VariableGroups)
	{
		for (uint8b i = 0; i < VariableGroupCount; i++)
		{
			if (VariableGroups[i])
			{
				delete (VariableGroups[i]);
				VariableGroups[i] = NULL;
			}
		}
		delete [] VariableGroups;
		VariableGroups = NULL;
	}
	
	delete m_LargeDataTransfer;
}

//-------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------

void SscpProtocolParser::GetBufferRequirements(uint32b & readBuffer, uint32b & writeBuffer, const IMessageServiceImplementation::ClientContext *clientContext,
	uint32b requestedSize)
{
	// pracujeme s SSCP_MAX_BUFFER_SIZE byte dlouhym bufferem pro prenosy dat
	ContextSpecificData *contextSpecificData = (ContextSpecificData *)clientContext->m_ContextSpecificData;
	
	if (requestedSize == 0)
		requestedSize = SSCP_MAX_DATA_SIZE;

	contextSpecificData->AllocatedBufferSize = requestedSize;

	//Logger.LogInfo("xxx", "Allocated context %d bytes", requestedSize);

	readBuffer = requestedSize + contextSpecificData->linkHeaderLength + sizeof(SSCPProtocolHeader) + 
		(contextSpecificData->enableCrcCheck ? sizeof(uint16b) : 0);
	writeBuffer = requestedSize + contextSpecificData->linkHeaderLength + sizeof(SSCPProtocolHeader) +
		(contextSpecificData->enableCrcCheck ? sizeof(uint16b) : 0);
}

//-------------------------------------------------------------------------------------------------

IMessageServiceImplementation::ClientContext * SscpProtocolParser::CreateContextData(void) const
{
	IMessageServiceImplementation::ClientContext *r = IMessageServiceImplementation::CreateContextData();
	r->m_ContextSpecificData = new ContextSpecificData(m_VariableGroupCount, m_VariableGroupSize);
	return r;
}

//-------------------------------------------------------------------------------------------------

void SscpProtocolParser::DeleteContextData(IMessageServiceImplementation::ClientContext * contextData) const
{
	delete (ContextSpecificData *)(contextData->m_ContextSpecificData);
	delete contextData;
}

//-------------------------------------------------------------------------------------------------

void SscpProtocolParser::ConnectionCreated(IMessageServiceImplementation::ClientContext * contextData)
{
	ContextSpecificData * data = (ContextSpecificData *)contextData->m_ContextSpecificData;
	data->LoggedIn = false;

	// zrusime registraci vsech drive zaregistrovanych skupin
	// fixme: a co kdyz bude nekvalitni spojeni - nejaky keep alive - nj, ale pak by musela byt vazba mezi contexty
	for (uint32b i = 0; i < m_VariableGroupCount; i++)
		data->VariableGroups[i]->Registered = false;
}

//-------------------------------------------------------------------------------------------------

void SscpProtocolParser::ConnectionClosed(IMessageServiceImplementation::ClientContext * contextData)
{
	ContextSpecificData * data = (ContextSpecificData *) contextData->m_ContextSpecificData;
	data->m_LargeDataTransfer->Cleanup();
	data->LoggedIn = false;

	for (uint32b i = 0; i < data->VariableGroupCount; i++)
	{
		VariableGroup *group = data->VariableGroups[i];
		for (uint32b j = 0; j < group->Count; j++)
		{
			VariableDefinition *item = &(data->VariableGroups[i]->Items[j]);
			m_EngineAccess->EnableCOVReporting(item->Uid, group->UseVmUid, false, this);
		}
	}
}

//-------------------------------------------------------------------------------------------------

IMessageServiceImplementation::Action SscpProtocolParser::DataRead(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData)
{
	ContextSpecificData *contextSpecificData = (ContextSpecificData *)contextData->m_ContextSpecificData;
	uint32b headerLen = contextSpecificData->linkHeaderLength + sizeof(SSCPProtocolHeader);
	uint32b footerLen = contextSpecificData->enableCrcCheck ? sizeof(uint16b) : 0;

	readBuffer.DataOffset = 0;
	writeBuffer.DataOffset = 0;

	if (readBuffer.DataLen >= (headerLen + footerLen))
	{
		IMessageServiceImplementation::Action nextAction = IMessageServiceImplementation::ReadData;
		
		SSCPProtocolHeader *msgHeader = (SSCPProtocolHeader *)(readBuffer.Buffer + contextSpecificData->linkHeaderLength);

		// uz vime, kolik mame mit dat
		uint16b frameLen = GetUint16FromBuffer((uint8b *)&msgHeader->m_DataLen, 0) + headerLen + footerLen;

		// prijali jsme vsechna data
		if (readBuffer.DataLen >= frameLen)
		{
			ContextSpecificData * ctxSpecificData = (ContextSpecificData *) contextData->m_ContextSpecificData;

			SSCPProtocolHeader header;
			if (CheckFrameAndSlaveAddress(readBuffer, ctxSpecificData->enableCrcCheck, m_LocalSlaveAddress, 
				ctxSpecificData->linkHeaderLength, &header))
			{
				// zareaguj na fci a posli reakci
				writeBuffer.DataOffset = readBuffer.DataOffset;

				ctxSpecificData->header = &header;

				nextAction = Sscp_Indication(readBuffer, writeBuffer, contextData);

				readBuffer.Clear();

				if (nextAction == IMessageServiceImplementation::WriteData)
				{
					if (ctxSpecificData->linkHeaderLength > 0)
						writeBuffer[0] = m_LocalSlaveAddress;

					if (ctxSpecificData->enableCrcCheck)
					{
						uint16b packetCrc = ModbusCrc::CRC16(writeBuffer.Buffer, writeBuffer.DataLen);
						SetUint16ToBuffer(writeBuffer.Buffer, writeBuffer.DataLen, packetCrc);

						writeBuffer.DataLen += footerLen;
					}
				}
			}
			else
			{
				writeBuffer.Clear();
			}
		}

		return nextAction;
	}

	return IMessageServiceImplementation::ReadData;
}

//-------------------------------------------------------------------------------------------------

IMessageServiceImplementation::Action SscpProtocolParser::Continue(CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData)
{
	return (Disconnect);
}
#endif /* SSCP_ENABLE_SERVER */

//-------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------

#if defined(SSCP_ENABLE_SERVER)
IMessageServiceImplementation::Action SscpProtocolParser::Sscp_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData)
{
	IMessageServiceImplementation::Action nextAction = IMessageServiceImplementation::Disconnect;
	ContextSpecificData *contextSpecificData = (ContextSpecificData *)contextData->m_ContextSpecificData;
	SSCPProtocolHeader *hdr = (SSCPProtocolHeader *)(readBuffer.Buffer + readBuffer.DataOffset);	
	SscpFunction::SscpFunction function = (SscpFunction::SscpFunction)GetUint16FromBuffer((uint8b *)&hdr->m_FunctionId, 0);

	if ((function != SscpFunction::ReceiveDataChunk) && (function != SscpFunction::SendDataChunk) && (function != SscpFunction::FinishDataSend) &&
		(function != SscpFunction::WriteVariableDirectly) && (function != SscpFunction::WriteVariableGroupDirectly) &&
		!(contextSpecificData->m_LargeDataTransfer->UsingExistingFileData() && (function == SscpFunction::InitiateDataReceive)))
	{		
		//dame prilezitost LargeDataTransferu po sobe uklidit pokud predchozi prikaz byl SSCP_FUNC_INITIATE_DATA_SEND nebo SSCP_FUNC_INITIATE_DATA_RECEIVE
		contextSpecificData->m_LargeDataTransfer->Cleanup();		
	}

	// HW konfigurace nevyzaduje prihlaseni
	if ((function == SscpFunction::GetBasicInfo) || (function == SscpFunction::SetBasicInfo))
	{
		IMemoryAllocator *ma = contextSpecificData->m_LargeDataTransfer->GetDataAllocator();
		if (function == SscpFunction::GetBasicInfo)
			nextAction = GetBasicInfo_Indication(readBuffer, writeBuffer);
		if (function == SscpFunction::SetBasicInfo)
			nextAction = SetBasicInfo_Indication(readBuffer, writeBuffer, contextSpecificData->AllocationId, contextSpecificData->LastSegmentId);
		
		return(nextAction);
	}

	// pokud volame nejakou fci a jeste nejsme prihlaseni nebo jsou povolene pouze zakladni funkce, tak se odlogujeme/odpojime
	if ((function != SscpFunction::Login && !contextSpecificData->LoggedIn) || (contextSpecificData->allowOnlyBasicFunctions))
	{
		//printf("Not logged %x\n", function);
		ComposeErrorResponse(function, SscpError::NotLogged, writeBuffer);
		nextAction = IMessageServiceImplementation::WriteData;
		return(nextAction);
	}

	switch (function)
	{
		case SscpFunction::Login: nextAction = Login_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::InitiateDataSend: nextAction = InitiateSendData_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::SendDataChunk: nextAction = SendDataChunk_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::FinishDataSend: nextAction = FinishSendData_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::InitiateDataReceive: nextAction = InitiateReceiveData_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::ReceiveDataChunk: nextAction = ReceiveDataChunk_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::ReadVariablesDirectly: nextAction = ReadVariablesDirectly_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::RegisterVariableGroup: nextAction = RegisterVariableGroup_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::ReadVariableGroupDirectly: nextAction = ReadVariableGroupDirectly_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::WriteVariableDirectly: nextAction = WriteVariableDirectly_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::WriteVariableGroupDirectly: nextAction = WriteVariableGroupDirectly_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::ReadVariableStructure: nextAction = ReadVariableStructure_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::GetPlcStatistics: nextAction = GetPlcStatistics_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::GetTaskStatistics: nextAction = GetTaskStatistics_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::ExecuteSystemCommand: nextAction = ExecuteSystemCommand_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::TimeSetup: nextAction = TimeSetup_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::DriverCommand: nextAction = DriverCommand_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::GetStacks: nextAction = GetStacks_Indication(readBuffer, writeBuffer, contextData); break;
		case SscpFunction::SetResetReadOnlyMode: nextAction = SetResetReadOnlyMode_Indication(readBuffer, writeBuffer, contextData); break;

		case SscpFunction::Logout:				
			nextAction = IMessageServiceImplementation::Disconnect;
			break;

		default:
			// invalid function
			Logger.LogError(FAC_SSCPSVR, "Invalid function %d/%04x", (int)function, (int)function);
			ComposeHeader(SscpError::UnknownFunction, 0, writeBuffer);	
			nextAction = IMessageServiceImplementation::WriteData;
			readBuffer.Clear(); //pravdepodobne chyba pri komunikaci, vymazeme buffer a zacne se znova nacisto
			break;
	}

	if ((function == SscpFunction::WriteVariableDirectly) || (function == SscpFunction::WriteVariableGroupDirectly))
	{		
		//dame prilezitost LargeDataTransferu po sobe uklidit pokud predchozi prikaz byl SSCP_FUNC_INITIATE_DATA_SEND nebo SSCP_FUNC_INITIATE_DATA_RECEIVE
		contextSpecificData->m_LargeDataTransfer->Cleanup();		
	}

	return nextAction;
}
#endif

//-------------------------------------------------------------------------------------------------

#if defined(SSCP_ENABLE_SERVER)
SscpProtocolParser::SscpProtocolParser(IEngineAccess * engineAccess, ILargeDataAccess * largeDataAccess, IMemoryAllocator *fileTransferAllocator, 
	IMemoryAllocator *memoryTransferAllocator)
{
	m_EngineAccess = engineAccess;
	m_LargeDataAccess = largeDataAccess;

	m_FileTransferMemoryAllocator = fileTransferAllocator;	
	m_MemoryTransferMemoryAllocator = memoryTransferAllocator;	
}

//-------------------------------------------------------------------------------------------------

void SscpProtocolParser::SetParameters(IAccessValidator *accessValidator, uint8b localSlaveAddress, uint8b variableGroupCount, uint8b variableGroupSize)
{
	m_AccessValidator = accessValidator;
	m_LocalSlaveAddress = localSlaveAddress;
	m_VariableGroupCount = variableGroupCount;
	m_VariableGroupSize = variableGroupSize;
}
#endif

//-------------------------------------------------------------------------------------------------

#if defined(SSCP_ENABLE_CLIENT)
	#if !defined(SSCP_STANDALONE_CLIENT)
SscpProtocolParser::SscpProtocolParser(IMemoryAllocator *memoryTransferAllocator)
	#else
SscpProtocolParser::SscpProtocolParser(void)
	#endif
{
#if defined(SSCP_ENABLE_SERVER)
	m_EngineAccess = NULL;
	m_LargeDataAccess = NULL;

	m_FileTransferMemoryAllocator = NULL;	
#endif

#if !defined(SSCP_STANDALONE_CLIENT)
	m_MemoryTransferMemoryAllocator = memoryTransferAllocator;	

	m_OneReadCyclePassed = false;
	m_EnableWritingCycles = false;
#endif
}
#endif

//-------------------------------------------------------------------------------------------------

SscpProtocolParser::~SscpProtocolParser(void)
{
}

//-------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------

// Vraci offset v writeBuffer, kde zacina vlastni SSCP hlavicka
uint32b SscpProtocolParser::ComposeHeader(uint16b functionId, uint16b dataLen, CommBuffer & writeBuffer)
{	
	sint32b iOffset = writeBuffer.DataOffset;

	SSCPProtocolHeader *responseHeader = (SSCPProtocolHeader *)(writeBuffer.Buffer + iOffset);	
	CommBuffer::SetUint16ToBuffer((uint8b *)&responseHeader->m_FunctionId, 0, functionId); 
	CommBuffer::SetUint16ToBuffer((uint8b *)&responseHeader->m_DataLen, 0, dataLen);
	
	writeBuffer.DataLen = iOffset + sizeof(SSCPProtocolHeader) + dataLen;
	writeBuffer.Offset = 0;

	return(iOffset);
}

//-------------------------------------------------------------------------------------------------

void SscpProtocolParser::ComposeErrorResponse(uint16b functionId, uint32b errorCode, CommBuffer & writeBuffer)
{
#if 0
	int iOffset = ComposeHeader(functionId | SSCP_STATUS_ERROR, sizeof(SSCPProtocolErrorResponse) - sizeof(SSCPProtocolHeader), writeBuffer);

	SSCPProtocolErrorResponse *response = (SSCPProtocolErrorResponse *)(writeBuffer.Buffer + iOffset);
	CommBuffer::SetUint32ToBuffer((uint8b *)&response->m_ErrorCode, 0, errorCode);
#endif
	ComposeErrorResponse(functionId, errorCode, writeBuffer, NULL, 0);
}

void SscpProtocolParser::ComposeErrorResponse(uint16b functionId, uint32b errorCode, CommBuffer & writeBuffer, const void *optionalData, uint32b optionalDataLength)
{
	int iOffset = ComposeHeader(functionId | SSCP_STATUS_ERROR, sizeof(SSCPProtocolErrorResponse) - sizeof(SSCPProtocolHeader) + optionalDataLength, writeBuffer);

	SSCPProtocolErrorResponse *response = (SSCPProtocolErrorResponse *)(writeBuffer.Buffer + iOffset);
	CommBuffer::SetUint32ToBuffer((uint8b *)&response->m_ErrorCode, 0, errorCode);

	if (optionalData && optionalDataLength)
		memcpy(writeBuffer.Buffer + writeBuffer.DataLen - optionalDataLength, optionalData, optionalDataLength);
}

//-------------------------------------------------------------------------------------------------

void SscpProtocolParser::ComposeSysCmdErrorResponse(uint16b functionId, uint32b errorCode, SystemCommand cmd, CommBuffer & writeBuffer)
{
	int iOffset = ComposeHeader(functionId | SSCP_STATUS_ERROR, sizeof(SSCPProtocolSysCmdErrorResponse) - sizeof(SSCPProtocolHeader), writeBuffer);

	SSCPProtocolSysCmdErrorResponse *response = (SSCPProtocolSysCmdErrorResponse *)(writeBuffer.Buffer + iOffset);

	CommBuffer::SetUint32ToBuffer((uint8b *)&response->CommonErrorResponse.m_ErrorCode, 0, errorCode);
	response->Command = cmd;
}

//-------------------------------------------------------------------------------------------------

void SscpProtocolParser::ComposeNoSuchVariableErrorResponse(uint16b functionId, uint32b errorCode, uint32b variableUid, CommBuffer & writeBuffer)
{
	int iOffset = ComposeHeader(functionId | SSCP_STATUS_ERROR, sizeof(SSCPProtocolNoSuchVariableErrorResponse) - sizeof(SSCPProtocolHeader), writeBuffer);

	SSCPProtocolNoSuchVariableErrorResponse *response = (SSCPProtocolNoSuchVariableErrorResponse *)(writeBuffer.Buffer + iOffset);

	CommBuffer::SetUint32ToBuffer((uint8b *)&response->m_CommonErrorResponse.m_ErrorCode, 0, errorCode);
	CommBuffer::SetUint32ToBuffer((uint8b *)&response->m_Uid, 0, variableUid);	
}

//-------------------------------------------------------------------------------------------------

bool SscpProtocolParser::CheckRightsComposeErrorResponse(ConnectionRights currentRights, ConnectionRights requieredRights, CommBuffer & writeBuffer)
{	
	if (currentRights < requieredRights)
	{
		ComposeHeader(SscpError::InsufficientRights, 0, writeBuffer);	
		
		return false;
	}

	return true;
}

//-------------------------------------------------------------------------------------------------

#if defined(SSCP_ENABLE_SERVER)
bool SscpProtocolParser::CheckDataLenComposeErrorResponse(uint32b dataLen, uint32b requieredDataLen,  CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData)
{
	uint32b maxReadLen, maxWriteLen;
	ContextSpecificData *contextSpecificData = (ContextSpecificData *)contextData->m_ContextSpecificData;
	GetBufferRequirements(maxReadLen, maxWriteLen, contextData, contextSpecificData->AllocatedBufferSize);	

	if ((dataLen != requieredDataLen) || (dataLen > (maxReadLen - sizeof(SSCPProtocolHeader))))
	{
		// fixme: odpovidame invalid function, ale asi by se melo odpovidat invalid parameters
		ComposeHeader(SscpError::UnknownFunction, 0, writeBuffer);	
		
		return false;
	}

	return true;
}

//-------------------------------------------------------------------------------------------------

bool SscpProtocolParser::HelperCheckLoginInfoFromBuffer(CommBuffer & readBuffer, uint32b *offset, uint32b *outRights)
{
	if ( (readBuffer.DataLen > 5) && (offset) )
	{
		sint8b cUsername[255] = "\0", cPassword[255] = "\0";
		uint32b iOffset = *offset;

		uint8b iUsernameLen = readBuffer[iOffset++];
		// fixme : michale co se asi tak stane, kdyz ti nahodou prijde iStringLen >= 255 ?
		// - nic, iStringLen je 8bit, takze to vetsi nez 255 byt nemuze
		if (iUsernameLen)
		{
			int iMaxLen = (iUsernameLen > sizeof(cUsername) - 1) ? sizeof(cUsername) - 1 : iUsernameLen; // kontrola na preteceni bufferu
			memcpy(cUsername, readBuffer.Buffer + iOffset, iMaxLen);
			cUsername[iMaxLen] = 0;
			iOffset += iUsernameLen;
		}
		uint8b iPasswordLen = readBuffer[iOffset++];
		if (iPasswordLen)
		{
			int iMaxLen = (iPasswordLen > sizeof(cPassword) - 1) ? sizeof(cPassword) - 1 : iPasswordLen; // kontrola na preteceni bufferu
			memcpy(cPassword, readBuffer.Buffer + iOffset, iMaxLen);
			cPassword[iMaxLen] = 0;
			iOffset += iPasswordLen;
		}

		if(offset)
			*offset = iOffset;

		return(m_AccessValidator && m_AccessValidator->CheckAccess((const sint8b*)cUsername, iUsernameLen, (const sint8b*)cPassword, iPasswordLen, outRights));
	}
	
	return (false);
}

#endif

//-------------------------------------------------------------------------------------------------

bool SscpProtocolParser::CheckFrameAndSlaveAddress(CommBuffer & buffer, bool enableCrc, uint8b slaveAddress, uint8b linkHeaderLength, SSCPProtocolHeader *header)
{	
	uint8b slaveAddressTlg = buffer[0];

	SSCPProtocolHeader *hdr = (SSCPProtocolHeader *)(buffer.Buffer + linkHeaderLength);	

	if (!header)
		return (false);

	header->m_FunctionId = CommBuffer::GetUint16FromBuffer((uint8b *)&hdr->m_FunctionId, 0);
	header->m_DataLen = CommBuffer::GetUint16FromBuffer((uint8b *)&hdr->m_DataLen, 0);

	if (buffer.BufferSize < header->m_DataLen)
	{
		//vadny paket, nebudeme na nej reagovat
		return (false);
	}

	if ((linkHeaderLength > 0) && (slaveAddress != slaveAddressTlg))
	{
		//paket urceny pro jine PLC, nebudeme na nej reagovat
		return (false);
	}

	// pripadna kontrola CRC
#if !defined(SSCP_STANDALONE_CLIENT)
	if (enableCrc)
	{
		uint16b crc;
		crc = CommBuffer::GetUint16FromBuffer(buffer.Buffer, buffer.DataOffset + sizeof(SSCPProtocolHeader) + header->m_DataLen);

		if (ModbusCrc::CRC16(buffer.Buffer, linkHeaderLength + sizeof(SSCPProtocolHeader) + header->m_DataLen) != crc)
		{
			//paket s vadnym kontrolnim souctem, nebudeme na nej reagovat
			return (false);
		}
	}
#endif

	buffer.DataOffset += linkHeaderLength;
	return (true);
}

//-------------------------------------------------------------------------------------------------
SscpError::SscpError SscpProtocolParser::CheckFunctionAndErrorStats(CommBuffer & buffer, SscpFunction::SscpFunction functionToCheck, uint32b *dataLength)
{
	return (CheckFunctionAndErrorStats(buffer, functionToCheck, dataLength, NULL));
}

SscpError::SscpError SscpProtocolParser::CheckFunctionAndErrorStats(CommBuffer & buffer, SscpFunction::SscpFunction functionToCheck, uint32b *dataLength, uint64b *optionalData)
{
	SscpFunction::SscpFunction function = (SscpFunction::SscpFunction)CommBuffer::GetUint16FromBuffer(buffer.Buffer, buffer.DataOffset);
	uint32b length = CommBuffer::GetUint16FromBuffer(buffer.Buffer, buffer.DataOffset + 2);
	if (dataLength)
		*dataLength = length;

	// pokud RT zadanou funkci neimplementuje
	if ((SscpError::SscpError)function == SscpError::UnknownFunction)
		return (SscpError::UnknownFunction);
	// kontrola opravneni
	if ((SscpError::SscpError)function == SscpError::InsufficientRights)
		return (SscpError::InsufficientRights);
	// kontrola, ze odpoved je na nasi funkci
	if (((uint16b)function & (~SSCP_STATUS_MASK)) != functionToCheck)
		return (SscpError::InvalidFunctionReceived);
	// kontrola, ze akce probehla uspesne
	if (((uint16b)function & SSCP_STATUS_MASK) == SSCP_STATUS_SUCCESS)
		return (SscpError::NoError);

	// pokud ne, tak je vracen jeji chybovy kod (min. 4B + volitelne podle chyby (ty uz si ale musi zpracovat uroven vyse))
	if (length >= 4)
	{
		uint32b errorCode = CommBuffer::GetUint32FromBuffer(buffer.Buffer, buffer.DataOffset + sizeof(SSCPProtocolHeader));

		if ((length >= 4) && (optionalData))
		{
			uint32b optionalDataLength = length - 4;
			if (optionalDataLength == 8)
			{
				// nasleduje jeden 64bit unsigned integer
				*optionalData = CommBuffer::GetUint64FromBuffer(buffer.Buffer, buffer.DataOffset + sizeof(SSCPProtocolHeader) + 4);
			}
		}

		return ((SscpError::SscpError)errorCode);
	}
	return (SscpError::Unknown);
}

//-------------------------------------------------------------------------------------------------

