#ifndef __SscpProtocolParser_h
#define __SscpProtocolParser_h

#include "./datatypes.h"
#include "./SscpProtocolCommon.h"

#if defined(SSCP_ENABLE_SERVER)
#include "./IMessageServiceImplementation.h"
#include "./../../RT/Engine/IEngineAccess.h"
#include "./ILargeDataAccess.h"
#include "./../../Shared/Utils/FragmentedMemoryAllocator.h"
#include "./LargeDataTransfer.h"
#include "./../../Shared/Utils/VmSpecificData.h"
#endif

#if defined(SSCP_ENABLE_CLIENT) && !defined(SSCP_STANDALONE_CLIENT)
#include "./../Communication/ICommunicationDriver.h"
#endif

#if !defined(SSCP_STANDALONE_CLIENT)
#include "./../../VM/csymboltable.h"
#else
#include "./NetworkingHelpers.h"
#include "./TcpClientEndpoint.h"
#endif

/*
 *    CLIENT    ||  SERVER
 * request
 *              ->
 *                 indication
 *                 response
 *              <-
 * confirmation
 */

/*!
 * \class SscpProtocolParser
 * \brief SSCP protocol parser
 * \details class for decode/encode communication function to SSCP messages (raw dat array)
 *
 * CLIENT    ||  SERVER
 * request
 *           ->
 *               indication
 *               response
 *            <-
 * confirmation
 *
 */

#if defined(SSCP_ENABLE_SERVER)
class SscpProtocolParser : public IMessageServiceImplementation, CVariableHook
#else
class SscpProtocolParser
#endif
{
	private:
#if defined(SSCP_ENABLE_SERVER)
		IEngineAccess *m_EngineAccess;
		ILargeDataAccess *m_LargeDataAccess;
		IMemoryAllocator *m_FileTransferMemoryAllocator;	
		IAccessValidator *m_AccessValidator;

		uint8b m_LocalSlaveAddress;
		uint8b m_VariableGroupCount;
		uint8b m_VariableGroupSize;
#endif
#if defined(SSCP_ENABLE_SERVER) || (defined(SSCP_ENABLE_CLIENT) && !defined(SSCP_STANDALONE_CLIENT))
		IMemoryAllocator * m_MemoryTransferMemoryAllocator;		
#endif

	#pragma region IMessageServiceImplementation - serverova cast
	public:
#if defined(SSCP_ENABLE_SERVER)
        /*!
         * \brief Context specific data
         * \details Used only with SSCP_ENABLE_SERVER
         * \todo doplnit popis
         */
		struct ContextSpecificData
		{
			uint16b AllocatedBufferSize;

			bool LoggedIn;
			uint16b ClientMaxDataSize;
			ConnectionRights Rights;

			LargeDataTransfer * m_LargeDataTransfer;
			uint8b VariableGroupCount;
			VariableGroup **VariableGroups;

			bool m_DataSendAllDataReceived;

			sint32b AllocationId;
			sint8b LastSegmentId;

			// zpracovani ramcu
			SSCPProtocolHeader *header; // ukazatel na zpracovanou hlavicku aktualniho ramce
			uint8b linkHeaderLength; // delka hlavicky linkove vrstvy pro dany kontext
			bool enableCrcCheck; // povoluje pouziti CRC v ramcich
			bool allowOnlyBasicFunctions; // povoluje pouze GetBasicInfo a SetBasicInfo

			ContextSpecificData(uint8b groupsCount, uint8b groupVariableCount);
			~ContextSpecificData(void);
		};		

		virtual void GetBufferRequirements(uint32b & readBuffer, uint32b & writeBuffer, const IMessageServiceImplementation::ClientContext *clientContext, 
			uint32b requestedSize = 0);
		virtual IMessageServiceImplementation::ClientContext * CreateContextData(void) const;
		virtual void DeleteContextData(IMessageServiceImplementation::ClientContext * contextData) const;
		virtual void ConnectionCreated(IMessageServiceImplementation::ClientContext * contextData);
		virtual void ConnectionClosed(IMessageServiceImplementation::ClientContext * contextData);
		virtual IMessageServiceImplementation::Action DataRead(CommBuffer & readBuffer, CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		virtual IMessageServiceImplementation::Action Continue(CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
#endif
	#pragma endregion

	#pragma region ICallbackCommunicationDriver - klientska cast
#if defined(SSCP_ENABLE_CLIENT) && !defined(SSCP_STANDALONE_CLIENT)
	private:
        /*!
         * \brief Driver state
         * \details Driver state. Used only with SSCP_ENABLE_CLIENT and not SSCP_STANDALONE_CLIENT definitions
         * \todo doplnit popis
         */
		enum DriverState 
		{ 
			// 0
			PrepareRequest, 
			CheckConfirmation,

			// 2
			RegisterGroup,
			RegisterGroupResponse,

			// 4
			ReadFileInitiate, 
			ReadFileInitiate2, 
			ReadFileRecvChunk, 
			ReadFileRecvChunk2,
			ReadFileComplete, 

			// 9
			WriteFileInitiate,
			WriteFileInitiate2,
			WriteFileSendChunk,
			WriteFileSendChunk2,
			WriteFileFinish,
			WriteFileFinish2,

			// 15
			WriteVariable,
			WriteVariable2,
			WriteVariableComplete,

			// 18
			ProtocolError, 
			Login, 
			LoginResponse 
		};
		enum CommCycle { ReadCycle, WriteCycle };

	public:
        /*!
         * \brief Client context specific data
         * \details Used only with SSCP_ENABLE_CLIENT and not SSCP_STANDALONE_CLIENT definitions
         * \todo Doplnit popis?
         */
		class ClientContextSpecificData : public ICallbackClientContext
		{
			public:
				DriverState State;
				CommCycle commCycle;
				CommBuffer Buffer;
				uint8b SlaveAddress;
				char *Username;
				char *Password;

				uint16b MaxServerDataSize;

				// promenne pro prenosy souboru
				DataTransferProperties DTP;
				sint32b AllocationId;
				char RequestFileName[32];

				SscpError::SscpError LastError;
				uint64b ErrorFlags;

				uint8b *ByteConversionBuffer;

				uint32b CurrentGroupIndex;
				uint32b GroupsCount;
				CommunicationVariableGroup ** VariableGroups;

				CriticalSection *DriverBufferAccessLock;
				CSymbolTable *SymbolTable;
		};

		virtual void Reset(ICallbackClientContext *context);
		virtual ICallbackCommunicationDriver::CallbackAction GetNextAction(ICallbackClientContext *context);
		virtual CommBuffer* GetCommunicationBuffer(ICallbackClientContext *context);

		virtual ICallbackClientContext* CreateClientContext(void);
		virtual void DeleteClientContext(ICallbackClientContext *context);
#endif /* SSCP_ENABLE_CLIENT && !SSCP_STANDALONE_CLIENT*/
	#pragma endregion

	#pragma region Funkce - requests
#if defined(SSCP_ENABLE_CLIENT)
        /*!
         * \brief Login request
         * \param[out] buffer Output data buffer
         * \param[in] username User name
         * \param[in] password MD5 hash of password
         * \param[in] proxyId Proxy ID
         * \details Create Login request message
         *
         * |                                       |     |                                                                                                                |
         * | ------------------------------------- | --- | -------------------------------------------------------------------------------------------------------------- |
         * | Function                              | 2B  | SscpFunction::Login                                                                                            |
         * | Data length                           | 2B  | -                                                                                                              |
         * | Requested version                     | 1B  | Protocol version requested by client (Currently 2)                                                             |
         * | [Client’s] Max data size in telegrams | 2B  | Client indicates the maximal buffer size to the server. Server cannot create frames that are bigger than this. |
         * | User name length                      | 1B  | Encoded user name is max 255 chars.                                                                            |
         * | User name                             | ?B  | Short string UTF8 encoded.                                                                                     |
         * | Password hash length                  | 1B  | 0x10                                                                                                           |
         * | Password hash                         | 16B | MD5                                                                                                            |
         * | ProxyID length                        | 1B  | ProxyID Length from version 2 of the protocol. Max 255 chars.                                                  |
         * | ProxyID                               | ?B  | ProxyID String UTF8 Encoded from version 2 of the protocol.                                                    |
         *
         */
		void Login_Request(CommBuffer & buffer, const char *username, const char *password, const char *proxyId);
        /*!
         * \brief Read variables directly request header
         * \param[out] buffer Output data buffer
         * \param[in] useOnlyCommUid Use only variables communication UID flag
         * \param[in] useVmUid Use VM UID flag
         * \param[in] varsCount Variables count
         * \details Create header for Read variables directly request. Variables are added via AppendVariable_Request function. This function is called \c varsCount times.
         *
         * |             |    |                                                                                                                                                                                                     |
         * | ----------- | -- | --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
         * | Function    | 2B | SscpFunction::ReadVariablesDirectly                                                                                                                                                                 |
         * | Data length | 2B | -                                                                                                                                                                                                   |
         * | Flags       | 1B | One of VarAccessFlags::GetFormatMask flags. _Note_: "Get Compact Values With Metadata" is not specified because the overhead of meta data is so big that it makes no sense to save individual bits. |
         *
         * \sa AppendVariable_Request, Login_Confirmation, ReadVariablesDirectlyVars_Confirmation
         */
		void ReadVariablesDirectlyHeader_Request(CommBuffer & buffer, bool useOnlyCommUid, bool useVmUid, uint32b varsCount);
        /*!
         * \brief Register group request header
         * \param[out] buffer Output data buffer
         * \param[in] groupId Group ID
         * \param[in] useOnlyCommUid Use only variables communication UID flag
         * \param[in] useVmUid Use VM UID flag
         * \param[in] varsCount Variables count
         * \details Create header for Register group request. Variables are added via AppendVariable_Request function. This function is called \c varsCount times.
         *
         * |             |    |                                                                 |
         * | ----------- | -- | --------------------------------------------------------------- |
         * | Function    | 2B | SscpFunction::RegisterVariableGroup                             |
         * | Data length | 2B | -                                                               |
         * | Flags       | 1B | VarAccessFlags::UseOffsetLength and or VarAccessFlags::UseVmUid |
         * | Group ID    | 1B | Identification of a group.                                      |
         *
         * \sa AppendVariable_Request
         */
        void RegisterGroupHeader_Request(CommBuffer & buffer, uint8b groupId, bool useOnlyCommUid, bool useVmUid, uint32b varsCount);
        /*!
         * \brief Append variable
         * \param[out] buffer Output data buffer
         * \param[in] commUid Variable ID
         * \param[in] offset Variable offset
         * \param[in] length Variable length
         * \details Append variable definition to request
         *
         * |        |    |                                                                                            |
         * | ------ | -- | ------------------------------------------------------------------------------------------ |
         * | UID    | 4B | Variable UID                                                                               |
         * | Offset | 4B | Offset is handy for accessing just a part of a struct/FB/array.                            |
         * | Length | 4B | Length of the data to copy as value (handy for accessing just a part of a struct/FB/array) |
         *
         * \sa RegisterGroupHeader_Request, ReadVariablesDirectlyHeader_Request
         */
		void AppendVariable_Request(CommBuffer & buffer, uint32b commUid, uint32b offset, uint32b length);
        /*!
         * \brief Append variable (only variable UID)
         * \param[out] buffer Output data buffer
         * \param[in] commUid Variable ID
         * \details Append variable definition to request. Variable is defined only with UID
         *
         * |        |    |              |
         * | ------ | -- | ------------ |
         * | UID    | 4B | Variable UID |
         *
         * \sa RegisterGroupHeader_Request, ReadVariablesDirectlyHeader_Request
         */
		void AppendVariable_Request(CommBuffer & buffer, uint32b commUid);
        /*!
         * \brief Read variable group directly request
         * \param[out] buffer Output data buffer
         * \param[in] groupId Group ID
         * \param[in] flags One of VarAccessFlags::GetFormatMask
         * \details Create Read variable group directly request
         *
         * |             |    |                                                                                                                                                         |
         * | ----------- | -- | ------------------------------------------------------------------------------------------------------------------------------------------------------- |
         * | Function    | 2B | SscpFunction::ReadVariableGroupDirectly                                                                                                                 |
         * | Data length | 2B | 0x00 0x02                                                                                                                                               |
         * | Group ID    | 1B | Identification of a group.                                                                                                                              |
         * | Flags       | 1B | _Note_: "Get Compact Values With Metadata" is not specified because the overhead of meta data is so big that it makes no sense to save individual bits. |
         *
         * \sa RegisterGroupHeader_Request, ReadVariablesDirectlyVars_Confirmation
         */
		void ReadVariableGroupDirectly_Request(CommBuffer & buffer, uint8b groupId, VarAccessFlags::VarAccessFlags flags);

        /*!
         * \brief Write variable directly request
         * \param[out] buffer Output data buffer
         * \param[in] useOnlyCommUid Sets VarAccessFlags::UseOffsetLength flag
         * \param[in] useVmUid Sets VarAccessFlags::UseVmUid flag
         * \param[in] flags Set other flags
         * \param[in] varCount Number of variables
         * \param[in] variablesLength Size of all variables
         * \param[out] totalLength Supposed request length
         * \todo Nejsou useOnlyCommUid a useVmUid zbytečné, když je použit parametr flags??
         * \details Create Write variable directly request header
         *
         * |                     |    |                                                                                                                                                   |
         * | ------------------- | -- | ------------------------------------------------------------------------------------------------------------------------------------------------- |
         * | Function            | 2B | SscpFunction::WriteVariableDirectly                                                                                                               |
         * | Data length         | 2B | -                                                                                                                                                 |
         * | Flags               | 1B | VarAccessFlags::UseOffsetLength, VarAccessFlags::UseVmUid and or VarAccessFlags::WriteLargeData                                                   |
         * | Number of variables | 1B | This field is only present if telegram contains variable values together with variable specification (VarAccessFlags::WriteLargeData not defined) |
         *
         * \sa WriteVariableDirectly_Confirmation
         */
		void WriteVariableDirectly_Request(CommBuffer & buffer, bool useOnlyCommUid, bool useVmUid, VarAccessFlags::VarAccessFlags flags, 
			uint32b varCount, uint32b variablesLength, uint32b *totalLength = NULL);
        /*!
         * \brief Write variable group directly request
         * \param[out] buffer Output data buffer
         * \param[in] groupId Group ID
         * \param[in] flags Flags
         * \param[in] variableDescCount Number of variables
         * \param[in] variablesLength Size of all variables
         * \details Create Write variable group directly request header
         *
         * |                     |    |                                                                                                                                          |
         * | ------------------- | -- | ---------------------------------------------------------------------------------------------------------------------------------------- |
         * | Function            | 2B | SscpFunction::WriteVariableGroupDirectly                                                                                                 |
         * | Data length         | 2B | -                                                                                                                                        |
         * | Group ID            | 1B | Identification of a group.                                                                                                               |
         * | Flags               | 1B | One of VarAccessFlags::SetAllValues, VarAccessFlags::SetSelectedValues, VarAccessFlags::SetCompactValues, VarAccessFlags::WriteLargeData |
         * | Number of variables | 1B | This field is only present if telegram contains variable values together with VarAccessFlags::SetSelectedValues flags                    |
         *
         * \sa WriteVariableGroupDirectly_Confirmation
         */
        void WriteVariableGroupDirectly_Request(CommBuffer & buffer, uint8b groupId, VarAccessFlags::VarAccessFlags flags, uint32b variableDescCount,
			uint32b variablesLength);
		
        /*!
         * \brief Read variable structure request
         * \param[out] buffer Output data buffer
         * \param[in] flags Flags
         * \param[in] varCount Number of variables
         * \details Create Read variable structure request header
         *
         * |             |    |                                     |
         * | ----------- | -- | ----------------------------------- |
         * | Function    | 2B | SscpFunction::ReadVariableStructure |
         * | Data length | 2B | 1 + variable count * 4              |
         * | Params      | 1B | VarAccessFlags::UseVmUid            |
         *
         * \sa ReadVariableStructure_Confirmation
         */
        void ReadVariableStructure_Request(CommBuffer & buffer, uint8b flags, uint32b varCount);

        /*!
         * \brief Initiate receive data request
         * \param[out] buffer Output data buffer
         * \param[in] filename Data identification
         * \details Create Initiate receive data request
         *
         *  |                            |    |                                    |
         *  | -------------------------- | -- | ---------------------------------- |
         *  | Function                   | 2B | SscpFunction::InitiateDataReceive  |
         *  | Data length                | 2B | -                                  |
         *  | Data identification length | 1B | Data identification == file name   |
         *  | Data identification        | ?B | Encoded user name is max 128 chars |
         *
         * \sa InitiateReceiveData_Confirmation
         */
		void InitiateReceiveData_Request(CommBuffer & buffer, const char *filename);
        /*!
         * \brief Receive data chunk request
         * \param[out] buffer Output data buffer
         * \param[in] dataOffset Offset
         * \details Create Receive data chunk request
         *
         * |             |    |                                |
         * | ----------- | -- | ------------------------------ |
         * | Function    | 2B | SscpFunction::ReceiveDataChunk |
         * | Data length | 2B | 0x00 0x04                      |
         * | Offset      | 4B | Offset to retrieve             |
         *
         * \sa ReceiveDataChunk_Confirmation
         */
		void ReceiveDataChunk_Request(CommBuffer & buffer, uint32b dataOffset);

        /*!
         * \brief Initiate send data request
         * \param[out] buffer Output data buffer
         * \param[in] filename Data identification
         * \param[in] dataProperties Data file information - Data size, Timestamp
         * \details Create Receive data chunk request
         *
         * |                            |    |                                     |
         * | -------------------------- | -- | ----------------------------------- |
         * | Function                   | 2B | SscpFunction::InitiateDataSend      |
         * | Data length                | 2B | -                                   |
         * | Data identification length | 1B | Data identification == file name.   |
         * | Data identification        | ?B | Encoded file name is max 128 chars. |
         * | Data size                  | 4B | Size of the data                    |
         * | Timestamp                  | 8B | DATE_TIME format                    |
         *
         * \sa DataTransferProperties, InitiateSendData_Confirmation
         */
        void InitiateSendData_Request(CommBuffer & buffer, const char *filename, DataTransferProperties *dataProperties);
        /*!
         * \brief Send data chunk request
         * \param[out] buffer Output data buffer
         * \param[in] maxFrameSize Sended data length
         * \param[in] src Data source
         * \param[in] dataSize Whole data length
         * \param[out] offset Current data offset
         * \param[out] transmissionComplete
         * \details Create Send data chunk request
         *
         * |             |    |                                                   |
         * | ----------- | -- | ------------------------------------------------- |
         * | Function    | 2B | SscpFunction::SendDataChunk                       |
         * | Data length | 2B | (4 + data length)                                 |
         * | Data offset | 4B | The data that are being sent are from this offset |
         * | Data        | ?B | Data itself                                       |
         *
         * \sa SendDataChunk_Request
         */
        void SendDataChunk_Request(CommBuffer & buffer, uint16b maxFrameSize, uint8b *src, uint32b dataSize, uint32b *offset, bool *transmissionComplete);
#if !defined(SSCP_STANDALONE_CLIENT)
        /*!
         * \brief Server function. Don't use
         */
		void SendDataChunk_Request(CommBuffer & buffer, uint16b maxFrameSize, IMemoryAllocator *ma, sint32b allocId, uint32b dataSize, uint32b *offset, 
			bool *transmissionComplete);
#endif
        /*!
         * \brief Finish data send request
         * \param[out] buffer Output data buffer
         * \param[in] crc Data CRC
         * \details Create Finish data send request
         *
         * |             |    |                              |
         * | ----------- | -- | ---------------------------- |
         * | Function    | 2B | SscpFunction::FinishDataSend |
         * | Data length | 2B | 0x00 0x02                    |
         * | CRC         | 2B | CRC of the data.             |
         *
         * \sa FinishDataSend_Confirmation
         */
		void FinishDataSend_Request(CommBuffer & buffer, uint16b crc);
        /*!
         * \brief Time setup request
         * \obsolete Obsoleted from protocol version 7
         * \param[out] buffer Output data buffer
         * \param[in] time Optional 64bit C# timestamp in UTC format for time preset
         * \details Create Time setup request
         *
         * |             |    |                                                           |
         * | ----------- | -- | --------------------------------------------------------- |
         * | Function    | 2B | SscpFunction::TimeSetup                                   |
         * | Data length | 2B | 0/8 bytes                                                 |
         * | Timestamp   | 8B | Optional 64bit C# timestamp in UTC format for time preset |
         *
         * \sa TimeSetup_Confirmation
         */
        void TimeSetup_Request(CommBuffer & buffer, uint64b time);
        /*!
         * \brief Time setup extended
         * \param[out] buffer Output data buffer
         * \param[in] cmd Time function command
         * \param[in] flags Flags (Not used in current version - set to zero)
         * \param[in] time Optional 64bit C# timestamp in UTC format for time preset
         * \details Create Time setup request
         *
         * |             |    |                                                                    |
         * | ----------- | -- | ------------------------------------------------------------------ |
         * | Function    | 2B | SscpFunction::TimeSetupEx                                            |
         * | Data length | 2B | 2/10 bytes                                                         |
         * | Command     | 1B | * 0x01 - Get RTC in UTC format                                     |
         * |             |    | * 0x02 - Get RTC in local format                                   |
         * |             |    | * 0x10 - Set RTC from UTC format (timestamp following after flags) |
         * |             |    | * 0x11 - Set RTC from local format                                 |
         * |             |    | * 0x20 - Get timezone offset (returned as timestamp)               |
         * |             |    | * 0x21 - Get daylight saving offset (returned as timestamp)        |
         * | Flags       | 1B | Not used in current version - set to zero                          |
         * | Timestamp   | 8B | Optional 64bit C# timestamp in UTC format for time preset          |
         *
         * \sa TimeSetupEx_Confirmation
         */
        void TimeSetupEx_Request(CommBuffer & buffer, uint8b cmd, uint8b flags, uint64b time);
        // fixme: schovat pro ReadVars/Group nejakou existenci prenosu souboru vs prime cteni ?
#endif
	#pragma endregion

	#pragma region Funkce - indication
	private:
        /*!
         * \brief Calculate data CRC
         * \param data
         * \param size
         * \param outOffset
         * \param args
         * \return
         * \todo Chybí tělo. Doplnit popisky! Funkce není použitá
         * \details Static function for calculating data CRC
         */
		static bool CalculateCrc(void *data, uint32b size, sint32b outOffset, void *args);

	public:
#if defined(SSCP_ENABLE_SERVER)
		IMessageServiceImplementation::Action Sscp_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);

		IMessageServiceImplementation::Action GetBasicInfo_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer);
		IMessageServiceImplementation::Action SetBasicInfo_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, sint32b & allocationId, sint8b & lastSequenceId);

		IMessageServiceImplementation::Action Login_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		
		IMessageServiceImplementation::Action InitiateSendData_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);	
		IMessageServiceImplementation::Action SendDataChunk_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action FinishSendData_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action InitiateReceiveData_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action ReceiveDataChunk_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		
		IMessageServiceImplementation::Action ReadVariablesDirectly_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action RegisterVariableGroup_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action ReadVariableGroupDirectly_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action WriteVariableDirectly_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action WriteVariableGroupDirectly_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action ReadVariableStructure_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action GetPlcStatistics_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action GetTaskStatistics_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action ExecuteSystemCommand_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action TimeSetup_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action DriverCommand_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action GetStacks_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		IMessageServiceImplementation::Action SetResetReadOnlyMode_Indication(CommBuffer & readBuffer, CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
#endif
	#pragma endregion

	#pragma region Funkce - response - commands
#if defined(SSCP_ENABLE_SERVER)
		void Login_Response(uint8b protocolVersion, uint16b maxDataSize, uint8b rights, CommBuffer & writeBuffer);	
		// InitiateDataSend_Response -> ComposeResponseHeader (CRH), ComposeErrorResponse (CER)
		void SendDataChunk_Response(uint32b offset, CommBuffer & writeBuffer);		
		// FinishDataSend_Response -> CRH/CER
		void InitiateReceiveData_Response(uint32b	totalSize, uint64b timestamp, uint16b	crc, CommBuffer & writeBuffer);			
		void ReceiveDataChunk_Response(uint32b offset, uint16b chunkLen, CommBuffer & writeBuffer);
		// ReadVariablesDirectly_Response
		// RegisterVariableGroup_Response -> CRH/CER
		// ReadVariableGroupDirectly_Response
		// WriteVariablesDirectly_Response -> CRH/CER
		// WriteVariableGroupDirectly_Response -> CRH/CER
		void ReadVariableStructure_Response(uint8b *structure, CommBuffer & writeBuffer);
		void TimeSetup_Response(uint64b timeStamp, CommBuffer & writeBuffer);
		void DriverCommand_Response(uint8b *data, uint8b length, CommBuffer & writeBuffer);
		void GetPlcStatistics_Response(uint8b version, const PlcStatistics & stats, EvaluatorState evalState, CommBuffer & writeBuffer);
		void GetTaskStatistics_Response(uint8b version, uint64b cycleCount, uint64b lastCycleDuration, uint64b averageCycleDuration, uint64b minimalCycleDuration, 
			uint64b maximalCycleDuration, CommBuffer & writeBuffer);
		// ExecuteSystemCommand_Response -> CRH/CER
		GetStacksError GetStacks_Response(uint8b version, const StacksSnapshot & stackSnapshot, uint8b taskId, StackType stackType, uint16b stackOffset, CommBuffer & writeBuffer);
		// SetResetReadOnlyMode_Response -> CRH/CER
#endif
	#pragma endregion

	#pragma region Funkce - confirmation
#if defined(SSCP_ENABLE_CLIENT)
        /*!
         * \brief Login confirmation
         * \param[in] buffer Data from socket
         * \param[out] protocolVersion Protocol version
         * \param[out] maxDataSize Max data size in telegrams
         * \param[out] rights Right group
         * \param[out] imageGuid Image GUID
         * \return Communication error
         * \details Create Login response from readed data
         *
         * |                                       |     |                                                                                                                |
         * | ------------------------------------- | --  | -------------------------------------------------------------------------------------------------------------- |
         * | Function                              | 2B  | 0x81 0x00                                                                                                      |
         * | Data length                           | 2B  | 0x00 0x14                                                                                                      |
         * | Protocol version                      | 1B  | 0x01                                                                                                           |
         * | [Server's] Max data size in telegrams | 2B  | Server indicates the maximal buffer size to the client. Client cannot create frames that are bigger than this. |
         * | Right group                           | 1B  | Users communication rights (ConnectionRights)                                                                  |
         * | Image GUID                            | 16B | Image identification                                                                                           |
         *
         * \sa Login_Request
         */
		SscpError::SscpError Login_Confirmation(CommBuffer & buffer, uint8b *protocolVersion, uint16b *maxDataSize, ConnectionRights *rights, uint8b *imageGuid);
        /*!
         * \brief Read variables directly vars confirmation
         * \param[in] buffer Data from socket
         * \param[out] valuesPtr Readed data from variables
         * \param[out] valuesLength Data length
         * \param[out] errorFlags Unsuccessful calling error code
         * \return Communication error
         * \details Create Read variables directly response from readed data
         *
         * |             |    |                                                                                                           |
         * | ----------- | -- | --------------------------------------------------------------------------------------------------------- |
         * | Function    | 2B | 0x85 0x00                                                                                                 |
         * | Data length | 2B | -                                                                                                         |
         * | Data        | ?B | Data format is appropriate to the specification of Flags in the request. See further the response format. |
         *
         * \sa ReadVariablesDirectlyHeader_Request
         */
		SscpError::SscpError ReadVariablesDirectlyVars_Confirmation(CommBuffer & buffer, uint8b **valuesPtr, uint32b *valuesLength, uint64b *errorFlags);
        /*!
         * \brief Read variable group directly confirmation
         * \param[in] buffer Data from socket
         * \param[out] valuesPtr Readed data from variables
         * \param[out] valuesLength Data length
         * \param[out] errorFlags Unsuccessful calling error code
         * \return Communication error
         * \details Create Read variable group directly response from readed data
         *
         * |             |    |                                                                                                           |
         * | ----------- | -- | --------------------------------------------------------------------------------------------------------- |
         * | Function    | 2B | 0x85 0x06                                                                                                 |
         * | Data length | 2B | -                                                                                                         |
         * | Data        | ?B | Data format is appropriate to the specification of Flags in the request. See further the response format. |
         *
         * \sa ReadVariableGroupDirectly_Request;
         */
		SscpError::SscpError ReadVariableGroupDirectly_Confirmation(CommBuffer & buffer, uint8b **valuesPtr, uint32b *valuesLength, uint64b *errorFlags);

        /*!
         * \brief Write variable directly confirmation
         * \param[in] buffer Data from socket
         * \param[out] errorFlags Unsuccessful calling error code
         * \return Communication error
         * \details Create Write variable directly response from readed data
         *
         * |             |    |                                                                                                           |
         * | ----------- | -- | --------------------------------------------------------------------------------------------------------- |
         * | Function    | 2B | 0x85 0x10                                                                                                 |
         * | Data length | 2B | -                                                                                                         |
         *
         * \sa WriteVariableDirectly_Request
         */
		SscpError::SscpError WriteVariableDirectly_Confirmation(CommBuffer & buffer, uint64b *errorFlags);
        /*!
         * \brief Write variable group directly confirmation
         * \param[in] buffer Data from socket
         * \param[out] errorFlags Unsuccessful calling error code
         * \return Communication error
         * \details Create Write variable group directly response from readed data
         *
         * |             |    |                                                                                                           |
         * | ----------- | -- | --------------------------------------------------------------------------------------------------------- |
         * | Function    | 2B | 0x85 0x11                                                                                                 |
         * | Data length | 2B | -                                                                                                         |
         *
         * \sa WriteVariableGroupDirectly_Request
         */
		SscpError::SscpError WriteVariableGroupDirectly_Confirmation(CommBuffer & buffer, uint64b *errorFlags);

        /*!
         * \brief Read variable structure confirmation
         * \param[in] buffer Data from socket
         * \param[out] structure Variable mem. layout
         * \param[out] length Data length
         * \return Communication error
         * \details Create Read variable structure response from readed data
         *
         * |                      |    |           |
         * | -------------------- | -- | --------- |
         * | Function             | 2B | 0x85 0x15 |
         * | Data length          | 2B | -         |
         * | Variable mem. layout | 2B | -         |
         *
         * \sa ReadVariableStructure_Request
         */
        SscpError::SscpError ReadVariableStructure_Confirmation(CommBuffer & buffer, uint8b **structure, uint32b *length);

        /*!
         * \brief Initiate receive data confirmation
         * \param[in] buffer Data from socket
         * \param[out] dataProperties Data properties (\c Total \c size, \c Timestamp, \c CRC)
         * \return Communication error
         * \details Create Initiate receive data response from readed data
         *
         * |             |    |                   |
         * | ----------- | -- | ----------------- |
         * | Function    | 2B | 0x82 0x10         |
         * | Data length | 2B | 0x00 0x0E         |
         * | Total size  | 4B | Size of the data. |
         * | Timestamp   | 8B | DATE_TIME format  |
         * | CRC         | 2B | CRC of the data.  |
         *
         * \sa InitiateReceiveData_Request
         */
        SscpError::SscpError InitiateReceiveData_Confirmation(CommBuffer & buffer, DataTransferProperties *dataProperties);
        /*!
         * \brief Receive data chunk confirmation
         * \param[in] buffer Data from socket
         * \param[out] dataOffset Data offset
         * \param[out] dest Reaaded data pointer
         * \param[out] transferTerminated Transfer complete or terminated flag
         * \param[out] dataProperties Received data properties
         * \return Communication error
         * \details Create Receive data chunk response from readed data
         *
         * |             |    |                                                   |
         * | ----------- | -- | ------------------------------------------------- |
         * | Function    | 2B | 0x82 0x11                                         |
         * | Data length | 2B | (4 + data length)                                 |
         * | Data offset | 4B | The data that are being sent are from this offset |
         * | Data        | ?B | Data itself                                       |
         *
         * \sa ReceiveDataChunk_Request
         */
		SscpError::SscpError ReceiveDataChunk_Confirmation(CommBuffer & buffer, uint32b *dataOffset, uint8b *dest, bool *transferTerminated, 
			DataTransferProperties *dataProperties);
#if !defined(SSCP_STANDALONE_CLIENT)
        /*!
         * \brief Server function. Don't use
         */
        SscpError::SscpError ReceiveDataChunk_Confirmation(CommBuffer & buffer, uint32b *dataOffset, IMemoryAllocator *ma, sint32b allocId,
			bool *transferTerminated, DataTransferProperties *dataProperties);
#endif
        /*!
         * \brief Initiate send data confirmation
         * \param[in] buffer Data from socket
         * \return Communication error
         * \details Create Initiate send data response from readed data
         *
         * |             |    |           |
         * | ----------- | -- | --------- |
         * | Function    | 2B | 0x82 0x00 |
         * | Data length | 2B | 0x00 0x00 |
         *
         * \sa InitiateSendData_Request
         */
		SscpError::SscpError InitiateSendData_Confirmation(CommBuffer & buffer);
        /*!
         * \brief Send data chunk confirmation
         * \param[in] buffer Data from socket
         * \param[out] confirmedOffset
         * \return Communication error
         * \details Create Send data chunk response from readed data
         *
         * |             |    |                           |
         * | ----------- | -- | ------------------------- |
         * | Function    | 2B | 0x82 0x01                 |
         * | Data length | 2B | 0x00 0x04                 |
         * | Offset      | 4B | Offset being acknowledged |
         *
         * \sa SendDataChunk_Request
         */
		SscpError::SscpError SendDataChunk_Confirmation(CommBuffer & buffer, uint32b *confirmedOffset);
        /*!
         * \brief Finish data send confirmation
         * \param[in] buffer
         * \return Communication error
         * \details Create Finish data send response from readed data
         *
         * |             |    |           |
         * | ----------- | -- | --------- |
         * | Function    | 2B | 0x82 0x02 |
         * | Data length | 2B | 0x00 0x00 |
         *
         * \sa FinishDataSend_Request
         */
		SscpError::SscpError FinishDataSend_Confirmation(CommBuffer & buffer);
        /*!
         * \brief Time setup confirmation
         * \param[in] buffer
         * \param[out] timeReturned Actual 64bit C# PLC time in UTC
         * \return Communication error
         * \details Create Time setup response from readed data
         *
         * |                   |    |                               |
         * | ----------------- | -- | ----------------------------- |
         * | Function          | 2B | 0x86 0x02                     |
         * | Data length       | 2B | 0x00 0x08                     |
         * | Current timestamp | 8B | Actual PLC time in UTC format |
         *
         * \sa TimeSetup_Request
         */
		SscpError::SscpError TimeSetup_Confirmation(CommBuffer & buffer, uint64b *timeReturned);
        /*!
         * \brief Time setup confirmation
         * \param[in] buffer
         * \param[out] timeReturned Optional 64bit C# timestamp in format specified by command for time preset
         * \return Communication error
         * \details Create Time setup response from readed data
         *
         * |                   |    |                               |
         * | ----------------- | -- | ----------------------------- |
         * | Function          | 2B | 0x86 0x04                     |
         * | Data length       | 2B | 0x00 0x08                     |
         * | Current timestamp | 8B | Optional 64bit C# timestamp in format specified by command for time preset |
         *
         * \sa TimeSetup_Request
         */
        SscpError::SscpError TimeSetupEx_Confirmation(CommBuffer & buffer, uint64b *timeReturned);
#endif
	#pragma endregion

	public:
		// Konstruktor serveru
#if defined(SSCP_ENABLE_SERVER)
        /*!
         * \brief Server function. Don't use
         */
        SscpProtocolParser(IEngineAccess * engineAccess, ILargeDataAccess * largeDataAccess, IMemoryAllocator *fileTransferAllocator,
			IMemoryAllocator *memoryTransferAllocator);

		void SetParameters(IAccessValidator *accessValidator, uint8b localSlaveAddress, uint8b variableGroupCount, uint8b variableGroupSize);
#endif
		// Konstruktor klienta
#if defined(SSCP_ENABLE_CLIENT)
	#if !defined(SSCP_STANDALONE_CLIENT)
        /*!
         * \brief Server function. Don't use
         */
        SscpProtocolParser(IMemoryAllocator *memoryTransferAllocator);
	#else
        /*!
         * \brief Constructor
         * \details Create an object instance
         */
		SscpProtocolParser(void);
	#endif
#endif
        /*!
         * \brief Destructor
         * \details Destruct object
         */
        ~SscpProtocolParser();

		// --------------------------
		// Funkce - response - common
		// --------------------------
	public:
        /*!
         * \brief Compose header
         * \param[in] functionId Function ID
         * \param[in] dataLen Function data length
         * \param[out] writeBuffer Data buffer
         * \return Offset in writeBuffer, where SSCP header starts.
         * \details Create function header for sended data to PLC. Doesn't shift offset over written data
         * \sa SscpFunction
         */
		static uint32b ComposeHeader(uint16b functionId, uint16b dataLen, CommBuffer & writeBuffer);	// pozor, neposouva offset za zapsana data !
        /*!
         * \brief Compose error response
         * \param[in] functionId Function ID
         * \param[in] errorCode Error code
         * \param[out] writeBuffer Data buffer
         * \details Create error header.
         * \sa SscpFunction, SscpError
         */
		static void ComposeErrorResponse(uint16b functionId, uint32b errorCode, CommBuffer & writeBuffer);
        /*!
         * \brief Compose error response
         * \param[in] functionId Function ID
         * \param[in] errorCode Error code
         * \param[out] writeBuffer Data buffer
         * \param[in] optionalData Optional data
         * \param[in] optionalDataLength Optional data length
         * \details Create error header with optional data (SSCP server)
         * \sa SscpFunction, SscpError
         * \todo Přepínač jenom pro SSCP SERVER??
         */
		static void ComposeErrorResponse(uint16b functionId, uint32b errorCode, CommBuffer & writeBuffer, const void *optionalData, uint32b optionalDataLength);
        /*!
         * \brief Compose system command error response
         * \param[in] functionId Function ID
         * \param[in] errorCode Error code
         * \param[in] cmd ??
         * \param[out] writeBuffer Data buffer
         * \details Create error message with for command SscpFunction::ExecuteSystemCommand
         * \sa SscpFunction, SscpError
         * \todo Přepínač jenom pro SSCP SERVER?? Není functionId zbytečné? Vžyť to platí pro SscpFunction::ExecuteSystemCommand ?
         */
		static void ComposeSysCmdErrorResponse(uint16b functionId, uint32b errorCode, SystemCommand cmd, CommBuffer & writeBuffer);
        /*!
         * \brief Compose no such variable error response
         * \param[in] functionId Function ID
         * \param[in] errorCode Error code
         * \param[in] variableUid Variable ID
         * \param[out] writeBuffer Data buffer
         * \details Create error message for commands SscpFunction::WriteVariableDirectly or SscpFunction::WriteVariableGroupDirectly
         * \sa SscpFunction
         * \todo Přepínač jenom pro SSCP SERVER??
         */
        void ComposeNoSuchVariableErrorResponse(uint16b functionId, uint32b errorCode, uint32b variableUid, CommBuffer & writeBuffer);
        /*!
         * \brief Check rights compose error response
         * \param[in] currentRights
         * \param[in] requieredRights
         * \param[out] writeBuffer
         * \return
         * \li \c true - User has rights
         * \li \c false - user has invalid rights
         * \details Create error message for invalid user rights. If user have invalid rights, create error message
         * \todo Přepínač jenom pro SSCP SERVER??
         */
		bool CheckRightsComposeErrorResponse(ConnectionRights currentRights, ConnectionRights requieredRights,  CommBuffer & writeBuffer);
#if defined(SSCP_ENABLE_SERVER)
		bool CheckDataLenComposeErrorResponse(uint32b dataLen, uint32b requieredDataLen,  CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		bool HelperCheckLoginInfoFromBuffer(CommBuffer & readBuffer, uint32b *offset, uint32b *outRights);
#endif
	
		// spolecne
        /*!
         * \brief Check frame and slave address
         * \param[in] buffer Data buffer
         * \param[in] enableCrc Check CRC message validation (not used with SSCP_STANDALONE_CLIENT definition)
         * \param[in] slaveAddress SSCP address
         * \param[in] linkHeaderLength Header data length
         * \param[in] header Command header
         * \return
         * \li \c true - Packet is valid
         * \li \c false - Packet is invalid
         * \details Check data frame and slave addres on current device.
         * \todo Je popis správně??
         */
		bool CheckFrameAndSlaveAddress(CommBuffer & buffer, bool enableCrc, uint8b slaveAddress, uint8b linkHeaderLength, SSCPProtocolHeader *header);
        /*!
         * \brief Check function and error stats
         * \param[in] buffer Input data
         * \param[in] functionToCheck Function ID
         * \param[out] dataLength Function's data length
         * \return Communication error
         * \details Parse data buffer. If error code is detected, return this error code.
         *
         * |               |    |                             |
         * | ------------- | -- | --------------------------- |
         * | Function      | 2B | (0xC0 0x00) | Function code |
         * | Data length   | 2B | 0x00 0x04                   |
         * | Error code    | 4B | SscpError code              |
         *
         * \sa SscpFunction
         */
		SscpError::SscpError CheckFunctionAndErrorStats(CommBuffer & buffer, SscpFunction::SscpFunction functionToCheck, uint32b *dataLength);
        /*!
         * \brief Check function and error stats
         * \param[in] buffer Input data
         * \param[in] functionToCheck Function ID
         * \param[out] dataLength Function's data length
         * \param[out] optionalData Optional data
         * \return Communication error
         * \details Parse data buffer. If error code is detected, return this error code.
         *
         * |               |    |                             |
         * | ------------- | -- | --------------------------- |
         * | Function      | 2B | (0xC0 0x00) | Function code |
         * | Data length   | 2B | -                           |
         * | Error code    | 4B | SscpError code              |
         * | Optional data | ?B | Optional data               |
         *
         * \sa SscpFunction
         */
		SscpError::SscpError CheckFunctionAndErrorStats(CommBuffer & buffer, SscpFunction::SscpFunction functionToCheck, uint32b *dataLength, uint64b *optionalData);

		// klientska cast
#if defined(SSCP_ENABLE_CLIENT) && !defined(SSCP_STANDALONE_CLIENT)
	private:
		bool m_OneReadCyclePassed;
		bool m_EnableWritingCycles;
	public:
		void CommonRequestPrepare(ClientContextSpecificData *context);
		bool CommonResponseCheck(ClientContextSpecificData *context);
		bool ProcessNextVariableGroup(ClientContextSpecificData *context);
		SscpProtocolParser::DriverState ClientProcessCurrentGroup(ClientContextSpecificData *context);
		SscpError::SscpError ClientUpdateValues(ClientContextSpecificData *context, ICallbackCommunicationDriver::CallbackAction *action);
		void CopyVarValueToCommBuffer(ClientContextSpecificData *context, CommunicationVariable *variable, bool useMemoryAllocator, uint32b *offset);

		virtual bool CheckOneReadCyclePassed(void);
		virtual void EnableWritingCycles(void);
#endif

#if defined(SSCP_ENABLE_SERVER)
	private:
		OperationStatus ParseProtocolVariable(CommBuffer & readBuffer, uint32b offset, bool commUidOnly, bool vmUid, 
			/* out */ SSCPProtocolVariableDefinition &variable, const IMessageServiceImplementation::ClientContext * contextData);
		SscpError::SscpError LoadVariableGroupToFile(uint8b groupId, uint8b flags, const IMessageServiceImplementation::ClientContext * contextData, uint32b *unknownUid);
		void BuildDataAccessErrorResponse(SscpFunction::SscpFunction function, OperationStatus error, 
			uint32b failedUid, CommBuffer & writeBuffer, uint64b errorFlags);

		virtual bool ChangedEnabled(void) { return (true); }
		virtual void ChangeHook(CVariable *var, uint32b threadId);
#endif

	#pragma region Function parsing - indication
#if defined(SSCP_ENABLE_SERVER)
	private:
		bool ParseLogin_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolLoginRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseInitiateSendData_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolInitiateDataSendRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseSendDataChunk_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolSendDataChunkRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseFinishSendData_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolFinishSendDataRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseInitiateReceiveData_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolInitiateDataReceiveRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseReceiveDataChunk_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolReceiveDataChunkRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseReadVariablesDirectly_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolReadVariablesDirectlyRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseRegisterVariableGroup_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolRegisterVariableGroupRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseReadVariableGroupDirectly_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolReadVariableGroupDirectlyRequest &request, 
			/* out */ CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseWriteVariablesDirectly_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolWriteVariablesDirectlyRequest &request, 
			/* out */ CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseWriteVariableGroupDirectly_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolWriteVariableGroupDirectlyRequest &request, 
			/* out */ CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseReadVariableStructure_Indication(CommBuffer & readBuffer, SSCPProtocolReadVariableStructureRequest &request,
			/* out */ CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseGetPlcStatistics_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolGetPlcStatisticsRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseGetTaskStatistics_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolGetTaskStatisticsRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseSystemCommand_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolSystemCommandRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseTimeSetup_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolTimeSetupRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseDriverCommand_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolDriverCommandRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseGetStacks_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolGetStacksRequest &request, /* out */ CommBuffer & writeBuffer, 
			const IMessageServiceImplementation::ClientContext * contextData);
		bool ParseSetResetReadOnlyMode_Indication(CommBuffer & readBuffer, /* out */ SSCPProtocolSetResetReadOnlyModeRequest &request, /* out */ sint32b & varCount, 
			/* out */ CommBuffer & writeBuffer, const IMessageServiceImplementation::ClientContext * contextData);
#endif
	#pragma endregion
};

#endif
