#include "./SystemCalls.h"
#include "./PlatformUtils.h"
#include <time.h>

#if defined(_MSC_VER) || defined(__BORLANDC__)
#include <WinSock2.h>
#include <windows.h>
#endif

#if defined(__POSIX__)
#include <time.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <libgen.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#if defined(__PIKEOS__)
#include <ddapi/ddapi.h>
extern "C" 
{
#include <p4.h>
}
#endif
#endif

#if defined( __FREERTOS__ )

#include "./../../HW/FreeRTOS-ARM/sysutils/rtc.h"

extern "C" {
#include "task.h"
}
#endif


void SystemCalls::Sleep(uint32b milliseconds)
{
#if defined(_MSC_VER) || defined(__BORLANDC__)
	::Sleep(milliseconds);
#endif

#if defined __POSIX__
	timespec req;

	req.tv_sec = milliseconds / 1000;
	req.tv_nsec = 1000000 * (milliseconds % 1000);

	while (nanosleep(&req, &req) == -1)
	{
		if (errno != EINTR)
			break;
	}
#endif

#if defined( __FREERTOS__ )
	portTickType tickDelay = milliseconds / portTICK_RATE_MS;

	if (tickDelay == 0)
	{
		SharkPrintf("Pozadavek na sleep kratsi nez 1 tik\n");
		tickDelay = 1;
	}

	vTaskDelay(tickDelay);
#endif
}

//-------------------------------------------------------------------------------------------------
void SystemCalls::Usleep(uint32b microseconds)
{
#if defined(_MSC_VER) || defined(__BORLANDC__)
	if (microseconds < 10000)
	{
		unsigned __int64 tmp;

		QueryPerformanceFrequency((LARGE_INTEGER *)&tmp);
		uint64b m_Frequency = (uint64b)tmp;

		QueryPerformanceCounter((LARGE_INTEGER *)&tmp);
		uint64b m_StartTime = (uint64b)tmp;

		while (true)
		{
			unsigned __int64 currentTicks;

			QueryPerformanceCounter((LARGE_INTEGER *)&currentTicks);

			double elapsed = 10000000.0 * (currentTicks - m_StartTime) / m_Frequency;
			if (elapsed / 10 > microseconds)
				break;
		}
	}
	else
		::Sleep(microseconds / 1000);
#endif

#if defined(__POSIX__)
	struct timespec req;

	req.tv_sec = microseconds / 1000000;
	req.tv_nsec = microseconds * 1000;

	while (nanosleep(&req, &req) == -1)
	{
		if (errno != EINTR)
			break;
	}
#endif

#if defined( __FREERTOS__ )
	portTickType tickDelay = (microseconds / 1000) / portTICK_RATE_MS;

	if (tickDelay == 0)
	{
		SharkPrintf("Pozadavek na usleep kratsi nez 1 tik\n");
		tickDelay = 1;
	}

	vTaskDelay(tickDelay);
#endif
}

//-------------------------------------------------------------------------------------------------

bool SystemCalls::DirectoryExists(const sint8b *dirName)
{
#if defined(_MSC_VER)
#if WINAPI_FAMILY==WINAPI_FAMILY_PHONE_APP
	DWORD dwAttrib;
	_WIN32_FILE_ATTRIBUTE_DATA fileInfo;
	if (!GetFileAttributesExA((const char*)dirName, GetFileExInfoStandard, &fileInfo)) dwAttrib = INVALID_FILE_ATTRIBUTES;
	else dwAttrib = fileInfo.dwFileAttributes;
#else
	DWORD dwAttrib = GetFileAttributesA((const char*)dirName);
#endif

	return (dwAttrib != INVALID_FILE_ATTRIBUTES && (dwAttrib & FILE_ATTRIBUTE_DIRECTORY));
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	struct stat sb;
	return (stat((const char*)dirName, &sb) == 0);
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
	return (false);
}

//-------------------------------------------------------------------------------------------------

bool SystemCalls::CreateDirectoryRecursive(const sint8b *dirName, bool containsFileName, void *owner, void *group)
{
#if defined(_MSC_VER)
	char dir[PATH_MAX];
	strncpy(dir, (const char*)dirName, MAX_PATH - 1);

	if (containsFileName)
	{
		// pokud cesta k adresari obsahuje i jmeno souboru, tak ho odrizneme
		char *end = dir + strlen(dir);
		while (end > dir)
		{
			if (*end == '\\')
			{
				*end = 0;
				break;
			}
			end--;
		}
	}

	if (!DirectoryExists((const sint8b*)dir))
	{
		char *ptr = dir;
		while (*ptr)
		{
			char *ptr2 = strstr(ptr, "\\");
			if (ptr2)
			{
				// po castech vytvori adresare
				if (!DirectoryExists((const sint8b*)dir))
				{
					*ptr2 = 0;
					CreateDirectoryA(dir, NULL);
					*ptr2 = '\\';
				}
			}
			else
			{
				if (!DirectoryExists((const sint8b*)dir))
					CreateDirectoryA(dir, NULL);
				break;
			}

			ptr = ptr2 + 1;
		}
	}
	return (true);
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	char dir[PATH_MAX];
	strncpy(dir, (const char*)dirName, PATH_MAX - 1);
	if (containsFileName)
		strncpy(dir, dirname((char*)dir), PATH_MAX - 1);
	if (!DirectoryExists((const sint8b*)dir))
	{
		// adresar neexistuje a je nutne ho vytvorit
		char *ptr = dir + 1;
		while (*ptr)
		{
			char *ptr2 = strstr(ptr, "/");
			if (ptr2)
			{
				*ptr2 = 0;
				if (mkdir((const char*)dir, 0777) == -1)
				{
					if (errno != EEXIST)
						return (false);
				}
				chown((const char*)dir, (sint32b)owner, (sint32b)group);
				*ptr2 = '/';
			}
			else
			{
				if (!DirectoryExists((const sint8b*)dir))
				{
					mkdir(dir, 0777);
					chown((const char*)dir, (sint32b)owner, (sint32b)group);
				}
				break;
			}
			ptr = ptr2 + 1;
		}
	}
	return (true);
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
}

//-------------------------------------------------------------------------------------------------

bool SystemCalls::ClearDirectory(const sint8b *dirName, bool recursive)
{
#if defined(_MSC_VER)
	if (recursive)
	{
		// muze byt staticky - pouzije se to jen pri findfirst a dalsi vnoreni si to prepisuje a nicemu to nevadi
		static char dirPath[PATH_MAX];
		strcpy(dirPath, (const char*)dirName);
		strcat(dirPath, "\\*");

		WIN32_FIND_DATAA ffd;
#if WINAPI_FAMILY==WINAPI_FAMILY_PHONE_APP
		HANDLE hFind = FindFirstFileExA((const char*)dirPath, FindExInfoStandard, &ffd, FindExSearchNameMatch, NULL, 0);
#else
		HANDLE hFind = FindFirstFileA((const char*)dirPath, &ffd);
#endif
		if (hFind != INVALID_HANDLE_VALUE)
		{
			do
			{
				char tmpPath[PATH_MAX]; // fixme: todle se mi nelibi - pri kazdem vnoreni 
				strncpy(tmpPath, (const char*)dirName, PATH_MAX - 1);
				strncat(tmpPath, "\\", PATH_MAX - 1);
				strncat(tmpPath, ffd.cFileName, PATH_MAX - 1);

				if (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				{
					if ((strcmp(ffd.cFileName, ".") != 0) && (strcmp(ffd.cFileName, "..") != 0))
					{
						ClearDirectory((const sint8b*)tmpPath, recursive);
						RemoveDirectoryA((const char*)tmpPath);
					}
				}
				else
					DeleteFileA(tmpPath);
			}
			while (FindNextFileA(hFind, &ffd) != 0);
			FindClose(hFind);
		}
	}
	else
		return (RemoveDirectoryA((const char*)dirName) == TRUE);
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	if (recursive)
	{
		DIR *dir;
		struct dirent *entry;
		dir = opendir((const char*)dirName);
		while ((entry = readdir(dir)) != NULL)
		{
			char tmpPath[PATH_MAX]; // fixme: todle se mi nelibi - pri kazdem vnoreni 
			strncpy(tmpPath, (const char*)dirName, PATH_MAX - 1);
			strncat(tmpPath, "/", PATH_MAX - 1);
			strncat(tmpPath, entry->d_name, PATH_MAX - 1);

			if (entry->d_type == DT_DIR)
			{
				if ((strcmp(entry->d_name, ".") != 0) && (strcmp(entry->d_name, "..") != 0))
				{
					ClearDirectory((const sint8b*)tmpPath, recursive);
					rmdir((const char*)tmpPath);
				}
			}
			else
				unlink(tmpPath);
		}
		if (dir)
			closedir(dir);
	}
	else
		return (rmdir((const char*)dirName) == 0);
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
	return (false);
}

//-------------------------------------------------------------------------------------------------

void SystemCalls::FixFileSystemPath(sint8b *path)
{
#if defined(_MSC_VER)
	char *ptr = (char*)path;
	while (*ptr)
	{
		// provede nahrazeni beznych lomitek za opacna
		char *ptr2 = strstr(ptr, "/");
		if (ptr2)
			*ptr2 = '\\';
		else
			break;
		ptr = ptr2 + 1;
	}
#elif defined(__POSIX__) && !defined(__PIKEOS__)
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
}

//-------------------------------------------------------------------------------------------------

file_t SystemCalls::OpenFile(const sint8b* fileName, bool create, bool append, bool readOnly)
{
#if defined(_MSC_VER)
	if (readOnly)
		return ((file_t)fopen((const char*)fileName, "rb"));
	else if (create)
		return ((file_t)fopen((const char*)fileName, "wb"));
	else if (append)
		return ((file_t)fopen((const char*)fileName, "a+b"));
	else
		return ((file_t)fopen((const char*)fileName, "r+b"));
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	uint32b flags = (readOnly) ? O_RDONLY : O_RDWR;
	if (create)
		flags |= O_CREAT | O_TRUNC;
	if (append)
		flags |= O_APPEND;
	file_t res = (file_t)open((const char*)fileName, flags, (mode_t)0600);
	if ((res == INVALID_FILE_VALUE) && append)
		res = (file_t)open((const char*)fileName, flags | O_CREAT & (~O_APPEND), (mode_t)0600);
	return (res);
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
	return (INVALID_FILE_VALUE);
}

//-------------------------------------------------------------------------------------------------

uint32b SystemCalls::WriteFile(file_t fd, const void *buffer, uint32b size)
{
#if defined(_MSC_VER)
	return (fwrite(buffer, 1, size, fd));
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	sint32b written = write(fd, buffer, size);
	fsync(fd);
	return (written);
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
}

//-------------------------------------------------------------------------------------------------

uint32b SystemCalls::ReadFile(file_t fd, void *buffer, uint32b maxSize)
{
#if defined(_MSC_VER)
	return (fread(buffer, 1, maxSize, fd));
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	return (read(fd, buffer, maxSize));
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
}

//-------------------------------------------------------------------------------------------------

uint32b SystemCalls::SeekFile(file_t fd, sint32b position, bool relative)
{
#if defined(_MSC_VER)
	return (fseek(fd, position, relative ? SEEK_CUR : SEEK_SET));
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	return (lseek(fd, position, relative ? SEEK_CUR : SEEK_SET));
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
}

//-------------------------------------------------------------------------------------------------

uint32b SystemCalls::Size(file_t fd)
{
#if defined(_MSC_VER)
	uint32b pos = ftell(fd);
	fseek(fd, 0, SEEK_END);
	uint32b res = ftell(fd);
	fseek(fd, pos, SEEK_SET);
	return (res);
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	uint32b pos = lseek(fd, 0, SEEK_CUR);
	sint32b res = lseek(fd, 0, SEEK_END);
	lseek(fd, pos, SEEK_SET);
	return (res >= 0 ? res : 0);
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
}

//-------------------------------------------------------------------------------------------------

void SystemCalls::CloseFile(file_t fd)
{
#if defined(_MSC_VER)
	fclose(fd);
#elif defined(__POSIX__) && !defined(__PIKEOS__)
	close(fd);
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
}

//-------------------------------------------------------------------------------------------------

void SystemCalls::RenameFile(const sint8b *fileFrom, const sint8b *fileTo)
{
#if defined(_MSC_VER) || (defined(__POSIX__) && !defined(__PIKEOS__))
	_unlink((const char*)fileTo);
	rename((const char*)fileFrom, (const char*)fileTo);
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
#else
#endif
}

//-------------------------------------------------------------------------------------------------
