#ifndef __SystemCalls_H
#define __SystemCalls_H

#include "./datatypes.h"

#include <stdio.h>

#if defined(_MSC_VER) || defined(__BORLANDC__)
typedef FILE* file_t;
#define INVALID_FILE_VALUE (file_t)NULL
#elif defined(__POSIX__) && !defined(__PIKEOS__)
typedef int file_t;
#define INVALID_FILE_VALUE (file_t)-1
#elif defined(__PIKEOS__)
#elif defined(__FREERTOS__)
typedef int file_t;
#define INVALID_FILE_VALUE (file_t)-1
#else
#endif
/*!
 * \brief System calls
 * \details Static class with helper functions
 * \todo Internal - Odkomentovat
 */
class SystemCalls
{
public:
    /*!
     * \brief Sleep
     * \param[in] milliseconds Timeout in ms
     * \details Stops continuing code running for specific timeout
     */
	static void Sleep(uint32b milliseconds);
    /*!
     * \brief Sleep
     * \param[in] microseconds Timeout in microseconds
     * \details Stops continuing code running for specific timeout
     */
    static void Usleep(uint32b microseconds);

    /*!
     * \brief Directory exists
     * \param[in] dirName Directory path
     * \return Directory exist flag
     * \details Check if selected directory exists
     */
	static bool DirectoryExists(const sint8b *dirName);
	// fixme: ty void na konci se mi uplne nelibi, je treba nejak zobecnit to, ze na posixu je nutne menit vlastnika a na win to 
	// bude treba uplne jinak
    /*!
     * \brief Create directory recursive
     * \param[in] dirName Directory path
     * \param[in] containsFileName ??
     * \param[in] owner Owner for created directory (linux)
     * \param[in] group Group for created directory (linux)
     * \return \c true - directory created
     * \details Create directory on selected path. If path didn't exists, function create it.
     * \todo Není použito. Kód je zakomentovaný. Co k tomu mam napsat?
     */
	static bool CreateDirectoryRecursive(const sint8b *dirName, bool containsFileName, void *owner, void *group);
    /*!
     * \brief Clear directory
     * \param[in] dirName Directory path
     * \param[in] recursive Delete subfolderd flag
     * \return \c true - action success
     * \details Delete selected directories with subdirectories (\c recursive = true)
     */
	static bool ClearDirectory(const sint8b *dirName, bool recursive);
    /*!
     * \brief Fix file system path
     * \param[out] path
     * \details Fixed file path to OS specified format
     */
	static void FixFileSystemPath(sint8b *path);

	// metody pro pristup k souborum
    /*!
     * \brief Open file
     * \param[in] fileName File name with path
     * \param[in] create Creation flag
     * \param[in] append Append flag
     * \param[in] readOnly Readonly flag
     * \return Pointer to opened file
     * \details. Create or open specified file in specifid mode.
     * \li \c create - Open file in write mode. If file doesn't exists, create it. If file exist, clear it.
     * \li \c append - Open file in write mode and continue at the last position.
     * \li \c readOnly - Open file in read mode
     *
     * \todo Není kontraproduktivní rozlyšovat create, append a readOnly? Chtělo by to asi enum a ve funkci potom dát elseif.
     * \code{.cpp}
     *  uint32b flags = (readOnly) ? O_RDONLY : O_RDWR;
     *  if (create)
     *      flags |= O_CREAT | O_TRUNC;
     *  if (append)
     *      flags |= O_APPEND;
     * \endcode
     * \sa WriteFile, ReadFile, SeekFile, CloseFile
     */
	static file_t OpenFile(const sint8b* fileName, bool create, bool append, bool readOnly);
    /*!
     * \brief Write file
     * \param[in] fd Pointer to file
     * \param[in] buffer Data to write
     * \param[in] size Data size
     * \return Size of written data
     * \details Write data to specified file. This file must be opened in write mode
     * \sa OpenFile
     */
	static uint32b WriteFile(file_t fd, const void *buffer, uint32b size);
    /*!
     * \brief Read file
     * \param[in] fd Pointer to file
     * \param[out] buffer Readed data buffer
     * \param[in] maxSize Maximum size of readed data
     * \return Size of readed data
     * \details Read data from file. Maximum data size is specifed with \c maxSize property. \c buffer must have at least \c maxSize capacity
     * \sa OpenFile
     */
	static uint32b ReadFile(file_t fd, void *buffer, uint32b maxSize);
    /*!
     * \brief Seek file
     * \param[in] fd Pointer to file
     * \param[in] position Position
     * \param[in] relative Flag for relative seek
     * \return Actual position
     * \details Send pointer in fil by \c position relative (\c relative = true) to current pointer position or absulutly (\c relative = true)
     * \sa OpenFile
     */
	static uint32b SeekFile(file_t fd, sint32b position, bool relative);
    /*!
     * \brief Size
     * \param[in] fd Pointer to file
     * \return File size
     * \details Get file size
     * \sa OpenFile
     */
	static uint32b Size(file_t fd);
    /*!
     * \brief CloseFile
     * \param[in] fd Pointer to file
     * \details Close opened file
     * \sa OpenFile
     */
	static void CloseFile(file_t fd);

    /*!
     * \brief Rename file
     * \param[in] fileFrom Path of current file name
     * \param[in] fileTo Path of new file name
     * \details Rename selected file from name \c fileFrom to name \c fileTo
     */
	static void RenameFile(const sint8b *fileFrom, const sint8b *fileTo);
};

#endif
