#include "./TcpClient.h"

#include "./NetworkingHelpers.h"
#include "./SystemCalls.h"

TcpUdpClient::TcpUdpClient(uint16b endpointsCount)
#if !defined(SSCP_STANDALONE_CLIENT)
	: ICallbackChannel()
#endif
{
	m_EndpointsCount = endpointsCount;
	m_Endpoints = new IClientEndpoint*[m_EndpointsCount];

	for (int i = 0; i < m_EndpointsCount; i++)
		m_Endpoints[i] = NULL;

	m_Timeout = 1000; // v ms, fixme: odnekud brat

#if defined(__THREAD_ID__) && defined(SSCP_STANDALONE_CLIENT)
	SetThreadName("Tcp client");
#endif
}

//-------------------------------------------------------------------------------------------------

TcpUdpClient::~TcpUdpClient(void)
{
	// fixme: nejak to sjednotit. takovyhl epoli je strasne moc
	for (uint32b a = 0; a < m_EndpointsCount; a++)
	{
		if (m_Endpoints[a])
		{
			delete (m_Endpoints[a]);
			m_Endpoints[a] = NULL;
		}
	}

	delete [] m_Endpoints;
}

//-------------------------------------------------------------------------------------------------

void TcpUdpClient::SetEndpoint(uint32b index, IClientEndpoint * endpoint)
{
	if (index < m_EndpointsCount)
		m_Endpoints[index] = endpoint;
}

//-------------------------------------------------------------------------------------------------

sint32b TcpUdpClient::FindEndpoint(uint32b ipv4, uint16b port)
{
	for (uint32b a = 0; a < m_EndpointsCount; a++)
	{
		if ((m_Endpoints[a]) && (m_Endpoints[a]->Match(ipv4, port)))
			return (a);
	}
	return (-1);
}

//-------------------------------------------------------------------------------------------------

void TcpUdpClient::ClearAndFreeEndpoint(void)
{
	for (uint32b a = 0; a < m_EndpointsCount; a++)
	{
		if (m_Endpoints[a])
		{
			delete m_Endpoints[a];
			m_Endpoints[a] = NULL;
		}
	}
}

//-------------------------------------------------------------------------------------------------

void TcpUdpClient::Initialize()
{
}

//-------------------------------------------------------------------------------------------------
void TcpUdpClient::Execute(volatile bool * running)
{
	FD_SET_TYPE reader;
	FD_SET_TYPE writer;	
	FD_SET_TYPE excepts;	

	sint32b totalEvents;

	bool endpointWantsWrite, endpointWantsRead;

	bool bRunning = true, lastRunning = true;
	while (bRunning)
	{
		if (lastRunning && (*running == false))
		{
			// only call while running transit from true to false
			for (uint32b a = 0; a < m_EndpointsCount; a++)
				if (m_Endpoints[a])
					m_Endpoints[a]->ShutdownRequest(false);
		}
		lastRunning = *running;

		if (! (*running))
		{
			// if client is stopping, check all endpoints if it has stopped
			bool bState = true;
			for (uint32b a = 0; (a < m_EndpointsCount) && bState; a++)
				if (m_Endpoints[a])
					if (!m_Endpoints[a]->ShutdownRequest(true))
						bState = false;
			if (bState)
				bRunning = false;
		}

		timeval timeout;
		sint32b maxSock = -1;
		
		FD_ZERO(&reader);
		FD_ZERO(&writer);
		FD_ZERO(&excepts);
		
		bool bRwAction = false;
		for (uint32b a = 0; a < m_EndpointsCount; a++)
		{		
			IClientEndpoint * endpoint = m_Endpoints[a];
			
			if (endpoint)
			{
				endpoint->WantRWAction(&endpointWantsRead, &endpointWantsWrite);

				SharkSocket sock = endpoint->GetSocket();

				if (sock != INVALID_SOCKET)
				{
					if ((sint32b)sock > maxSock)
						maxSock = sock;

					if (endpointWantsRead)
						FD_SET(sock, &reader);

					if (endpointWantsWrite)
						FD_SET(sock, &writer);

					if (endpoint->IsConnected())
						bRwAction = endpointWantsRead | endpointWantsWrite;

					FD_SET(sock, &excepts);
				}
			}
		}

		if (bRwAction)
		{
			// timeout pro pripad rw operace
			timeout.tv_sec = m_Timeout / 1000;
			timeout.tv_usec = (m_Timeout % 1000) * 1000;
		}
		else
		{
			// timeout pouze pro pripad cekani na vyjimku
			timeout.tv_sec = 0;
			timeout.tv_usec = 1000; // 1 ms
		}

		//Protoze na PikeOS je vse blokujici nemuzeme si dovolit cteni nebo zapis ze socketu pred selectem
		//Vzdy je potreba nastavit requesty a otestovat selectem jestli to projde bez blokovani
		//Na PikeOS nefunguje timeout bez excepts
		SystemCalls::Sleep(40);

		totalEvents = select(maxSock + 1, &reader, &writer, &excepts, &timeout);

		for (uint32b a = 0; a < m_EndpointsCount; a++)
		{
			if (m_Endpoints[a])
			{
				bool checkTimeout = true;
				SharkSocket sock = m_Endpoints[a]->GetSocket();
			
				if (sock != INVALID_SOCKET)
				{
					if (FD_ISSET(sock, &excepts))
					{
						m_Endpoints[a]->OnExceptionEvent();
					}
					else
					{
						if (FD_ISSET(sock, &reader))
						{
							if (m_Endpoints[a]->OnReadEvent())
							{
								checkTimeout = false;
							}
						}

						if (FD_ISSET(sock, &writer))
						{
							if (m_Endpoints[a]->OnWriteEvent())
							{
								checkTimeout = false;
							}
						}

						if (FD_ISSET(sock, &excepts))
						{
							int sockError;
#if defined(__FREERTOS__)
							u32_t len = 4;
#else
							int len = 4;
#endif
#if defined(__POSIX__)
							getsockopt(sock, SOL_SOCKET, SO_ERROR, (char*)&sockError, (socklen_t*)&len);
#else
							getsockopt(sock, SOL_SOCKET, SO_ERROR, (char*)&sockError, &len);
#endif
							len++;
						}
					}
				}
			
				if (checkTimeout)
				{
					m_Endpoints[a]->CheckTimeout();
				}
			}
		}				
	}
}

//-------------------------------------------------------------------------------------------------

void TcpUdpClient::Deinitialize()
{
#if defined(SSCP_STANDALONE_CLIENT)
	for (uint32b a = 0; a < m_EndpointsCount; a++)
	{
		if (m_Endpoints[a])
			m_Endpoints[a]->Disconnect();
	}
#endif
}

//-------------------------------------------------------------------------------------------------

#if !defined(SSCP_STANDALONE_CLIENT)
bool TcpUdpClient::CheckOneReadCyclePassed(void)
{
	bool result = true;
	for (uint32b i = 0; (i < m_EndpointsCount) && result; i++)
		result &= m_Endpoints[i]->CheckOneReadCyclePassed();
	return (result);
}

//-------------------------------------------------------------------------------------------------

void TcpUdpClient::EnableWritingCycles(void)
{
	for (uint32b i = 0; i < m_EndpointsCount; i++)
		m_Endpoints[i]->EnableWritingCycles();
}

//-------------------------------------------------------------------------------------------------
#endif