#if defined(__POSIX__)
#include <errno.h>
#endif
#include "./TcpClientEndpoint.h"
#if defined(STANDALONE_SUPPORT)
#include "./TimeRanges.h"
#include "./SystemLogger.h"
#else
#include "./../../Shared/Utils/TimeRanges.h"
#endif

TcpClientEndpoint::TcpClientEndpoint(uint32b ipv4, uint32b timeoutInMs)
{
	m_State = EState_NotConnected;
	m_Socket = INVALID_SOCKET;
	m_Ipv4 = ipv4;
	m_Timeout = timeoutInMs * TimeRanges::TicksPerMilisecs;

	m_IdleTimer.Restart();
	m_WokenAction = ICallbackCommunicationDriver::CallbackAction_None;
}

//-------------------------------------------------------------------------------------------------

TcpClientEndpoint::~TcpClientEndpoint(void)
{
	Disconnect();
}

//-------------------------------------------------------------------------------------------------

void TcpClientEndpoint::SetParameters(uint32b port, ICallbackCommunicationDriver * driver, uint32b tag)
{
	m_Port = port;
	m_Driver = driver;
	m_Tag = tag;

#if !defined(SSCP_STANDALONE_CLIENT)
	m_Buffer = driver->GetCommunicationBuffer();
#else
	m_Buffer = driver->GetCommunicationBuffer(m_Tag);
#endif
}

//-------------------------------------------------------------------------------------------------

bool TcpClientEndpoint::Match(uint32b ipv4, uint16b port)
{
	return ((m_Ipv4 == ipv4) && (m_Port == port));
}

//-------------------------------------------------------------------------------------------------

void TcpClientEndpoint::CheckTimeout(void)
{
	uint64b t = m_Timer.GetTotalTime();
	if ((t > m_Timeout) && (m_State != EState_Disconnected))
	{
#if !defined(SSCP_STANDALONE_CLIENT)
		Logger.LogDebug(NULL, "Timeout");
#endif
		Disconnect();
	}
}

//-------------------------------------------------------------------------------------------------

bool TcpClientEndpoint::Start(void)
{
#if !defined(SSCP_STANDALONE_CLIENT)
	m_Driver->Reset();
#else
	m_Driver->Reset(m_Tag);
#endif

	return Connect();
}

//-------------------------------------------------------------------------------------------------

void TcpClientEndpoint::WantRWAction(bool *wantRead, bool *wantWrite)
{
	*wantWrite = *wantRead = false;

	switch (m_State)
	{
		case EState_NotConnected:
			{
				if (Start())
				{
					m_State = EState_Connecting;
					m_IdleTimer.Restart();
					*wantWrite = true;
				}
				else
				{
					m_State = EState_NotConnected;
				}
				m_Timer.Restart();
			} break;
		case EState_Connecting:
			{
				*wantWrite = true;
				if (m_WokenAction == ICallbackCommunicationDriver::CallbackAction_None)
				{
					// pokud se probouzime z disconnectu, tak buffer nesmime nulovat - obsahuje jiz platna data
					m_Buffer->Offset = 0;
					m_Buffer->DataLen = 0;
				}
			} break;
		case EState_WritingData:
			{
				*wantWrite = true;
			} break;
		case EState_ReadingData:
			{
				*wantRead = true;
			} break;
		case EState_Connected:
			{
				ICallbackCommunicationDriver::CallbackAction action = ICallbackCommunicationDriver::CallbackAction_None;
				if (m_WokenAction == ICallbackCommunicationDriver::CallbackAction_None)
				{
#if !defined(SSCP_STANDALONE_CLIENT)
					action = m_Driver->GetNextAction();
#else
					action = m_Driver->GetNextAction(m_Tag);
#endif
				}
				else
				{
					action = m_WokenAction;
					m_WokenAction = ICallbackCommunicationDriver::CallbackAction_None;
				}

				if (action == ICallbackCommunicationDriver::CallbackAction_WriteData)
				{
					m_State = EState_WritingData;
					m_IdleTimer.Restart();
					*wantWrite = true;
				}
				else if (action == ICallbackCommunicationDriver::CallbackAction_ReadData)
				{
					m_State = EState_ReadingData;
					m_IdleTimer.Restart();
					*wantRead = true;
				}
				else
				{
					// ICallbackCommunicationDriver::CallbackAction_None
					if (m_IdleTimer.GetTotalTimeInMs() > 10000)
					{
						// pokud driver 10 s nic nechce, uzavreme spojeni a nechame endpoint v klidu
						Disconnect();
						m_State = EState_Disconnected;
					}
				}
			} break;
		case EState_Disconnected:
			{
#if !defined(SSCP_STANDALONE_CLIENT)
				ICallbackCommunicationDriver::CallbackAction action = m_Driver->GetNextAction();
#else
				ICallbackCommunicationDriver::CallbackAction action = m_Driver->GetNextAction(m_Tag);
#endif

				m_WokenAction = action;
				if (action != ICallbackCommunicationDriver::CallbackAction_None)
				{
					// nejdrive musime projit celym systemem connectu a pak teprve muzeme zapisovat/cist
					m_State = EState_NotConnected;
				}
			} break;
		default:
			break;
	}
}

//-------------------------------------------------------------------------------------------------

bool TcpClientEndpoint::OnReadEvent(void)
{
	bool result = true;

	switch (m_State)
	{
		case EState_Connecting:
			{
				m_State = EState_Connected;
			} break;
		case EState_ReadingData:
			{
				result = (ReadData() > 0);
			} break;
		default: break;
	}

	return result;
}

//-------------------------------------------------------------------------------------------------

bool TcpClientEndpoint::OnWriteEvent(void)
{
	bool result = true;

	switch (m_State)
	{
		case EState_Connecting:
			{
				m_State = EState_Connected;
			} break;
		case EState_WritingData:
			{
				// zpet na stav connected se prechazi po odvysilani vseho ve funkci WriteData
				result = (WriteData() > 0);
			} break;
		default: break;
	}

	return result;
}

//-------------------------------------------------------------------------------------------------

void TcpClientEndpoint::OnExceptionEvent(void)
{
	Disconnect();
}

//-------------------------------------------------------------------------------------------------

bool TcpClientEndpoint::Connect(void)
{
	bool result = false;

	Disconnect();

	m_Socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

	if (m_Socket != INVALID_SOCKET)
	{
		// fixme: chyby
#if defined(__POSIX__)
		if (fcntl(m_Socket, F_SETFL, fcntl(m_Socket, F_GETFL, 0) | O_NONBLOCK) != SOCKET_ERROR)
#else
		unsigned long nonBlock = 1;

		if (ioctlsocket(m_Socket, FIONBIO, &nonBlock) != SOCKET_ERROR)
#endif
		{
			sockaddr_in clientService;

			clientService.sin_family = AF_INET;
			clientService.sin_addr.s_addr = m_Ipv4;
			clientService.sin_port = htons(m_Port);

			if (connect(m_Socket, reinterpret_cast<SOCKADDR*>(&clientService), sizeof(clientService)) == -1)
			{
#if defined(_MSC_VER) || defined(__BORLANDC__)
				result = WSAGetLastError() == WSAEWOULDBLOCK;
#elif defined(__POSIX__) || defined(__FREERTOS__)
				result = errno == EINPROGRESS || errno == EALREADY;
#endif
			}
			else
				result = true;
		}
	}

	return (result);
}

//-------------------------------------------------------------------------------------------------

void TcpClientEndpoint::Disconnect(void)
{
	m_WokenAction = ICallbackCommunicationDriver::CallbackAction_None;

	if (m_Socket != INVALID_SOCKET)
	{
#if defined(_MSC_VER) || defined(__BORLANDC__)
		shutdown(m_Socket, SD_BOTH);
#endif
#if defined(__POSIX__)
#if defined(__PIKEOS__)
		shutdown(m_Socket, 2);
#else
		shutdown(m_Socket, SHUT_RDWR);
#endif
#endif
#if defined( __FREERTOS__ )
		//parameter 2 (how) neni v lwip stacku pouzit
		shutdown(m_Socket, 2);
#endif
		closesocket(m_Socket);

		m_Socket = INVALID_SOCKET;
	}
	m_State = EState_NotConnected;
}

//-------------------------------------------------------------------------------------------------

sint32b TcpClientEndpoint::WriteData(void)
{
	sint32b written;
	uint32b toWrite = m_Buffer->DataLen - m_Buffer->Offset;

#if defined(_MSC_VER) || defined(__BORLANDC__)
	written = send(m_Socket, (const char *)(m_Buffer->Buffer + m_Buffer->Offset), toWrite, 0);
#endif
#if defined(__POSIX__)
	written = send(m_Socket, (char *)(m_Buffer->Buffer + m_Buffer->Offset), toWrite, 0);
#endif
#if defined( __FREERTOS__ )
	written = send(m_Socket, (char *)(m_Buffer->Buffer + m_Buffer->Offset), toWrite, 0);
#endif

	if ((uint32b)written == toWrite)
	{
		// cely buffer byl uz odvysilan, prejdeme tedy zpet do stavu connected
		m_State = EState_Connected;
		m_Buffer->Offset = 0;
		m_Buffer->DataLen = 0;
	}
	else
	{
		// odeslal jsem jen neco -> musim si posunout ukazovatka a psat budu, az bude socket volny
		if (written > 0)
		{
			m_Buffer->Offset += written;
		}

#if defined(_MSC_VER) || defined(__BORLANDC__)
		if (written == -1 && WSAGetLastError() != WSAEWOULDBLOCK)
#endif
#if defined(__POSIX__)
			volatile int er = errno;
		if ((written == -1) && (er != EMSGSIZE) && (er != EWOULDBLOCK))
#endif
		{
			Disconnect();
		}
	}
	return (written);
}

//-------------------------------------------------------------------------------------------------

sint32b TcpClientEndpoint::ReadData(void)
{
	sint32b received, result = 0;

	do
	{
		received = recv(m_Socket, (char*)m_Buffer->Buffer + m_Buffer->Offset, m_Buffer->BufferSize - m_Buffer->Offset,
#if defined(__PIKEOS__)
			MSG_DONTWAIT
#else
			0
#endif
			);

		// fixme: we have to see whether the data is complete
		if (received > 0)
		{
			m_Timer.Restart();

			// driver must decide, whether it has everything
			m_Buffer->Offset += received;
			m_Buffer->DataLen = m_Buffer->Offset;
			result += received;

#if !defined(SSCP_STANDALONE_CLIENT)
			ICallbackCommunicationDriver::CallbackAction action = m_Driver->GetNextAction();
#else
			ICallbackCommunicationDriver::CallbackAction action = m_Driver->GetNextAction(m_Tag);
#endif

			if (action != ICallbackCommunicationDriver::CallbackAction_ReadData)
			{
				m_State = EState_Connected;

				break;
			}
		}
	}
	while (received > 0);

	//recv return value of 0 means that the connection was closed
	if (received == 0)
	{
		Disconnect();
	}
	else
	{
		// fixme: cteni telegramu dal...
		// jak do toho zakomponovat chyby? WSAEWOULDBLOCK?
#if defined(_MSC_VER) || defined(__BORLANDC__)
		sint32b error = 0;

		if (received == -1 && (error = WSAGetLastError()) != WSAEWOULDBLOCK)
#else
		if ((received == -1) && (errno != EWOULDBLOCK))
#endif
		{
			Disconnect();
		}
	}
	return result;
}

//-------------------------------------------------------------------------------------------------

#if !defined(SSCP_STANDALONE_CLIENT)
bool TcpClientEndpoint::CheckOneReadCyclePassed(void)
{
	if (m_Driver)
		return (m_Driver->CheckOneReadCyclePassed());
	return (true);
}

//-------------------------------------------------------------------------------------------------

void TcpClientEndpoint::EnableWritingCycles(void)
{
	if (m_Driver)
		m_Driver->EnableWritingCycles();
}

//-------------------------------------------------------------------------------------------------

bool TcpClientEndpoint::ShutdownRequest(bool request)
{
	if (request)
	{
		// get status if it is stopped
		uint8b tmp = DriverCommandInt_Finished;
		OperationStatus res = m_Driver->DriverCommand(DriverCommandType_Internal, &tmp, NULL);
		if (res == OperationStatus_Unspecified)
			return (true);
		return (res == OperationStatus_Ok);
	}
	else
	{
		// inform about stop
		uint8b tmp = DriverCommandInt_StopRequest;
		m_Driver->DriverCommand(DriverCommandType_Internal, &tmp, NULL);
	}
	return (true);
}

//-------------------------------------------------------------------------------------------------
#endif

