#ifndef __TcpClientEndpoint_h
#define __TcpClientEndpoint_h

#include "./datatypes.h"
#if !defined(SSCP_STANDALONE_CLIENT)
#include "./ICommunicationDriver.h"
#endif
#include "./NetworkingHelpers.h"
#include "./StopWatch.h"
#include "./IClientEndpoint.h"

#if defined(SSCP_STANDALONE_CLIENT)
/*!
 * \brief Callback communication driver interface
 * \details
 */
class ICallbackCommunicationDriver
{
	public:
        /*!
         * \brief Callback actions
         * \details Callback actions
         */
		enum CallbackAction
		{
            /*!
             * \brief No action
             */
			CallbackAction_None,
            /*!
             * \brief Write data to PLC
             */
            CallbackAction_WriteData,
            /*!
             * \brief Read data from PLC
             */
            CallbackAction_ReadData
		};

        /*!
         * \brief Destructor
         * \details Make some deinitializations steps
         */
        virtual ~ICallbackCommunicationDriver(void) { };

        /*!
         * \brief Reset
         * \param[in] tag Endpoint ID
         * \details Reset selected driver state
         */
        virtual void Reset(uint32b tag) = 0;
        /*!
         * \brief Get next action
         * \param[in] tag Endpoint ID
         * \return Next action to do
         * \details Check and perform selected drivers next action.
         */
        virtual CallbackAction GetNextAction(uint32b tag) = 0;
        /*!
         * \brief GetCommunicationBuffer
         * \param[in] tag Endpoint ID
         * \return Pointer to data buffer
         * \details Get the drivers communication buffer for selected device
         */
        virtual CommBuffer* GetCommunicationBuffer(uint32b tag) = 0;
};
#endif

/*!
 * \brief TCP client endpoint
 * \details Communication canal with one device
 */
class TcpClientEndpoint : public IClientEndpoint
{
	private:
        /*!
         * \brief Endpoint state
         * \details Device communication status
         */
		enum EndpointState
		{
            /*!
             * \brief Device is not connected
             */
            EState_NotConnected,
            /*!
             * \brief Device is connectiong
             */
            EState_Connecting,
            /*!
             * \brief Device is connected
             */
            EState_Connected,
            //EState_Disconnecting,
            /*!
             * \brief Device is disconnected
             */
            EState_Disconnected,
            /*!
             * \brief Writing data to device
             */
            EState_WritingData,
            /*!
             * \brief Reading data from device
             */
            EState_ReadingData
		};

        /*!
         * \brief IP address
         * \details IP v4 address
         */
		uint32b m_Ipv4;
        /*!
         * \brief Port
         * \details IP Port
         */
        uint32b m_Port;
		// fixme: host, ipv6,...

        /*!
         * \brief Tag
         * \details Endpoint identificator
         */
		uint32b m_Tag;

        /*!
         * \brief Driver
         * \details Device communication driver
         */
		ICallbackCommunicationDriver * m_Driver;

        /*!
         * \brief State
         * \details Communication state
         */
		EndpointState m_State;
        /*!
         * \brief Socket
         * \details TCP IP socket
         */
		SharkSocket m_Socket;

        /*!
         * \brief Buffer
         * \details Communicatioon buffer
         */
		CommBuffer *m_Buffer;

        /*!
         * \brief Buffer max length
         * \details Maximum size of communication buffer (\c m_Buffer)
         */
		uint32b m_CommBufferMaxLength;

        /*!
         * \var TcpClientEndpoint::m_Timer
         * Timer
         */
        /*!
         * \var TcpClientEndpoint::m_IdleTimer
         * Timer
         */
        StopWatch m_Timer, m_IdleTimer;
        /*!
         * \brief Timeout
         * \details Communication timeout in ms
         */
		uint32b m_Timeout;

        /*!
         * \brief Woken action
         * \details Current action
         */
		ICallbackCommunicationDriver::CallbackAction m_WokenAction;

        /*!
         * \brief Connect
         * \return Success flag
         * \details Create socket and connect to the device.
		 * New version of compiler (_MSC_VER >= 1800) uses WinSoc2. It is nessessary to call WSAStartup before initialise socket!
		 * Don't forget to clean WSASocket by calling WSACleanup functioon at the end of your application!
		 * \code{.cpp}
		 * #if _MSC_VER >= 1800
		 *     WSADATA wsaData;
		 *     WSAStartup(MAKEWORD(2, 2), &wsaData);
		 * #endif
		 * \endcode
         */
		bool Connect(void);
	public:
        /*!
         * \brief Disconnect
         * \details Disconnect from device
         */
		virtual void Disconnect(void);
	private:
        /*!
         * \brief Write data
         * \return Writen data size
         * \details Write buffered data to device
         */
		sint32b WriteData(void);
        /*!
         * \brief Read data
         * \return Readed data size
         * \details Read from device to buffer
         */
        sint32b ReadData(void);

	public:
        /*!
         * \brief Constructor
         * \param[in] ipv4 IP Address
         * \param[in] timeoutInMs Timeout in ms
         * \details Create and initializae object
         */
		TcpClientEndpoint(uint32b ipv4, uint32b timeoutInMs);
        /*!
         * \brief Destructor
         * \details Disconnect device
         */
        virtual ~TcpClientEndpoint(void);

        /*!
         * \brief Set parameters
         * \param[in] port IP port
         * \param[in] driver Used driver
         * \details Call extended SetParameters
         */
		void SetParameters(uint32b port, ICallbackCommunicationDriver * driver) { SetParameters(port, driver, 0); }
        /*!
         * \brief Set parameters
         * \param[in] port IP port
         * \param[in] driver Used driver
         * \param[in] tag Endpoint ID
         * \details Set port, driver and id on this object
         */
		void SetParameters(uint32b port, ICallbackCommunicationDriver * driver, uint32b tag);
        /*!
         * \brief Match
         * \param[in] ipv4 IP Address
         * \param[in] port IP port
         * \return
         * \li \c true - Endpoint has selected settings
         * \li \c false - Endpoint hasn't selected settings
         * \details Check selected setting on this device
         */
		virtual bool Match(uint32b ipv4, uint16b port);

		virtual SharkSocket GetSocket(void) { return m_Socket; }
        /*!
         * \brief Get timeout
         * \return Setted timeout
         * \details Get current communication timeout
         */
		uint32b GetTimeout(void) { return m_Timeout; }

        /*!
         * \brief Start
         * \return Success flag
         * \details Starts drive. If driver is started, reset it
         *
         * \sa Connect
         */
		bool Start(void);

		virtual void WantRWAction(bool *wantRead, bool *wantWrite);	

		virtual bool OnReadEvent(void);
		virtual bool OnWriteEvent(void);
		virtual void OnExceptionEvent(void);

		virtual void CheckTimeout(void);

        bool IsConnected() {return (m_State != EState_Connecting && m_State != EState_NotConnected);}
#if !defined(SSCP_STANDALONE_CLIENT)
		virtual bool CheckOneReadCyclePassed(void);
		virtual void EnableWritingCycles(void);
		virtual bool ShutdownRequest(bool request);
#endif

		friend class TcpClientFactory;
};

#endif
