#ifndef __Thread_h
#define __Thread_h

#if defined(_POSIX_THREADS)
#include <pthread.h>
#endif

#include "PlatformUtils.h"
/*!
 * \brief Thread class
 * \details Common class for threading functionality
 * \code{.cpp}
 * \endcode
 * \todo Internal - Odkomentovat
 */
class Thread
{
	protected:
        /*!
         * \brief Running
         * \details Thread safe running flag. \c true means thread is running
         */
		volatile bool m_Running;
        /*!
         * \brief Stopped
         * \details Thread safe stopped flag. \c true means stop the thread
         */
        volatile bool m_Stopped;

#if defined(__FREERTOS__)
		// Bude to stacit? Asi by melo. Funkce SharkPrintf z knihovny newlib vyzaduje pomerne dost mista na zasobniku
		// experimantalne zjisteno 400 bytes je malo, 512 bytes teto funkci staci
		// Pokud zasobnik pretece, tak program havaruje a vyvola se preruseni ktere vola rutinu HardFault_Handler
#if (__FPU_PRESENT == 1) && (__FPU_USED == 1)
		enum { DEFAULT_STACK_SIZE = 2048 + 1024 };
#else
		enum { DEFAULT_STACK_SIZE = /*1400*/2048 };
#endif
#endif

#if defined(_POSIX_THREADS)
        /*!
         * \brief Thread ID
         * \details Thread identificator
         */
		pthread_t m_ThreadId;
#endif

#if defined(__THREAD_ID__)
#if !defined(MAX_THREAD_ID_LENGTH)
#define MAX_THREAD_ID_LENGTH	32
#endif

#if MAX_THREAD_ID_LENGTH < 16
#error MAX_THREAD_ID_LENGTH have to be at least 16 chars
#endif

	public:
        /*!
         * \brief Thread name
         * \details Gets the thread name
         */
		char m_ThreadName[MAX_THREAD_ID_LENGTH];
        /*!
         * \brief Set thread name
         * \details Sets the thread name
         * \todo Proč existuje, když je m_ThreadName public???
         */
        void SetThreadName(const char *threadName);
#endif

	public:
        /*!
         * \brief Constructor
         * \details Initialize class objects
         */
		Thread(void);
        /*!
         * \brief Destructor
         * \details Destroy the object
         * \todo Nešlo by zde udělat stopnutí thredu?
         */
        virtual ~Thread(void);

		// je to volane z ThreadProc, aby se uzivatel nemusel starat o nastavovani flagu
		// nemel by to nikdo z user kodu volat
        /*!
         * \brief On started
         * \details Made actions before running thread
         * \todo nemel by to nikdo z user kodu volat - Proč to není protected?
         */
		void OnStarted(void);
        /*!
         * \brief On stopped
         * \details Stop thread event
         */
		void OnStopped(void);

        /*!
         * \brief Start
         * \param[in] priority Thread priority
         * \return Thread is running flag
         * \details Function for starting thread
         * \todo Virtual se mi tu moc nelíbí, protože tato funkce spustí thread, který volá Execute funkci. Kyž je to takdo, kdekdo si jí může přepsat a k vykonání kódu v threadu nemusí dojít a účel je pak v háji.
         */
		virtual bool Start(uint32b priority);

		// pouze nastavi, ze se thread ma ukoncit (~m_Running = false;)
		// pretizena volani, musi udelat to same
        /*!
         * \brief Set stop signal
         * \details Function sets protected property m_Running to false. Overloaded functions must at least do the same.
         */
		virtual void SetStopSignal(void);

		// ceka, az bude m_Stopped == true; (zajisteno OnStopped())
        /*!
         * \brief Join
         * \param[in] wdtClientId WTD Client ID - used with GLOBAL_WDTMANAGER definition
         * \param[in] wdtPeriod WTD period - used with GLOBAL_WDTMANAGER definition
         * \param[in] timeoutInMs - Waiting timeout in ms
         * \return Thread end signaled (true)
         * \details Wait for end of thread main function fo specified timeout (\c timeoutInMs)
         */
		bool Join(uint8b wdtClientId, uint32b wdtPeriod, uint32b timeoutInMs = 0); // fixme: timeout?

		// signalizace ukonceni zajistena pres OnStopped()
        /*!
         * \brief Execute
         * \details Main execution fünction. This function runs in thread
         * \todo Neměla by být jako protected?
         */
		virtual void Execute(void) = 0;
};

#endif

